#include <fstream>
#include "learning_quadrangulations.h"
using namespace vcg::tri::pl;

typedef typename MT_PolygonEnumerator<PolyMesh>::CornersType      CornersType;

class Notifiable : public MT_PolygonEnumerationNotifiable {
public:
  Notifiable() : len(0) { }

  void notify(NotificationType notificationType, num_type nTrees, count_type nPolygons,
              std::chrono::nanoseconds duration) {
    std::stringstream stream;
    stream << nTrees << " trees completed. ";
    stream << nPolygons << " polygons enumerated. ";
    stream << "Elapsed: ";
    std::string durationStr;
    TimeElements timeElements;
    TimeSupport::SplitTimeElements(duration, timeElements);
    TimeSupport::TimeElementsToString(timeElements, durationStr);
    stream << durationStr.c_str();

    switch (notificationType) {
    case TREE_BUILT:
      stream << " Still enumerating";
      break;
    case FLUSH_DB:
      stream << " Enumeration complete. Flushing";
      break;
    case COMPLETE:
      stream << " Enumeration complete!";
      break;
    default:
      break;
    }

    if (notificationType != EMPTY && notificationType != COMPLETE)
      for (int i = 0; i <= timeElements.s.count() % 3; ++i)
        stream << '.';

    std::cout << '\r';
    for (size_t c = 0; c < len; ++c)
      std::cout << ' ';
    len = stream.str().length();
    std::cout << '\r' << stream.str();
    if (notificationType == EMPTY || notificationType == COMPLETE)
      std::cout << '\n';
    std::cout.flush();
  }

private:
  size_t len;
};

/**
 * @brief The Parameters struct
 */
struct Parameters {
  std::string       meshFilename;           // -mesh <filename>
  std::string       meshList;               // -list <filename>
  std::string       patchDBFilename;        // -patchDB <filename>
  num_type          minNFaces;              // -minF <#>
  num_type          maxNFaces;              // -maxF <#>
  CornersType       cornersType;            // -corners [convex | concave | irregular | any]
  count_type        insertBufferSize;       // -ibs <#>
  count_type        updateBufferSize;       // -ubs <#>
  bool              useRegister;            // -reg <yes | no>
  count_type        registerBufferSize;     // -rbs <#>
  std::string       polygonDBFilename;      // -polygonDB <filename>
  count_type        polygonCounterSize;     // -pcs <#>
  u_small_int_type  nThreads;               // -nt <#>

  /**
   * @brief Parameters Default constructor.
   */
  Parameters() : patchDBFilename(PatchDBInterface::PATCHDBFILENAME())
               , minNFaces(0)
               , maxNFaces(std::numeric_limits<num_type>::max())
               , cornersType(CornersType::CONVEX)
               , insertBufferSize(3000000)
               , updateBufferSize(300000000)
               , useRegister(true)
               , registerBufferSize(300000000)
               , polygonDBFilename(PolygonDBInterface::POLYGONDBFILENAME())
               , polygonCounterSize(0)
               , nThreads(std::thread::hardware_concurrency()) { }

  /**
   * @brief PrintUsage
   */
  static void PrintUsage(const char *name) {
    std::string cmd(name);
    size_t pos = cmd.find_last_of("/\\");
    if (pos != std::string::npos)
      cmd = cmd.substr(pos+1);
    std::cout << "Learning Quadrangulations." << std::endl << std::endl;
    std::cout << "Usage:" << std::endl;
    std::cout << "\t" << cmd << " [options] -mesh|-list <filename>" << std::endl << std::endl;
    std::cout << "Options:" << std::endl;
    std::cout << "-patchDB <filename>                       \tDefaults to '" << PatchDBInterface::PATCHDBFILENAME() << "'" << std::endl;
    std::cout << "-minF <n>                                 \tDefaults to 0" << std::endl;
    std::cout << "-maxF <n>                                 \tDefaults to " << std::numeric_limits<num_type>::max() << std::endl;
    std::cout << "-corners <convex|concave|irregular|any>   \tDefaults to 'convex'" << std::endl;
    std::cout << "-ibs <n>                                  \tDefaults to 3000000" << std::endl;
    std::cout << "-ubs <n>                                  \tDefaults to 300000000" << std::endl;
    std::cout << "-reg <yes | no>                           \tDefaults to yes" << std::endl;
    std::cout << "-rbs <n>                                  \tDefaults to 300000000" << std::endl;
    std::cout << "-polygonDB <filename>                     \tDefaults to '" << PolygonDBInterface::POLYGONDBFILENAME() << "'" << std::endl;
    std::cout << "-pcs <n>                                  \tDefaults to 0" << std::endl;
    std::cout << "-nt <n>                                   \tDefaults to " << std::thread::hardware_concurrency() << std::endl;
    std::cout << std::endl;
  }

  /**
   * @brief CheckParameters reads parameters from command line and sets them.
   * @param argc
   * @param argv
   * @param parameters
   * @return 0 if all ok, -1 otherwise.
   */
  static int CheckParameters(int argc, char *argv[], Parameters &parameters) {
    for (int i = 1; i < argc; ++i) {
      if (argv[i] == std::string("-mesh")) {
        ++i;
        if (i < argc)
          parameters.meshFilename = argv[i];
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-list")) {
        ++i;
        if (i < argc)
          parameters.meshList = argv[i];
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-patchDB")) {
        ++i;
        if (i < argc)
          parameters.patchDBFilename = argv[i];
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-minF")) {
        ++i;
        if (i < argc)
          parameters.minNFaces = ::atoi(argv[i]);
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-maxF")) {
        ++i;
        if (i < argc)
          parameters.maxNFaces = ::atoi(argv[i]);
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-corners")) {
        ++i;
        if (i < argc) {
          if (argv[i] == std::string("convex"))
            parameters.cornersType = CornersType::CONVEX;
          else if (argv[i] == std::string("concave"))
            parameters.cornersType = CornersType::CONCAVE;
          else if (argv[i] == std::string("irregular"))
            parameters.cornersType = CornersType::IRREGULAR;
          else if (argv[i] == std::string("any"))
            parameters.cornersType = CornersType::ANY;
          else {
            PrintUsage(argv[0]);
            return -1;
          }
        } else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-ibs")) {
        ++i;
        if (i < argc)
          parameters.insertBufferSize = ::atoi(argv[i]);
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-ubs")) {
        ++i;
        if (i < argc)
          parameters.updateBufferSize = ::atoi(argv[i]);
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-reg")) {
        ++i;
        if (i < argc) {
          if (argv[i] == std::string("yes"))
            parameters.useRegister = true;
          else if (argv[i] == std::string("no"))
            parameters.useRegister = false;
          else {
            PrintUsage(argv[0]);
            return -1;
          }
        } else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-rbs")) {
        ++i;
        if (i < argc)
          parameters.registerBufferSize = ::atoi(argv[i]);
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-polygonDB")) {
        ++i;
        if (i < argc)
          parameters.polygonDBFilename = argv[i];
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-mesh")) {
        ++i;
        if (i < argc)
          parameters.meshFilename = argv[i];
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-pcs")) {
        ++i;
        if (i < argc)
          parameters.polygonCounterSize = ::atoi(argv[i]);
        else {
          PrintUsage(argv[0]);
          return -1;
        }
      } else if (argv[i] == std::string("-nt")) {
        ++i;
        if (i < argc)
          parameters.nThreads = ::atoi(argv[i]);
        else {
          PrintUsage(argv[0]);
          return -1;
        }
        if (parameters.nThreads < 1) {
          PrintUsage(argv[0]);
          return -1;
        }
      } else {
        PrintUsage(argv[0]);
        return -1;
      }
    }
    if ((parameters.meshFilename.empty() && parameters.meshList.empty()) ||
        (!parameters.meshFilename.empty() && !parameters.meshList.empty())) {
      PrintUsage(argv[0]);
      return -1;
    }
    return 0;
  }
};

int main(int argc, char *argv[]) {
  std::ifstream listStream;
  TS_PolygonRegister<PolyMesh> *polygonRegister = 0;
  TS_PatchLearner<PolyMesh> patchLearner;
  MT_PolygonEnumerator<PolyMesh> polygonEnumerator;
  MT_PolygonEnumerator<PolyMesh>::MT_Notifiable_Ptr notifiable;
  PolyMesh mesh;
  Parameters parameters;
  notifiable = std::dynamic_pointer_cast<MT_PolygonEnumerationNotifiable>(std::shared_ptr<Notifiable>(new Notifiable));

  // read parameters
  if (Parameters::CheckParameters(argc, argv, parameters) != 0)
    return 0;

  if (!parameters.meshList.empty()) {
    listStream.open(parameters.meshList.c_str());
    if (!listStream.is_open()) {
      std::cout << "Error. File '" << parameters.meshList << "' not open." << std::endl;
      return 1;
    }
  }

  patchLearner.setPatchDBFilename(parameters.patchDBFilename);
  patchLearner.setInsertBufferSize(parameters.insertBufferSize);
  patchLearner.setUpdateBufferSize(parameters.updateBufferSize);
  patchLearner.setPolygonDBFilename(parameters.polygonDBFilename);
  patchLearner.setMeshFilename(parameters.meshFilename);
  patchLearner.setPolygonCounterSize(parameters.polygonCounterSize);
  if (parameters.useRegister) {
    polygonRegister = new TS_PolygonRegister<PolyMesh>();
    polygonRegister->setBufferSize(parameters.registerBufferSize);
  }
  polygonEnumerator.setNumberOfThreads(parameters.nThreads);
  polygonEnumerator.registerNotifiable(notifiable);

  do {
    if (listStream.is_open())
      std::getline(listStream, parameters.meshFilename);

    if (MeshSupport<PolyMesh>::OpenMesh(mesh, parameters.meshFilename.c_str()) != 0) {
      std::cout << "File '" << parameters.meshFilename << "' not open." << std::endl;
      continue;
    }

    std::cout << "Learning mesh '" << parameters.meshFilename << "' with " << mesh.face.size()
              << " faces into DB '" << parameters.patchDBFilename << "' with max " << parameters.maxNFaces
              << " faces and using " << parameters.nThreads << " threads. Counter size: " << parameters.polygonCounterSize
              << std::endl;

    patchLearner.initialize();
    if (polygonRegister) {
      polygonRegister->initialize();
      polygonRegister->resetDB();
    }


    polygonEnumerator.enumerate(mesh, polygonRegister, patchLearner, parameters.minNFaces, parameters.maxNFaces,
                                parameters.cornersType, (parameters.polygonCounterSize > 0));
    polygonEnumerator.waitForThreads();

    std::cout << "End of processing." << std::endl;
    std::cout << "Found " << polygonEnumerator.numberOfPolygons() << " polygons with " <<
                 polygonEnumerator.maxNumberOfSides() << " max number of sides and " <<
                 polygonEnumerator.maxNumberOfSingularities() << " max number of singularities, " <<
                 patchLearner.getNumberOfPatches() << " patches (" << patchLearner.getNumberOfNewPatches() << " new)." <<
                 std::endl << std::endl;
  } while (listStream.is_open() && !listStream.eof());

  if (polygonRegister) {
    polygonRegister->finalize();
    delete polygonRegister;
    polygonRegister = 0;
  }
  patchLearner.finalize();
  if (listStream.is_open())
    listStream.close();

  return 0;
}

