#include "glarea.h"
#if __cplusplus >= 201103L
  #include <utility>
#else
  #include <algorithm>
#endif
#include <QKeyEvent>
#include <QMouseEvent>
#include <QWheelEvent>
#include <QStaticText>
#include <QSvgGenerator>
#include <QInputDialog>
#if QT_VERSION >= QT_VERSION_CHECK(5,0,0)
  #include <QGuiApplication>
#else
  #include <QApplication>
#endif
#if defined(__APPLE__) || defined(MACOSX)
  #include <GLUT/glut.h>
  #include <glu.h>
#else
  #include <GL/glut.h>
  #include <GL/glu.h>
#endif
#include <wrap/gl/space.h>
#include <wrap/qt/trackball.h>
#include <wrap/gl/picking.h>
#include <wrap/gl/addons.h>
#include "mesh_utilities.h"

/**
 * @brief GLArea::GLArea
 * @param parent
 */
GLArea::GLArea(QWidget *parent) : QGLWidget(QGLFormat(QGL::SampleBuffers), parent)
                                , _inputMode(VIEW)
                                , _isTrackballActive(true)
                                , _showTrackball(true)
                                , _renderData(0)
                                #if QT_VERSION >= QT_VERSION_CHECK(5,0,0)
                                  , _font(QGuiApplication::font())
                                #else
                                  _font(QApplication::font());
                                #endif
                                , _hasToPick(false)
                                , _startNewPolyLine(false)
                                , _endPolyLine(false)
                                , _nSamples(NSAMPLES) {
  _trackball.center = vcg::Point3f(0, 0, 0);
  _trackball.radius = 0.8;
  _font.setFamily("Palatino");
  _font.setItalic(true);
  _font.setPointSize(18);
  _picked2DPoint.SetZero();
}

/**
 * @brief GLArea::clearButtons releases any pressed buttons.
 */
void GLArea::clearButtons() {
  _trackball.ButtonUp(QT2VCG(Qt::NoButton, Qt::ControlModifier));
  _trackball.ButtonUp(QT2VCG(Qt::NoButton, Qt::ShiftModifier));
  _trackball.ButtonUp(QT2VCG(Qt::NoButton, Qt::AltModifier));
  update();
}

/**
 * @brief GLArea::setRenderData
 * @param renderData
 */
void GLArea::setRenderData(RenderData *renderData) {
  _renderData = renderData;
  _trackball.Reset();
  clearButtons();
}

/**
 * @brief GLArea::setInputMode
 * @param im
 */
void GLArea::setInputMode(InputMode im) {
  _inputMode = im;
  clearButtons();
}

/**
 * @brief GLArea::setNumberOfSamples
 * @param nSamples
 */
void GLArea::setNumberOfSamples(const size_t nSamples) {
  _nSamples = nSamples;
}

/**
 * @brief GLArea::resetTrackball
 */
void GLArea::resetTrackball() {
  _trackball.Reset();
  clearButtons();
}

/**
 * @brief GLArea::setTrackballActive
 * @param active
 */
void GLArea::setTrackballActive(bool active) {
  _isTrackballActive = active;
  clearButtons();
}

/**
 * @brief GLArea::setTrackballVisible
 * @param visible
 */
void GLArea::setTrackballVisible(bool visible) {
  _showTrackball = visible;
  clearButtons();
}

/**
 * @brief GLArea::takeSnapShot saves the current opengl buffer on a disk file.
 * @param filename
 */
void GLArea::takeSnapShot(const QString &filename) {
  // read pixels from buffer
  unsigned char *pixelsInv = new unsigned char[4 * QTDeviceWidth(this) * QTDeviceHeight(this)];
  glReadPixels(0, 0, (GLint)QTDeviceWidth(this), (GLint)QTDeviceHeight(this), GL_RGBA, GL_UNSIGNED_BYTE, pixelsInv);

  // create an image
  QImage image(pixelsInv, QTDeviceWidth(this), QTDeviceHeight(this), QImage::Format_RGBA8888_Premultiplied);

  // swap upside down
  image = image.mirrored(false, true);

  // save the image on file
  QString ext("png");
  QString savename(filename);
  if (savename.right(4) != QString('.') + ext)
    savename += QString('.') + ext;
  image.save(savename, ext.toStdString().c_str());

  delete[] pixelsInv;
}

/**
 * @brief GLArea::initializeGL
 */
void GLArea::initializeGL() {
  glewInit();
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_NORMALIZE);
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_MULTISAMPLE);
  glEnable( GL_LINE_SMOOTH );
  glHint( GL_LINE_SMOOTH_HINT, GL_NICEST );
  glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_NICEST);
  glEnable(GL_POINT_SMOOTH);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA_SATURATE, GL_ONE);
  glClearColor(1.f, 1.f, 1.f, 1.f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

/**
 * @brief GLArea::resizeGL
 * @param w
 * @param h
 */
void GLArea::resizeGL(int w, int h) {
  glViewport(0, 0, QTLogicalToDevice<GLsizei>(this, w), QTLogicalToDevice<GLsizei>(this, h));
  update();
}

/**
 * @brief GLArea::paintEvent
 * @param event
 */
void GLArea::paintEvent(QPaintEvent *event) {
  Q_UNUSED(event)

  GLsizei previewWindowSize = 0, previewWindowX = 0;

  QString str;
  QPainter painter(this);
  painter.setFont(_font);
  painter.setRenderHints(QPainter::Antialiasing | QPainter::HighQualityAntialiasing);
  painter.beginNativePainting();
  makeCurrent();

  glViewport(0, 0, QTLogicalToDevice<GLsizei>(this, width()), QTLogicalToDevice<GLsizei>(this, height()));
  glEnable(GL_DEPTH_TEST);
  glClearColor(1.f, 1.f, 1.f, 1.f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  prepareOpenGL(QTLogicalToDevice<GLsizei>(this, width()), QTLogicalToDevice<GLsizei>(this, height()));
  _trackball.GetView();
  _trackball.Apply();

  ScalarType scaleFactor;
  PointType center;

  if (!_renderData ||
      ((_renderData->currentMesh >= _renderData->meshes.size() || _renderData->meshes[_renderData->currentMesh].mesh.IsEmpty())
       && _renderData->tMesh.IsEmpty())) {
    if (_renderData && !_renderData->boundary.sides.empty())
      renderBoundary();
    if (_showTrackball)
      _trackball.DrawPostApply();
    painter.endNativePainting();
    return;
  }

  // draw main data
  if (!_renderData->tMesh.IsEmpty()) {
    scaleFactor = 2.5f / _renderData->tMesh.bbox.Diag();
    center = _renderData->tMesh.bbox.Center();
  } else {
    scaleFactor = 2.5f / _renderData->meshes[_renderData->currentMesh].mesh.bbox.Diag();
    center = _renderData->meshes[_renderData->currentMesh].mesh.bbox.Center();
  }
  if (_renderData->showTriMesh && !_renderData->tMesh.IsEmpty())
    renderTriMesh(scaleFactor, center);
  if (_renderData->showProvMesh) {
    renderData(&painter, _renderData->provMesh, 2.5f / _renderData->provMesh.mesh.bbox.Diag(),
               _renderData->provMesh.mesh.bbox.Center(), _renderData->showVertices, false, false);
  } else {
    renderData(&painter, _renderData->meshes[_renderData->currentMesh], scaleFactor, center,
               _renderData->showVertices, _renderData->showStrings, true);
  }
  if (!_renderData->flows.empty() && !_renderData->meshes[_renderData->currentMesh].polychordMap.empty()) {
    enter2D(&painter);
    _font.setPointSize(22);
    painter.setFont(_font);
    str = QString::number(_renderData->meshes[_renderData->currentMesh].value);
    QRectF rect(painter.fontMetrics().boundingRect(str));
    rect.moveCenter(QPointF(width() / 2, height() - rect.height() / 2 - 1));
    painter.setPen(QColor(0, 0, 255));
    painter.drawText(rect, str);
    exit2D(&painter);
    _font.setPointSize(18);
    painter.setFont(_font);
  }

  if (_showTrackball)
    _trackball.DrawPostApply();

  // draw preview
  if (_renderData->showPreview && _renderData->meshes.size() > 0) {
    glClear(GL_DEPTH_BUFFER_BIT);
    previewWindowSize = std::min<GLsizei>(QTDeviceHeight(this) / 5, QTDeviceWidth(this) / _renderData->meshes.size());
    previewWindowX = (QTDeviceWidth(this) - previewWindowSize * _renderData->meshes.size()) / 2;
    for (size_t i = 0; i < _renderData->meshes.size(); ++i, previewWindowX += previewWindowSize) {
      // draw rectangle
      if (!_renderData->meshes[i].mesh.IsEmpty()) {
        enter2D(&painter);
        if (i == _renderData->currentMesh)
          painter.setPen(QColor(0, 0, 255));
        else
          painter.setPen(QColor(250, 250, 250));
        painter.setBrush(QBrush(QColor(250, 250, 250)));
        painter.drawRect(QTDeviceToLogical(this, previewWindowX), 0,
                         QTDeviceToLogical(this, previewWindowSize) - 1, QTDeviceToLogical(this, previewWindowSize));
        if (!_renderData->flows.empty() && !_renderData->meshes[i].polychordMap.empty()) {
          str = QString::number(_renderData->meshes[i].value);
          QRectF rect(painter.fontMetrics().boundingRect(str));
          rect.moveCenter(QPointF(QTDeviceToLogical(this, previewWindowX + previewWindowSize / 2),
                                  QTDeviceToLogical(this, previewWindowSize) - rect.height() / 2 - 1));
          if (i == _renderData->currentMesh)
            painter.setPen(QColor(0, 0, 255));
          else
            painter.setPen(QColor(250, 0, 0));
          painter.drawText(rect, str);
        }
        exit2D(&painter);
        // draw mesh
        painter.translate(0, QTDeviceToLogical(this, QTDeviceHeight(this) - previewWindowSize - 1));
        glViewport(previewWindowX, QTDeviceHeight(this) - previewWindowSize - 1, previewWindowSize, previewWindowSize);
        prepareOpenGL(previewWindowSize, previewWindowSize);
        renderData(&painter, _renderData->meshes[i],  scaleFactor, center, _renderData->showVertices, false, true);
        painter.translate(0, - QTDeviceToLogical(this, QTDeviceHeight(this) - previewWindowSize - 1));
      }
    }
  }

  // draw base mesh
  if (_renderData->showBaseMesh && !_renderData->baseMesh.mesh.IsEmpty()) {
    glClear(GL_DEPTH_BUFFER_BIT);
    previewWindowSize = QTDeviceHeight(this) / 5;
    previewWindowX = QTDeviceWidth(this) - previewWindowSize - 1;
    // draw rectangle
    enter2D(&painter);
    painter.setPen(QColor(0, 0, 255));
    painter.setBrush(QBrush(QColor(250, 250, 250)));
    painter.drawRoundRect(QTDeviceToLogical(this, previewWindowX), QTDeviceToLogical(this, QTDeviceHeight(this) - previewWindowSize) - 1,
                          QTDeviceToLogical(this, previewWindowSize), QTDeviceToLogical(this, previewWindowSize));
    exit2D(&painter);
    // draw base mesh
    painter.translate(0, QTDeviceToLogical(this, QTDeviceHeight(this) - previewWindowSize - 1));
    glViewport(previewWindowX, 0, previewWindowSize, previewWindowSize);
    prepareOpenGL(previewWindowSize, previewWindowSize);
    renderData(&painter, _renderData->baseMesh,  2.5f / _renderData->baseMesh.mesh.bbox.Diag(), _renderData->baseMesh.mesh.bbox.Center(),
               _renderData->showVertices, _renderData->showStrings, false);
    painter.translate(0, - QTDeviceToLogical(this, QTDeviceHeight(this) - previewWindowSize - 1));
  }

  painter.endNativePainting();
}

void GLArea::keyPressEvent(QKeyEvent *e) {
  switch (_inputMode) {
  case VIEW:
    if (_isTrackballActive) {
      if (e->key() == Qt::Key_Control)
        _trackball.ButtonDown(QT2VCG(Qt::NoButton, Qt::ControlModifier));
      if (e->key() == Qt::Key_Shift)
        _trackball.ButtonDown(QT2VCG(Qt::NoButton, Qt::ShiftModifier));
      if (e->key() == Qt::Key_Alt)
        _trackball.ButtonDown(QT2VCG(Qt::NoButton, Qt::AltModifier));
    }
    break;
  case DRAW2DLINES:
    break;
  default:
    break;
  }
  e->ignore();
  update();
}

void GLArea::keyReleaseEvent(QKeyEvent *e) {
  if (e->key() == Qt::Key_C) {
    char cam_str[512];
    _trackball.ToAscii(cam_str);
    QString camera(cam_str);
    bool ok = false;
    camera = QInputDialog::getText(0, tr("Camera settings"), tr("Set camera"), QLineEdit::Normal, camera, &ok);
    if (ok && !camera.isEmpty())
      _trackball.SetFromAscii(camera.toStdString().c_str());
    e->accept();
    return;
  }
  switch (_inputMode) {
  case VIEW:
    if (_isTrackballActive) {
      if (e->key() == Qt::Key_Control)
        _trackball.ButtonUp(QT2VCG(Qt::NoButton, Qt::ControlModifier));
      if (e->key() == Qt::Key_Shift)
        _trackball.ButtonUp(QT2VCG(Qt::NoButton, Qt::ShiftModifier));
      if (e->key() == Qt::Key_Alt)
        _trackball.ButtonUp(QT2VCG(Qt::NoButton, Qt::AltModifier));
    }
    break;
  case DRAW2DLINES:
    break;
  default:
    break;
  }
  e->ignore();
  update();
}

void GLArea::mousePressEvent(QMouseEvent *e) {
  switch (_inputMode) {
  case VIEW:
    if (_isTrackballActive) {
      setFocus();
      _trackball.MouseDown(QT2VCG_X(this, e), QT2VCG_Y(this, e), QT2VCG(e->button(), e->modifiers()));
    }
    break;
  case DRAW2DLINES:
    _hasToPick = true;
    _startNewPolyLine = true;
    _picked2DPoint = vcg::Point2i(QT2VCG_X(this, e), QT2VCG_Y(this, e));
    update();
    break;
  default:
    break;
  }
  e->accept();
  update();
}

void GLArea::mouseMoveEvent(QMouseEvent *e) {
  switch (_inputMode) {
  case VIEW:
    if (_isTrackballActive && e->buttons())
      _trackball.MouseMove(QT2VCG_X(this, e), QT2VCG_Y(this, e));
    break;
  case DRAW2DLINES:
    _hasToPick = true;
    _picked2DPoint = vcg::Point2i(QT2VCG_X(this, e), QT2VCG_Y(this, e));
    update();
    break;
  default:
    break;
  }
  e->accept();
  update();
}

void GLArea::mouseReleaseEvent(QMouseEvent *e) {
  switch (_inputMode) {
  case VIEW:
    if (_isTrackballActive)
      _trackball.MouseUp(QT2VCG_X(this, e), QT2VCG_Y(this, e), QT2VCG(e->button(), e->modifiers()));
    break;
  case DRAW2DLINES:
    _hasToPick = true;
    _endPolyLine = true;
    _picked2DPoint = vcg::Point2i(QT2VCG_X(this, e), QT2VCG_Y(this, e));
    update();
    break;
  default:
    break;
  }
  e->accept();
  update();
}

void GLArea::wheelEvent(QWheelEvent *e) {
  const int WHEEL_STEP = 120;
  switch (_inputMode) {
  case DRAW2DLINES:
  case VIEW:
    _trackball.MouseWheel(e->delta() / (float)WHEEL_STEP, QTWheel2VCG(e->modifiers()));
    break;
  default:
    break;
  }
  e->accept();
  update();
}

/**
 * @brief GLArea::enter2D
 * @param painter
 */
void GLArea::enter2D(QPainter *painter) {
  glPushAttrib(GL_ENABLE_BIT|GL_VIEWPORT_BIT);
  glDisable(GL_DEPTH_TEST);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  painter->endNativePainting();
  painter->save();
}

/**
 * @brief GLArea::exit2D
 * @param painter
 */
void GLArea::exit2D(QPainter *painter) {
  painter->restore();
  painter->beginNativePainting();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glPopAttrib();
}

/**
 * @brief GLArea::drawRichText
 * @param painter
 * @param p
 * @param text
 * @param c
 */
void GLArea::drawRichText(QPainter *painter, const PointType &p, const QString &text, const vcg::Color4b &c) {
  GLdouble model[16];
  GLdouble proj[16];
  GLint view[4];
  glGetDoublev(GL_MODELVIEW_MATRIX, model);
  glGetDoublev(GL_PROJECTION_MATRIX, proj);
  glGetIntegerv(GL_VIEWPORT, view);
  GLdouble winx, winy, winz;
  gluProject(p[0], p[1], p[2], model, proj, view, &winx, &winy, &winz);
  QStaticText t(text);
  t.setTextFormat(Qt::RichText);
  t.prepare(QTransform(), painter->font());
  QPointF point(QTDeviceToLogical(painter, winx) - t.size().width()/2, QTDeviceToLogical(painter, view[3] - winy) - t.size().height()/2);
  enter2D(painter);
  painter->setPen(QPen(QBrush(QColor(c[0], c[1], c[2], c[3])), 2, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
  painter->drawStaticText(point, t);
  exit2D(painter);
}

/**
 * @brief GLArea::prepareOpenGL
 * @param w
 * @param h
 */
void GLArea::prepareOpenGL(const float w, const float h) {
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(30, w / h, 1, 20);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(0,0,5,   0,0,0,   0,1,0);
  glPointSize(10);
}

/**
 * @brief GLArea::pickOnPlaneGL
 * @param pickedPoint
 */
void GLArea::pickOnPlaneGL(PointType &pickedPoint) {
  PointType nearestPoint, farthestPoint;
  GLdouble res[3];
  GLdouble mm[16], pm[16];
  GLint vp[4];
  GLfloat depthrange[2] = {0, 0};
  glGetDoublev(GL_MODELVIEW_MATRIX, mm);
  glGetDoublev(GL_PROJECTION_MATRIX, pm);
  glGetIntegerv(GL_VIEWPORT, vp);
  glGetFloatv(GL_DEPTH_RANGE, depthrange);
  gluUnProject(_picked2DPoint.X(), _picked2DPoint.Y(), depthrange[0], mm, pm, vp, &res[0], &res[1], &res[2]);
  nearestPoint[0] = res[0];
  nearestPoint[1] = res[1];
  nearestPoint[2] = res[2];
  gluUnProject(_picked2DPoint.X(), _picked2DPoint.Y(), depthrange[1], mm, pm, vp, &res[0], &res[1], &res[2]);
  farthestPoint[0] = res[0];
  farthestPoint[1] = res[1];
  farthestPoint[2] = res[2];
  vcg::Plane3<ScalarType> plane(ScalarType(0), PointType(0, 0, 1));
  vcg::Line3<ScalarType> pickedDir(nearestPoint, (farthestPoint - nearestPoint).Normalize());
  vcg::IntersectionPlaneLine(plane, pickedDir, pickedPoint);
}

/**
 * @brief GLArea::nearestBorder
 * @param p
 * @param closestPoint
 * @return
 */
n_corners_type GLArea::nearestSide(const PointType &p, PointType &closestPoint, num_type &edgeIndex) {
  PointConstIterator it, nextIt;
  PointType tmpPoint(0, 0, 0);
  ScalarType dist, bestDist = std::numeric_limits<ScalarType>::max();
  n_corners_type nSides = _renderData->boundary.sides.size();
  n_corners_type bestIndex = std::numeric_limits<n_corners_type>::max();
  num_type eIndex = 0;
  for (n_corners_type i = 0; i < nSides; ++i) {
    for (it = _renderData->boundary.sides[i].begin(), nextIt = it, ++nextIt, eIndex = 0;
         nextIt != _renderData->boundary.sides[i].end();
         it = nextIt, ++nextIt, ++eIndex) {
      vcg::SegmentPointDistance(vcg::Segment3<ScalarType>(*it, *nextIt), p, tmpPoint, dist);
      if (dist < bestDist) {
        bestDist = dist;
        closestPoint = tmpPoint;
        bestIndex = i;
        edgeIndex = eIndex;
      }
    }
  }
  if (bestDist <= _renderData->pickRadius)
    return bestIndex;
  edgeIndex = -1;
  closestPoint = p;
  return std::numeric_limits<n_corners_type>::max();
}

/**
 * @brief GLArea::picking
 */
void GLArea::picking() {
  vcg::Color4b c;
  PointType picked3DPoint, closestPoint;
  FlowMapIterator flowsIt;
  RenderData::FlowColorsMap::iterator flowsColorsIt;
  if (_hasToPick) {
    pickOnPlaneGL(picked3DPoint);
    if (_startNewPolyLine) {
      _lastFlowTopology.startSide = nearestSide(picked3DPoint, closestPoint, _lastFlowPoints.startEdge);
      _lastFlowTopology.endSide = std::numeric_limits<n_corners_type>::max();
      _lastFlowPoints.points.clear();
      _lastFlowPoints.points.push_back(closestPoint);
      _startNewPolyLine = false;
    } else if (_endPolyLine) {
      if (_lastFlowPoints.points.size() > 2) {
        if (_lastFlowTopology.startSide >= _renderData->boundary.sides.size()) {
          if ((picked3DPoint - _lastFlowPoints.points.front()).Norm() <= _renderData->pickRadius) {
            _lastFlowPoints.points.push_back(_lastFlowPoints.points.front());
            for (int i = 0; i < 3; ++i)
              for (size_t p = 1; p < _lastFlowPoints.points.size() - 1; ++p)
                _lastFlowPoints.points[p] = (_lastFlowPoints.points[p-1] + _lastFlowPoints.points[p] + _lastFlowPoints.points[p+1]) / 3.0;
            _lastFlowPoints.updateArcLengths();
            _lastFlowPoints.computeSamples(_nSamples);
            flowsIt = _renderData->flows.find(_lastFlowTopology);
            if (flowsIt == _renderData->flows.end())
              _renderData->flows.insert(std::make_pair(_lastFlowTopology, FlowContainer(1, _lastFlowPoints)));
            else
              flowsIt->second.push_back(_lastFlowPoints);
            c = vcg::Color4b::Scatter(_renderData->boundary.sides.size() + 20, _renderData->boundary.sides.size() + _renderData->nFlows);
            ++_renderData->nFlows;
            flowsColorsIt = _renderData->flowsColors.find(_lastFlowTopology);
            if (flowsColorsIt == _renderData->flowsColors.end())
              _renderData->flowsColors.insert(std::make_pair(_lastFlowTopology, RenderData::FlowColors(1, c)));
            else
              flowsColorsIt->second.push_back(c);
          }
        } else {
          _lastFlowTopology.endSide = nearestSide(picked3DPoint, closestPoint, _lastFlowPoints.endEdge);
          if (_lastFlowTopology.endSide < _renderData->boundary.sides.size()) {
            _lastFlowPoints.points.push_back(closestPoint);
            for (int i = 0; i < 3; ++i)
              for (size_t p = 1; p < _lastFlowPoints.points.size() - 1; ++p)
                _lastFlowPoints.points[p] = (_lastFlowPoints.points[p-1] + _lastFlowPoints.points[p] + _lastFlowPoints.points[p+1]) / 3.0;
            if (_lastFlowTopology.endSide < _lastFlowTopology.startSide) {
              _lastFlowTopology.swap();
              _lastFlowPoints.invertPoints();
            } else if (_lastFlowTopology.startSide == _lastFlowTopology.endSide && _lastFlowPoints.endEdge < _lastFlowPoints.startEdge) {
              _lastFlowPoints.invertPoints();
            }
            _lastFlowPoints.updateArcLengths();
            _lastFlowPoints.computeSamples(_nSamples);
            flowsIt = _renderData->flows.find(_lastFlowTopology);
            if (flowsIt == _renderData->flows.end())
              _renderData->flows.insert(std::make_pair(_lastFlowTopology, FlowContainer(1, _lastFlowPoints)));
            else
              flowsIt->second.push_back(_lastFlowPoints);
            c = vcg::Color4b::Scatter(_renderData->boundary.sides.size() + 20, _renderData->boundary.sides.size() + _renderData->nFlows);
            ++_renderData->nFlows;
            flowsColorsIt = _renderData->flowsColors.find(_lastFlowTopology);
            if (flowsColorsIt == _renderData->flowsColors.end())
              _renderData->flowsColors.insert(std::make_pair(_lastFlowTopology, RenderData::FlowColors(1, c)));
            else
              flowsColorsIt->second.push_back(c);
          }
        }
      }
      _lastFlowPoints.startEdge = -1;
      _lastFlowPoints.endEdge = -1;
      _lastFlowPoints.points.clear();
      _lastFlowPoints.arcLengths.clear();
      _lastFlowPoints.samplePoints.clear();
      _endPolyLine = false;
    } else {
      if ((picked3DPoint - _lastFlowPoints.points.back()).Norm() > _renderData->pickRadius)
        _lastFlowPoints.points.push_back(picked3DPoint);
    }
    _hasToPick = false;
  }
}

/**
 * @brief GLArea::renderTriMesh
 * @param tMesh
 */
void GLArea::renderTriMesh(const ScalarType scaleFactor, const PointType &center) {
  FlowMapConstIterator flowMapIt;
  RenderData::FlowColorsMap::const_iterator flowColorsMapIt;
  PointConstIterator it, nextIt;

  if (_renderData->tMesh.IsEmpty())
    return;

  glPushAttrib(GL_ALL_ATTRIB_BITS);
  glPushMatrix();
  vcg::glScale(scaleFactor);
  glTranslate(-center);

  // put here mesh painting code
  glEnable(GL_LIGHTING);
  glPolygonOffset(1.0f, 1.0f);          // set offset
  glEnable(GL_POLYGON_OFFSET_FILL);     // turn on offset
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  for (size_t i = 0; i < _renderData->tMesh.face.size(); ++i) {
    if (_renderData->tMesh.face[i].IsD())
      continue;

    vcg::glColor(_renderData->tMesh.face[i].cC());
    PointType n = _renderData->tMesh.face[i].cN();
    glBegin(GL_POLYGON);
    for (int j = 0; j < _renderData->tMesh.face[i].VN(); ++j) {
      vcg::glNormal(n);
      vcg::glVertex(_renderData->tMesh.face[i].cP(j));
    }
    glEnd();
  }
  glDisable(GL_POLYGON_OFFSET_FILL);    // turn off offset

  glDisable(GL_LIGHTING);

  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);  // draw polygon outlines
  glLineWidth(1);
  vcg::glColor(DARKGRAY);
  for (size_t i = 0; i < _renderData->tMesh.face.size(); i++) {
    if (_renderData->tMesh.face[i].IsD())
      continue;

    glBegin(GL_POLYGON);
    for (int j = 0; j < _renderData->tMesh.face[i].VN(); j++)
      vcg::glVertex(_renderData->tMesh.face[i].cP(j));
    glEnd();
  }

  // draw boundary
  glDisable(GL_LIGHTING);
  glLineWidth(1);

  // draw lines
  glDepthRange(0.0f, 0.999999f);
  for (size_t i = 0; i < _renderData->boundary.sides.size(); ++i) {
    vcg::glColor(_renderData->boundary.colors[i]);
    glBegin(GL_LINE_STRIP);
    for (it = _renderData->boundary.sides[i].begin(); it != _renderData->boundary.sides[i].end(); ++it)
      vcg::glVertex(*it);
    glEnd();
  }
  // draw points
  if (_renderData->showVertices == ShowVertices::ALL || _renderData->showVertices == ShowVertices::SINGS) {
    for (size_t i = 0; i < _renderData->boundary.sides.size(); ++i) {
      vcg::glColor(CYAN);
      it = _renderData->boundary.sides[i].begin();
      vcg::Add_Ons::glPoint<vcg::Add_Ons::DMSolid>(vcg::Point3f::Construct(*it), _renderData->vertexRadius, 32, 32);
      if (_renderData->showVertices == ShowVertices::ALL) {
        vcg::glColor(DARKGRAY);
        for (++it, nextIt = it, ++nextIt;
             nextIt != _renderData->boundary.sides[i].end();
             it = nextIt, ++nextIt)
          vcg::Add_Ons::glPoint<vcg::Add_Ons::DMSolid>(vcg::Point3f::Construct(*it), _renderData->vertexRadius, 32, 32);
      }
    }
  }

  // draw drawnlines
  glDepthRange(0.0f, 0.999998f);
  if (!_renderData->flows.empty()) {
    for (flowMapIt = _renderData->flows.begin(); flowMapIt != _renderData->flows.end(); ++flowMapIt) {
      flowColorsMapIt = _renderData->flowsColors.find(flowMapIt->first);
      assert(flowColorsMapIt != _renderData->flowsColors.end());
      const FlowContainer &flowContainer = flowMapIt->second;
      const std::deque<vcg::Color4b> &flowColors = flowColorsMapIt->second;
      assert(flowContainer.size() == flowColors.size());
      for (size_t i = 0; i < flowContainer.size() && i < flowColors.size(); ++i) {
        vcg::glColor(flowColors[i]);
        // draw flows
        const FlowPoints &flowPoints = flowContainer[i];
        glLineWidth(5);
        glBegin(GL_LINE_STRIP);
        for (size_t p = 0; p < flowContainer[i].points.size(); ++p)
          vcg::glVertex(flowPoints.points[p]);
        glEnd();
      }
    }
  }

  glDepthRange(0.0f, 1.0f);
  glPopAttrib();
  glPopMatrix();
}

/**
 * @brief GLArea::renderBoundary
 */
void GLArea::renderBoundary() {
  if (_renderData->boundary.sides.empty())
    return;

  vcg::Box3<ScalarType> bbox;
  PointConstIterator it, nextIt;
  FlowMapConstIterator flowMapIt;
  FlowContainerConstIterator flowsIt;
  RenderData::FlowColorsMap::const_iterator flowsColorsMapIt;
  RenderData::FlowColors::const_iterator flowsColorsIt;
  vcg::Color4b c;

  for (size_t i = 0; i < _renderData->boundary.sides.size(); ++i)
    for (it = _renderData->boundary.sides[i].begin(); it != _renderData->boundary.sides[i].end(); ++it)
      bbox.Add(*it);
  ScalarType d = 2.5f / bbox.Diag();
  glPushAttrib(GL_ALL_ATTRIB_BITS);
  glPushMatrix();
  vcg::glScale(d);
  glTranslate(-bbox.Center());

  glDisable(GL_LIGHTING);
  glLineWidth(1);

  // draw lines
  for (size_t i = 0; i < _renderData->boundary.sides.size(); ++i) {
    vcg::glColor(_renderData->boundary.colors[i]);
    glBegin(GL_LINE_STRIP);
    for (it = _renderData->boundary.sides[i].begin(); it != _renderData->boundary.sides[i].end(); ++it)
      vcg::glVertex(*it);
    glEnd();
  }
  // draw points
  if (_renderData->showVertices == ShowVertices::ALL || _renderData->showVertices == ShowVertices::SINGS) {
    for (size_t i = 0; i < _renderData->boundary.sides.size(); ++i) {
      vcg::glColor(CYAN);
      it = _renderData->boundary.sides[i].begin();
      vcg::Add_Ons::glPoint<vcg::Add_Ons::DMSolid>(vcg::Point3f::Construct(*it), _renderData->vertexRadius, 32, 32);
      if (_renderData->showVertices == ShowVertices::ALL) {
        vcg::glColor(DARKGRAY);
        for (++it, nextIt = it, ++nextIt;
             nextIt != _renderData->boundary.sides[i].end();
             it = nextIt, ++nextIt)
          vcg::Add_Ons::glPoint<vcg::Add_Ons::DMSolid>(vcg::Point3f::Construct(*it), _renderData->vertexRadius, 32, 32);
      }
    }
  }

  // picking
  picking();

  // draw drawnlines
  glLineWidth(5);
  c = vcg::Color4b::Scatter(_renderData->boundary.sides.size() + 20, _renderData->boundary.sides.size() + _renderData->nFlows);
  vcg::glColor(c);
  glBegin(GL_LINE_STRIP);
  for (size_t p = 0; p < _lastFlowPoints.points.size(); ++p)
    vcg::glVertex(_lastFlowPoints.points[p]);
  glEnd();
  if (!_renderData->flows.empty()) {
    for (flowMapIt = _renderData->flows.begin(); flowMapIt != _renderData->flows.end(); ++flowMapIt) {
      flowsColorsMapIt = _renderData->flowsColors.find(flowMapIt->first);
      assert(flowsColorsMapIt != _renderData->flowsColors.end());
      for (flowsIt = flowMapIt->second.begin(), flowsColorsIt = flowsColorsMapIt->second.begin();
           flowsIt != flowMapIt->second.end() && flowsColorsIt != flowsColorsMapIt->second.end();
           ++flowsIt, ++flowsColorsIt) {
        vcg::glColor(*flowsColorsIt);
        glBegin(GL_LINE_STRIP);
        for (size_t p = 0; p < flowsIt->points.size(); ++p)
          // draw the lines
          vcg::glVertex(flowsIt->points[p]);
        glEnd();
      }
    }
  }

  glPopAttrib();
  glPopMatrix();
}

/**
 * @brief GLArea::renderData
 * @param painter
 * @param mesh
 * @param strings
 * @param polyLines
 * @param polyVectors
 * @param value
 * @param showVertices
 * @param showStrings
 */
void GLArea::renderData(QPainter *painter, const MeshData &meshData, const ScalarType scaleFactor, const PointType &center,
                        const ShowVertices showVertices, const bool showStrings, const bool showDrawnPolyLines) {
  std::list<PolyLine>::const_iterator polyLinesIt;
  std::list<PointType>::const_iterator pointsIt;
  std::list<Text>::const_iterator stringsIt;
  ScalarType minW, maxW;
  vcg::Color4b c;
  FlowMapConstIterator flowMapIt;
  RenderData::FlowColorsMap::const_iterator flowColorsMapIt;
  PolychordMapConstIterator polychordMapIt;

  if (meshData.mesh.IsEmpty())
    return;

  glPushAttrib(GL_ALL_ATTRIB_BITS);
  glPushMatrix();
  vcg::glScale(scaleFactor);
  glTranslate(-center);

  // put here mesh painting code
  glEnable(GL_LIGHTING);
  glPolygonOffset(1.0f, 1.0f);          // set offset
  glEnable(GL_POLYGON_OFFSET_FILL);     // turn on offset
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  for (size_t i = 0; i < meshData.mesh.face.size(); ++i) {
    if (meshData.mesh.face[i].IsD())
      continue;

    vcg::glColor(meshData.mesh.face[i].cC());
    PointType n = meshData.mesh.face[i].cN();
    glBegin(GL_POLYGON);
    for (int j = 0; j < meshData.mesh.face[i].VN(); ++j) {
      vcg::glNormal(n);
      vcg::glVertex(meshData.mesh.face[i].cP(j));
    }
    glEnd();
  }
  glDisable(GL_POLYGON_OFFSET_FILL);    // turn off offset

  glDisable(GL_LIGHTING);

  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);  // draw polygon outlines
  glLineWidth(1);
  vcg::glColor(DARKGRAY);
  for (size_t i = 0; i < meshData.mesh.face.size(); i++) {
    if (meshData.mesh.face[i].IsD())
      continue;

    glBegin(GL_POLYGON);
    for (int j = 0; j < meshData.mesh.face[i].VN(); j++)
      vcg::glVertex(meshData.mesh.face[i].cP(j));
    glEnd();
  }

  glDepthRange(0.0f, 0.999999f);

  // draw other lines
  if (!meshData.polyLines.empty()) {
    glLineWidth(3);
    for (polyLinesIt = meshData.polyLines.begin(); polyLinesIt != meshData.polyLines.end(); ++polyLinesIt) {
      // set the color
      vcg::glColor(polyLinesIt->color);
      // draw the lines
      glBegin(GL_LINE_STRIP);
      for (pointsIt = polyLinesIt->points.begin(); pointsIt != polyLinesIt->points.end(); ++pointsIt) {
        vcg::glVertex(*pointsIt);
      }
      glEnd();
    }
  }

  // draw vertices
  if (showVertices == ShowVertices::ALL || showVertices == ShowVertices::SINGS) {
    for (size_t i = 0; i < meshData.mesh.vert.size(); ++i) {
      if (meshData.mesh.vert[i].IsD())
        continue;
      if (showVertices == ShowVertices::SINGS && !meshData.mesh.vert[i].IsS())
        continue;
      vcg::glColor(meshData.mesh.vert[i].cC());
      vcg::Add_Ons::glPoint<vcg::Add_Ons::DMSolid>(vcg::Point3f::Construct(meshData.mesh.vert[i].cP()), meshData.vertexRadius, 32, 32);
    }
  }

  // picking
  picking();

  // draw drawnlines
  glDepthRange(0.0f, 0.999998f);
  glLineWidth(5);
  c = vcg::Color4b::Scatter(_renderData->boundary.sides.size() + 20, _renderData->boundary.sides.size() + _renderData->nFlows);
  vcg::glColor(c);
  glBegin(GL_LINE_STRIP);
  for (size_t p = 0; p < _lastFlowPoints.points.size(); ++p)
    vcg::glVertex(_lastFlowPoints.points[p]);
  glEnd();
  if (showDrawnPolyLines && !_renderData->flows.empty()) {
    if (_renderData->showSampleVectors) {
      // crossed lines
      minW = std::numeric_limits<ScalarType>::max();
      maxW = 0.0;
      for (polychordMapIt = meshData.polychordMap.begin(); polychordMapIt != meshData.polychordMap.end(); ++polychordMapIt) {
        const PolychordContainer &polychordContainer = polychordMapIt->second;
        for (size_t i = 0; i < polychordContainer.size(); ++i) {
          const PolychordPoints &polychordPoints = polychordContainer[i];
          for (size_t p = 0; p < polychordPoints.distances.size(); ++p) {
            if (polychordPoints.distances[p] < minW)
              minW = polychordPoints.distances[p];
            if (polychordPoints.distances[p] > maxW)
              maxW = polychordPoints.distances[p];
          }
        }
      }
    }
    for (flowMapIt = _renderData->flows.begin(); flowMapIt != _renderData->flows.end(); ++flowMapIt) {
      flowColorsMapIt = _renderData->flowsColors.find(flowMapIt->first);
      assert(flowColorsMapIt != _renderData->flowsColors.end());
      const FlowContainer &flowContainer = flowMapIt->second;
      const std::deque<vcg::Color4b> &flowColors = flowColorsMapIt->second;
      assert(flowContainer.size() == flowColors.size());
      polychordMapIt = meshData.polychordMap.find(flowMapIt->first);
      for (size_t i = 0; i < flowContainer.size() && i < flowColors.size(); ++i) {
        vcg::glColor(flowColors[i]);
        // draw flows
        const FlowPoints &flowPoints = flowContainer[i];
        glLineWidth(5);
        glBegin(GL_LINE_STRIP);
        for (size_t p = 0; p < flowContainer[i].points.size(); ++p)
          vcg::glVertex(flowPoints.points[p]);
        glEnd();
        // draw polychords
        if (_renderData->showPolychords && polychordMapIt != meshData.polychordMap.end() && i < polychordMapIt->second.size()) {
          const PolychordPoints &polychordPoints = polychordMapIt->second[i];
          glLineWidth(2);
          glBegin(GL_LINE_STRIP);
          for (size_t p = 0; p < polychordPoints.points.size(); ++p)
            vcg::glVertex(polychordPoints.points[p]);
          glEnd();
        }
        // draw sample vectors
        if (_renderData->showSampleVectors && polychordMapIt != meshData.polychordMap.end() && i < polychordMapIt->second.size()) {
          const PolychordPoints &polychordPoints = polychordMapIt->second[i];
          glLineWidth(2);
          glBegin(GL_LINES);
          for (size_t s = 0; s < polychordPoints.samplePoints.size() && s < flowPoints.samplePoints.size(); ++s) {
            Coloring::SetColorScaled(c, minW, maxW, polychordPoints.distances[s], (minW + maxW) * 0.5);
            vcg::glColor(c);
            vcg::glVertex(flowPoints.samplePoints[s]);
            vcg::glVertex(polychordPoints.samplePoints[s]);
          }
          glEnd();
        }
      }
    }
  }
  glDepthRange(0.0f, 1.0f);

  glPopAttrib();

  // draw strings
  if (showStrings && !meshData.strings.empty())
    for (stringsIt = meshData.strings.begin(); stringsIt != meshData.strings.end(); ++stringsIt)
      drawRichText(painter, stringsIt->point, QString(stringsIt->text.c_str()), stringsIt->color);

  glPopMatrix();
}
