#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <QActionGroup>
#include <QString>
#include <QDir>
#include <QFileDialog>
#include <QInputDialog>
#include <QMessageBox>
#include <QElapsedTimer>
#include <QTime>
#include <sstream>
#include "mesh_utilities.h"

MainWindow::MainWindow(QWidget *parent): QMainWindow(parent)
                                       , ui(new Ui::MainWindow)
                                       , patchRenderer(0)
                                       , viewMode(DEFAULT)
                                       , pDBServer(new PatchDBServer<PolyMesh>)
                                       , patchGeometrizator(new PatchFlattener<PolyMesh>)
                                       , patchRetopologizer(new PatchRetopologizer<PolyMesh,TMesh>)
                                       , keepNodesInMemory(false)
                                       , maxNFaces(std::numeric_limits<num_type>::max())
                                       , side(0)
#ifndef SINGLE_THREAD
                                       , nThreads(std::thread::hardware_concurrency())
#endif
{
    ui->setupUi(this);
    // shortcuts
    ui->actionOpen->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_O));
    ui->actionSave->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_S));
    ui->actionCollapse->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_L));
    ui->actionEnumerate_Polygons->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_E));
    ui->actionNavigate->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_N));
    ui->actionFind_Patches->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_F));
    ui->actionSplit_polychord->setShortcut(QKeySequence(Qt::CTRL | Qt::SHIFT | Qt::Key_Y));
#if defined(__APPLE__) || defined(MACOSX)
    ui->actionFullscreen->setShortcut(QKeySequence(Qt::CTRL | Qt::META | Qt::Key_F));
#else
    ui->actionFullscreen->setShortcut(QKeySequence(Qt::Key_F11));
#endif
    ui->actionTrackball->setShortcut(QKeySequence(Qt::CTRL | Qt::SHIFT | Qt::Key_T));
    ui->actionLabels->setShortcut(QKeySequence(Qt::CTRL | Qt::SHIFT | Qt::Key_L));
    ui->actionSnapshot->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_P));
    ui->actionBase_Patch->setShortcut(QKeySequence(Qt::CTRL | Qt::SHIFT | Qt::Key_B));
    ui->actionNo_vertices->setShortcut(QKeySequence(Qt::CTRL | Qt::ALT | Qt::SHIFT | Qt::Key_V));
    ui->actionAll_vertices->setShortcut(QKeySequence(Qt::CTRL | Qt::ALT | Qt::Key_V));
    ui->actionSingularities_only->setShortcut(QKeySequence(Qt::CTRL | Qt::SHIFT | Qt::Key_V));
    ui->actionKeep_Polygons->setShortcut(QKeySequence(Qt::CTRL | Qt::SHIFT | Qt::Key_K));
    ui->actionPreview->setShortcut(QKeySequence(Qt::CTRL | Qt::SHIFT | Qt::Key_P));
    ui->actionAligned_Vectors->setShortcut(QKeySequence(Qt::CTRL | Qt::ALT | Qt::SHIFT | Qt::Key_S));
    ui->actionAligned_Polychords->setShortcut(QKeySequence(Qt::CTRL | Qt::ALT | Qt::SHIFT | Qt::Key_P));
    ui->actionRetopo_Patches->setShortcut(QKeySequence(Qt::CTRL | Qt::Key_R));
    // end shortcut

    meshToLearnData.meshes.resize(1);
    meshToLearnData.pickRadius = 0;
    meshToLearnData.vertexRadius = 0;
    patchesData.vertexRadius = PatchFlattener<PolyMesh>::RADIUS / 50.0;
    patchesData.pickRadius = PatchFlattener<PolyMesh>::RADIUS / 20.0;
    retopoPatchesData.pickRadius = 0;
    retopoPatchesData.vertexRadius = 0;

    dbFilename = QDir::homePath().toStdString() + std::string("/") + PatchDBInterface::PATCHDBFILENAME();
    polygonDBFilename = QDir::homePath().toStdString() + std::string("/") + PolygonDBInterface::POLYGONDBFILENAME();

    pDBServer->setFilename(dbFilename);
    polygonRegister.setBufferSize(300000000);
    patchLearner.setPolygonDBFilename(polygonDBFilename);
    patchLearner.setPatchDBFilename(dbFilename);
    patchLearner.setInsertBufferSize(3000000);
    patchLearner.setUpdateBufferSize(300000000);
    patchLookup.setPatchDBServer(pDBServer);
    patchLookup.showStatusMessages(true);
    patchLookupRetopo.setPatchDBServer(pDBServer);
    patchLookupRetopo.showStatusMessages(true);
#ifndef SINGLE_THREAD
    polygonEnumerator.setNumberOfThreads(nThreads);
    patchLookup.setNumberOfThreads(nThreads);
    patchLookupRetopo.setNumberOfThreads(nThreads);
#endif

    connect(ui->actionTrackball, SIGNAL(toggled(bool)), ui->centralWidget, SLOT(setTrackballVisible(bool)));
    QActionGroup *verticesGroup = new QActionGroup(this);
    verticesGroup->addAction(ui->actionNo_vertices);
    verticesGroup->addAction(ui->actionAll_vertices);
    verticesGroup->addAction(ui->actionSingularities_only);
    connect(verticesGroup, SIGNAL(triggered(QAction*)), this, SLOT(on_actionShowVertices(QAction*)));
    if (ui->actionAll_vertices->isChecked())
      meshToLearnData.showVertices = ShowVertices::ALL;
    else if (ui->actionSingularities_only->isChecked())
      meshToLearnData.showVertices = ShowVertices::SINGS;
    else
      meshToLearnData.showVertices = ShowVertices::NONE;
    patchesData.showVertices = meshToLearnData.showVertices;
    retopoPatchesData.showVertices = meshToLearnData.showVertices;
    connect(ui->actionLabels, SIGNAL(toggled(bool)), this, SLOT(on_actionLabels_toggled(bool)));
    meshToLearnData.showStrings = false;
    patchesData.showStrings = ui->actionLabels->isChecked();
    retopoPatchesData.showStrings = ui->actionLabels->isChecked();
    meshToLearnData.showPreview = false;
    patchesData.showPreview = ui->actionPreview->isChecked();
    retopoPatchesData.showPreview = ui->actionPreview->isChecked();
    meshToLearnData.showBaseMesh = false;
    patchesData.showBaseMesh = ui->actionBase_Patch->isChecked();
    retopoPatchesData.showBaseMesh = ui->actionBase_Patch->isChecked();
    meshToLearnData.showSampleVectors = false;
    patchesData.showSampleVectors = ui->actionAligned_Vectors->isChecked();
    retopoPatchesData.showSampleVectors = ui->actionAligned_Vectors->isChecked();
    meshToLearnData.showPolychords = false;
    patchesData.showPolychords = ui->actionAligned_Polychords->isChecked();
    retopoPatchesData.showPolychords = ui->actionAligned_Polychords->isChecked();
    meshToLearnData.showTriMesh = false;
    patchesData.showTriMesh = false;
    retopoPatchesData.showTriMesh = true;

    patchRenderer = std::make_shared<PatchRenderer>(&patchesData, pDBServer, &patchLookup);
#ifndef SINGLE_THREAD
    patchRenderer->registerForNotifications();
#endif
    patchRenderer->setWindowFlags(Qt::Tool | Qt::WindowStaysOnTopHint);
    patchRenderer->setWindowTitle(tr("Find Patches"));
    patchRenderer->hide();
    connect(&*patchRenderer, SIGNAL(toStatus(QString)), ui->statusBar, SLOT(showMessage(QString)));
    connect(&*patchRenderer, SIGNAL(changed()), ui->centralWidget, SLOT(update()));
    connect(&*patchRenderer, SIGNAL(drawLines(bool)), this, SLOT(setInputMode(bool)));
    connect(&*patchRenderer, SIGNAL(closed()), this, SLOT(closeFindPatchesTool()));
    connect(&*patchRenderer, SIGNAL(executing(bool)), ui->actionOpen, SLOT(setDisabled(bool)));
    connect(&*patchRenderer, SIGNAL(activateTrackball(bool)), this, SLOT(activateTrackBall(bool)));
    connect(this, SIGNAL(toStatusBarText(QString)), ui->statusBar, SLOT(showMessage(QString)));

    patchRetopo = std::make_shared<PatchRetopo>(&retopoPatchesData, &patchLookupRetopo);
#ifndef SINGLE_THREAD
    patchRetopo->registerForNotifications();
#endif
    patchRetopo->setWindowFlags(Qt::Tool | Qt::WindowStaysOnTopHint);
    patchRetopo->setWindowTitle(tr("Retopo Patches"));
    patchRetopo->hide();
    connect(&*patchRetopo, SIGNAL(toStatus(QString)), ui->statusBar, SLOT(showMessage(QString)));
    connect(&*patchRetopo, SIGNAL(changed()), ui->centralWidget, SLOT(update()));
    connect(&*patchRetopo, SIGNAL(closed()), this, SLOT(closeFindPatchesTool()));
    connect(&*patchRetopo, SIGNAL(executing(bool)), ui->actionOpen, SLOT(setDisabled(bool)));
    connect(&*patchRetopo, SIGNAL(executing(bool)), ui->menuVertices, SLOT(setDisabled(bool)));
    connect(&*patchRetopo, SIGNAL(executing(bool)), ui->actionLabels, SLOT(setDisabled(bool)));
    connect(this, SIGNAL(toStatusBarText(QString)), ui->statusBar, SLOT(showMessage(QString)));
}

/**
 * @brief MainWindow::~MainWindow
 */
MainWindow::~MainWindow() {
#ifndef SINGLE_THREAD
  // register callback functions for multi-threaded enumeration
  std::shared_ptr<MT_PolygonEnumerationNotifiable> shared_this =
      std::dynamic_pointer_cast<MT_PolygonEnumerationNotifiable>(shared_from_this());
  polygonEnumerator.deregisterNotifiable(shared_this);
#endif
  delete ui;
}

#ifndef SINGLE_THREAD
/**
 * @brief MainWindow::registerForNotifications
 */
void MainWindow::registerForNotifications() {
  // register callback functions for multi-threaded enumeration
  std::shared_ptr<MT_PolygonEnumerationNotifiable> shared_this =
      std::dynamic_pointer_cast<MT_PolygonEnumerationNotifiable>(shared_from_this());
  polygonEnumerator.registerNotifiable(shared_this);
}

/**
 * @brief MainWindow::notify receive a notification on the ongoing enumeration process.
 * @param notificationType The type of event.
 * @param nTrees The number of trees currently built.
 * @param nPolygons The number of polygons currently enumerated.
 * @param duration The time elapsed since the beginning of the process.
 */
void MainWindow::notify(NotificationType notificationType,
                        num_type nTrees, count_type nPolygons,
                        std::chrono::nanoseconds duration) {

  QString msg = "Faces " + QString::number(meshToLearnData.meshes.front().mesh.FN()) + ". ";
  msg += QString::number(nTrees) + " trees completed. ";
  msg += QString::number(nPolygons) + " polygons enumerated. ";
  msg += "Elapsed: ";
  std::string durationStr;
  TimeElements timeElements;
  TimeSupport::SplitTimeElements(duration, timeElements);
  TimeSupport::TimeElementsToString(timeElements, durationStr);
  msg += durationStr.c_str();

  switch (notificationType) {
  case TREE_BUILT:
    msg += " Still enumerating";
    break;
  case FLUSH_DB:
    msg += " Enumeration complete. Flushing";
    break;
  case COMPLETE:
    msg += " Enumeration complete! ";
    msg += QString::number(polygonEnumerator.maxNumberOfSides()) + " max sides. ";
    msg += QString::number(polygonEnumerator.maxNumberOfSingularities()) + " max singularities. ";
    msg += QString::number(patchLearner.getNumberOfPatches()) + " patches (" + QString::number(patchLearner.getNumberOfNewPatches()) + " new).";
    // update GUI
    viewMode = DEFAULT;
    ui->actionCollapse->setEnabled(true);
    ui->actionKeep_Polygons->setEnabled(true);
    ui->actionEnumerate_Polygons->setEnabled(true);
    ui->actionNavigate->setChecked(false);
    ui->actionNavigate->setEnabled(true);
    ui->actionFind_Patches->setEnabled(true);
    ui->actionOpen->setEnabled(true);
    ui->actionSave->setEnabled(true);
    break;
  default:
    break;
  }

  if (notificationType != EMPTY && notificationType != COMPLETE)
    for (int i = 0; i <= timeElements.s.count() % 3; ++i)
      msg += '.';
  emit toStatusBar(msg);
}
#endif

/**
 * @brief MainWindow::on_actionOpen_triggered
 */
void MainWindow::on_actionOpen_triggered() {
  // select a mesh file
  filename = QFileDialog::getOpenFileName(0, tr("Open mesh"), QString(),
                                                  tr("Mesh (*.ply *.off *.obj *.vmi)"));

  if(!filename.isEmpty()) {
    // open the mesh
    int err = MeshSupport::OpenMesh(meshToLearnData.meshes.front().mesh, filename.toStdString().c_str());

    if (err == 0) {
      if (patchRenderer) {
        closeFindPatchesTool();
        patchRenderer->setVisible(false);
        patchRetopo->setVisible(false);
      }
      viewMode = DEFAULT;
      vcg::tri::UpdateColor<PolyMesh>::PerFaceConstant(meshToLearnData.meshes.front().mesh, LIGHTGRAY);
      Coloring::SetVertexColorByValence(meshToLearnData.meshes.front().mesh);
      meshToLearnData.vertexRadius = Coloring::ComputeVertexRadius(meshToLearnData.meshes.front().mesh);
      meshToLearnData.meshes.front().vertexRadius = meshToLearnData.vertexRadius;
      polygonEnumerator.resize(meshToLearnData.meshes.front().mesh.face.size());
      currentPos.SetNull();
      ui->centralWidget->setRenderData(&meshToLearnData);
      ui->actionCollapse->setEnabled(true);
      ui->actionKeep_Polygons->setEnabled(true);
      ui->actionEnumerate_Polygons->setEnabled(true);
      ui->actionNavigate->setEnabled(true);
      ui->actionNavigate->setChecked(false);
      ui->statusBar->showMessage(tr("Faces: ") + QString::number(meshToLearnData.meshes.front().mesh.FN()) + ".");
    } else {
      viewMode = DEFAULT;
      ui->centralWidget->setRenderData(0);
      polygonEnumerator.resize(0);
      currentPos.SetNull();
      ui->actionCollapse->setEnabled(false);
      ui->actionKeep_Polygons->setEnabled(false);
      ui->actionEnumerate_Polygons->setEnabled(false);
      ui->actionNavigate->setEnabled(false);
      ui->actionNavigate->setChecked(false);
      ui->actionSplit_polychord->setEnabled(false);
      QMessageBox::critical(0, tr("Error"), tr("Unable to open file"));
    }
  }
  ui->centralWidget->clearButtons();
}

/**
 * @brief MainWindow::on_actionCollapse_triggered
 */
void MainWindow::on_actionCollapse_triggered() {
  if (!meshToLearnData.meshes.front().mesh.IsEmpty()) {
    // collapse all
    vcg::tri::PolychordCollapse<PolyMesh>::CollapseAllPolychords(meshToLearnData.meshes.front().mesh);
    // delete faces
    vcg::tri::Allocator<PolyMesh>::CompactFaceVector(meshToLearnData.meshes.front().mesh);
    // delete vertices
    vcg::tri::Allocator<PolyMesh>::CompactVertexVector(meshToLearnData.meshes.front().mesh);
    // update normals
    vcg::tri::UpdateNormal<PolyMesh>::PerPolygonalFaceNormalized(meshToLearnData.meshes.front().mesh);
    // update the bounding box
    vcg::tri::UpdateBounding<PolyMesh>::Box(meshToLearnData.meshes.front().mesh);
    // update faces and vertices color
    vcg::tri::UpdateColor<PolyMesh>::PerFaceConstant(meshToLearnData.meshes.front().mesh, LIGHTGRAY);
    Coloring::SetVertexColorByValence(meshToLearnData.meshes.front().mesh);

    ui->actionKeep_Polygons->setEnabled(true);
    ui->actionNavigate->setEnabled(true);
    ui->actionNavigate->setChecked(false);
    viewMode = DEFAULT;
    polygonEnumerator.resize(meshToLearnData.meshes.front().mesh.face.size());
    currentPos.SetNull();
    ui->statusBar->showMessage(tr("Faces: ") + QString::number(meshToLearnData.meshes.front().mesh.FN()) + ".");
  }
  ui->centralWidget->clearButtons();
}

/**
 * @brief MainWindow::on_actionSave_triggered
 */
void MainWindow::on_actionSave_triggered(){
  if (!meshToLearnData.meshes.front().mesh.IsEmpty() || !patchesData.meshes[patchesData.currentMesh].mesh.IsEmpty()) {
    QString filename = QFileDialog::getSaveFileName(0, tr("Save mesh"), QString(),
                                                    tr("Mesh (*.obj *.ply *.off *.vmi)"));
    if(!filename.isEmpty()) {
      int err = 0;
      if (!patchesData.meshes[patchesData.currentMesh].mesh.IsEmpty())
        err = MeshSupport::SaveMesh(patchesData.meshes[patchesData.currentMesh].mesh, filename.toStdString().c_str());
      else
        err = MeshSupport::SaveMesh(meshToLearnData.meshes.front().mesh, filename.toStdString().c_str());

      if (err == 0)
        QMessageBox::information(0, tr("Saved"), tr("File saved"));
      else
        QMessageBox::critical(0, tr("Error"), tr("Unable to save file"));
    }
  }
  ui->centralWidget->clearButtons();
}

/**
 * @brief MainWindow::on_actionEnumerate_Polygons_triggered
 */
void MainWindow::on_actionEnumerate_Polygons_triggered() {
  // get the faces number bound
  bool ok;
  maxNFaces = QInputDialog::getInt(0, tr("Max number of faces into each polygon"),
                                   tr("Set the maximum number of faces each polygon is allowed to have"),
                                   std::numeric_limits<int>::max() - 1, 0, std::numeric_limits<int>::max() - 1,
                                   1, &ok);
  if (maxNFaces == (num_type)std::numeric_limits<int>::max())
    maxNFaces = std::numeric_limits<num_type>::max();
  if (!ok)
    return;

  CornersType cornersType;
  QMessageBox cornersTypeBox;
  cornersTypeBox.setText(tr("Corners Type"));
  cornersTypeBox.setInformativeText(tr("Which kind of corners are the patches allowed to have?"));
  QPushButton *convexB = cornersTypeBox.addButton(tr("Convex"), QMessageBox::ActionRole);
  QPushButton *concaveB = cornersTypeBox.addButton(tr("Concave"), QMessageBox::ActionRole);
  QPushButton *irregularB = cornersTypeBox.addButton(tr("Irregular"), QMessageBox::ActionRole);
  QPushButton *edgesB = cornersTypeBox.addButton(tr("Edges"), QMessageBox::ActionRole);
  cornersTypeBox.exec();
  if ((QPushButton *)cornersTypeBox.clickedButton() == convexB)
    cornersType = CornersType::CONVEX;
  else if ((QPushButton *)cornersTypeBox.clickedButton() == concaveB)
    cornersType = CornersType::CONCAVE;
  else if ((QPushButton *)cornersTypeBox.clickedButton() == irregularB)
    cornersType = CornersType::IRREGULAR;
  else if ((QPushButton *)cornersTypeBox.clickedButton() == edgesB)
    cornersType = CornersType::NOTCONVEX;
  else
    return;

  count_type counterSize = QInputDialog::getInt(0, tr("Max number of polygons bound to each template patch"),
                                                tr("Set the maximum number of polygons to bind to each template patch"),
                                                0, 0, std::numeric_limits<int>::max() - 1,
                                                1, &ok);
  if (counterSize == (count_type)std::numeric_limits<int>::max())
    counterSize = std::numeric_limits<count_type>::max();
  if (!ok)
    return;

  // update GUI
  viewMode = DEFAULT;
  ui->actionCollapse->setDisabled(true);
  ui->actionKeep_Polygons->setDisabled(true);
  ui->actionEnumerate_Polygons->setDisabled(true);
  ui->actionNavigate->setChecked(false);
  ui->actionNavigate->setDisabled(true);
  ui->actionOpen->setDisabled(true);
  ui->actionSave->setDisabled(true);
  ui->actionFind_Patches->setDisabled(true);
  ui->statusBar->showMessage(tr("Enumerating..."));

  // initialize mark
  vcg::tri::IMark(meshToLearnData.meshes.front().mesh) = 0;
  vcg::tri::InitFaceIMark(meshToLearnData.meshes.front().mesh);
  vcg::tri::InitVertexIMark(meshToLearnData.meshes.front().mesh);
  vcg::tri::UpdateFlags<PolyMesh>::Clear(meshToLearnData.meshes.front().mesh);
  Coloring::SetVertexColorByValence(meshToLearnData.meshes.front().mesh);

  // prepare the DBs
  pDBServer->close();
  polygonRegister.initialize();
  polygonRegister.resetDB();
  if (counterSize > 0)
    patchLearner.setMeshFilename(filename.toStdString());
  else
    patchLearner.setMeshFilename(std::string());
  patchLearner.setPolygonCounterSize(counterSize);
  patchLearner.initialize();

  // enumerate
  polygonEnumerator.enumerate(meshToLearnData.meshes.front().mesh, &polygonRegister, patchLearner, 0, maxNFaces,
                              cornersType, (counterSize > 0), true, keepNodesInMemory);
}

/**
 * @brief MainWindow::keyPressEvent
 * @param e
 */
void MainWindow::keyPressEvent(QKeyEvent *e) {
  vcg::Color4b c;
  std::list<PolyMesh::CoordType> l;
  FaceSet faces;
  FaceSetIterator fIt;
  std::deque<PosType> corners;
  PosType startPos, endPos;
  n_corners_type nSides = 0;

  if (viewMode & NAVIGATE_MESH) {
    switch (e->key()) {
    case Qt::Key_Left:
      if (viewMode & NAVIGATE_TREE) {
        nSides = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].numberOfSides();
        side = (side + nSides - 1) % nSides;
        c = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].isChildNull(side) ? RED : GREEN;
        polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].corners(corners);
        startPos = corners[side];
        endPos = corners[(side + 1) % nSides];
      } else {
        endPos = currentPos;
        currentPos.FlipE();
        currentPos.FlipV();
        side = currentPos.E();
        c = currentPos.IsBorder() || !currentPos.IsManifold() ? RED : GREEN;
        startPos = currentPos;
      }
      break;

    case Qt::Key_Right:
      if (viewMode & NAVIGATE_TREE) {
        nSides = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].numberOfSides();
        side = (side + 1) % nSides;
        c = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].isChildNull(side) ? RED : GREEN;
        polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].corners(corners);
        startPos = corners[side];
        endPos = corners[(side + 1) % nSides];
      } else {
        currentPos.FlipV();
        currentPos.FlipE();
        side = currentPos.E();
        c = currentPos.IsBorder() || !currentPos.IsManifold() ? RED : GREEN;
        endPos = startPos = currentPos;
        endPos.FlipV();
        endPos.FlipE();
      }
      break;

    case Qt::Key_Up:
      if (viewMode & NAVIGATE_TREE) {
        if (!polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].isParentNull()) {
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].faces(faces);
          fIt = faces.begin();
          for (; fIt != faces.end(); fIt++)
            meshToLearnData.meshes.front().mesh.face[*fIt].C() = LIGHTGRAY;
          side = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].parentSide();
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].selectParent();
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].faces(faces);
          fIt = faces.begin();
          for (; fIt != faces.end(); fIt++)
            meshToLearnData.meshes.front().mesh.face[*fIt].C() = BLUE;
          c = GREEN;
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].corners(corners);
          startPos = corners[side];
          nSides = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].numberOfSides();
          endPos = corners[(side + 1) % nSides];
        } else {
          startPos.SetNull();
          endPos.SetNull();
        }
      } else {
        startPos.SetNull();
        endPos.SetNull();
      }
      break;

    case Qt::Key_Down:
      if (viewMode & NAVIGATE_TREE) {
        if (!polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].isChildNull(side)) {
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].faces(faces);
          fIt = faces.begin();
          for (; fIt != faces.end(); fIt++)
            meshToLearnData.meshes.front().mesh.face[*fIt].C() = LIGHTGRAY;
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].selectChild(side);
          side = 0;
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].faces(faces);
          fIt = faces.begin();
          for (; fIt != faces.end(); fIt++)
            meshToLearnData.meshes.front().mesh.face[*fIt].C() = BLUE;
          c = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].isChildNull(side) ? RED : GREEN;
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].corners(corners);
          startPos = corners[side];
          nSides = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].numberOfSides();
          endPos = corners[(side + 1) % nSides];
        } else {
          startPos.SetNull();
          endPos.SetNull();
        }
      } else {
        if (!currentPos.IsBorder() && currentPos.IsManifold()) {
          currentPos.F()->C() = LIGHTGRAY;
          currentPos.FlipF();
          currentPos.FlipV();
          currentPos.F()->C() = LIGHTBLUE;
          side = currentPos.E();
          c = GREEN;
          endPos = startPos = currentPos;
          endPos.FlipV();
          endPos.FlipE();
        } else {
          startPos.SetNull();
          endPos.SetNull();
        }
      }
      break;

    case Qt::Key_Return:
      if (viewMode == NAVIGATE_MESH) {
        if (!polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].isEmpty()) {
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].selectRoot();
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].faces(faces);
          fIt = faces.begin();
          for (; fIt != faces.end(); fIt++)
            meshToLearnData.meshes.front().mesh.face[*fIt].C() = BLUE;
          side = currentPos.E();
          c = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].isChildNull(side) ? RED : GREEN;
          polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].corners(corners);
          startPos = corners[side];
          nSides = polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].numberOfSides();
          endPos = corners[(side + 1) % nSides];
          viewMode |= NAVIGATE_TREE;
          ui->actionSplit_polychord->setEnabled(false);
        } else {
          startPos.SetNull();
          endPos.SetNull();
        }
      } else {
        startPos.SetNull();
        endPos.SetNull();
      }
      break;

    case Qt::Key_Escape:
      if (viewMode & NAVIGATE_TREE) {
        polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].faces(faces);
        fIt = faces.begin();
        for (; fIt != faces.end(); fIt++)
          meshToLearnData.meshes.front().mesh.face[*fIt].C() = LIGHTGRAY;
        side = currentPos.E();
        currentPos.F()->C() = LIGHTBLUE;
        c = GREEN;
        polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].corners(corners);
        startPos = corners[side];
        endPos = startPos = currentPos;
        endPos.FlipV();
        endPos.FlipE();
        viewMode &= ~NAVIGATE_TREE;
        ui->actionSplit_polychord->setEnabled(true);
      } else {
        startPos.SetNull();
        endPos.SetNull();
      }
      break;

    default:
      QMainWindow::keyPressEvent(e);
      return;
      break;
    }

    if (!startPos.IsNull()) {
      do {
        l.push_back(startPos.V()->P());
        startPos.FlipV();
        startPos.FlipE();
        if (startPos != endPos) {
          startPos.FlipF();
          startPos.FlipE();
        }
      } while (startPos != endPos);
      l.push_back(endPos.V()->P());
      // draw
      meshToLearnData.meshes.front().polyLines.clear();
      meshToLearnData.meshes.front().polyLines.push_back(PolyLine(c, l));
      ui->centralWidget->clearButtons();
    }

  } else {
    QMainWindow::keyPressEvent(e);
  }
}

/**
 * @brief MainWindow::on_actionNavigate_triggered
 */
void MainWindow::on_actionNavigate_triggered() {
  if (viewMode & NAVIGATE_MESH) {
    if (viewMode & NAVIGATE_TREE) {
      FaceSet faces;
      FaceSetIterator fIt;
      polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].faces(faces);
      fIt = faces.begin();
      for (; fIt != faces.end(); ++fIt)
        meshToLearnData.meshes.front().mesh.face[*fIt].C() = LIGHTGRAY;
      polygonEnumerator[vcg::tri::Index(meshToLearnData.meshes.front().mesh, currentPos.F())].selectRoot();
    }
    side = currentPos.E();
    currentPos.F()->C() = LIGHTGRAY;
    meshToLearnData.meshes.front().polyLines.clear();
    ui->centralWidget->clearButtons();
    viewMode = DEFAULT;
    ui->actionSplit_polychord->setEnabled(false);
    ui->actionNavigate->setChecked(false);
    ui->actionCollapse->setEnabled(true);
    ui->actionKeep_Polygons->setEnabled(true);
    ui->actionEnumerate_Polygons->setEnabled(true);
    ui->actionOpen->setEnabled(true);
    ui->actionFind_Patches->setEnabled(true);
    ui->actionRetopo_Patches->setEnabled(true);

  } else {
    if (currentPos.IsNull()) {
      size_t i = 0;
      while (meshToLearnData.meshes.front().mesh.face[i].IsD() && i < meshToLearnData.meshes.front().mesh.face.size())
        i++;
      if (i >= meshToLearnData.meshes.front().mesh.face.size()) {
        viewMode = DEFAULT;
        ui->actionNavigate->setChecked(false);
        return;
      }
      currentPos.Set(&meshToLearnData.meshes.front().mesh.face[i], 0, meshToLearnData.meshes.front().mesh.face[i].V(0));
    }
    currentPos.F()->C() = LIGHTBLUE;
    side = currentPos.E();
    std::list<PolyMesh::CoordType> l;
    l.push_back(currentPos.V()->P());
    l.push_back(currentPos.VFlip()->P());
    vcg::Color4b c = currentPos.IsBorder() || !currentPos.IsManifold() ? RED : GREEN;
    meshToLearnData.meshes.front().polyLines.push_back(PolyLine(c, l));
    ui->centralWidget->clearButtons();
    viewMode = NAVIGATE_MESH;
    ui->actionNavigate->setChecked(true);
    ui->actionSplit_polychord->setEnabled(true);
    ui->actionCollapse->setEnabled(false);
    ui->actionKeep_Polygons->setEnabled(false);
    ui->actionEnumerate_Polygons->setEnabled(false);
    ui->actionOpen->setEnabled(false);
    ui->actionFind_Patches->setEnabled(false);
    ui->actionRetopo_Patches->setEnabled(false);
  }
}

/**
 * @brief MainWindow::on_actionFind_Patches_triggered
 */
void MainWindow::on_actionFind_Patches_triggered() {
  if (!pDBServer->isInitialized()) {
    pDBServer->initialize();
    std::cout << "Loading DB...";
    std::cout.flush();
    pDBServer->load();
#ifdef WIN32
    _pclose(_popen("play C:\\Windows\\Media\\notify.wav", "r"));
#endif
    std::cout << "done. " << pDBServer->numberOfCachedPatches() << " templates loaded." << std::endl;
  }
  ui->statusBar->showMessage(tr(""));
  ui->actionCollapse->setDisabled(true);
  ui->actionKeep_Polygons->setDisabled(true);
  ui->actionEnumerate_Polygons->setDisabled(true);
  ui->actionNavigate->setChecked(false);
  ui->actionNavigate->setDisabled(true);
  ui->actionFind_Patches->setDisabled(true);
  ui->actionLabels->setEnabled(true);
  ui->actionRetopo_Patches->setDisabled(true);
  ui->actionBase_Patch->setEnabled(true);
  ui->actionPreview->setEnabled(true);
  ui->menuAlignment->setEnabled(true);
  viewMode = DEFAULT;
  patchRenderer->setPolygonDBFilename(polygonDBFilename);
  patchRenderer->move(x() + ui->centralWidget->x() + 10, y() + ui->centralWidget->y() + 25);
  patchRenderer->show();
  ui->centralWidget->setRenderData(&patchesData);
  ui->centralWidget->setTrackballActive(false);
}

/**
 * @brief MainWindow::on_actionRetopo_Patches_triggered
 */
void MainWindow::on_actionRetopo_Patches_triggered() {
  if (!pDBServer->isInitialized()) {
    pDBServer->initialize();
    std::cout << "Loading DB...";
    std::cout.flush();
    pDBServer->load();
#ifdef WIN32
    _pclose(_popen("play C:\\Windows\\Media\\notify.wav", "r"));
#endif
    std::cout << "done. " << pDBServer->numberOfCachedPatches() << " templates loaded." << std::endl;
  }
  ui->statusBar->showMessage(tr(""));
  ui->actionCollapse->setDisabled(true);
  ui->actionKeep_Polygons->setDisabled(true);
  ui->actionEnumerate_Polygons->setDisabled(true);
  ui->actionNavigate->setChecked(false);
  ui->actionNavigate->setDisabled(true);
  ui->actionFind_Patches->setDisabled(true);
  ui->actionLabels->setEnabled(true);
  ui->actionRetopo_Patches->setDisabled(true);
  ui->actionBase_Patch->setEnabled(true);
  ui->actionPreview->setEnabled(true);
  ui->menuAlignment->setEnabled(true);
  viewMode = DEFAULT;
  patchRetopo->move(x() + ui->centralWidget->x() + 10, y() + ui->centralWidget->y() + 25);
  patchRetopo->show();
  ui->centralWidget->setRenderData(&retopoPatchesData);
  ui->centralWidget->setTrackballActive(true);
}

/**
 * @brief MainWindow::closeFindPatchesTool
 */
void MainWindow::closeFindPatchesTool() {
  ui->centralWidget->setRenderData(0);
  if (!meshToLearnData.meshes.front().mesh.IsEmpty()) {
    ui->actionCollapse->setEnabled(true);
    ui->actionKeep_Polygons->setEnabled(true);
    ui->actionEnumerate_Polygons->setEnabled(true);
    ui->actionNavigate->setEnabled(true);
    ui->actionNavigate->setChecked(false);
    ui->centralWidget->setRenderData(&meshToLearnData);
  }
  ui->actionFind_Patches->setEnabled(true);
  ui->actionRetopo_Patches->setEnabled(true);
  ui->actionLabels->setDisabled(true);
  ui->actionPreview->setDisabled(true);
  ui->actionBase_Patch->setDisabled(true);
  ui->menuAlignment->setDisabled(true);
  viewMode = DEFAULT;
  ui->centralWidget->setTrackballActive(true);
  if (meshToLearnData.meshes.front().mesh.IsEmpty())
    ui->statusBar->showMessage(tr(""));
  else
    ui->statusBar->showMessage(tr("Faces: ") + QString::number(meshToLearnData.meshes.front().mesh.FN()) + ".");
}

/**
 * @brief MainWindow::on_actionSplit_polychord_triggered
 */
void MainWindow::on_actionSplit_polychord_triggered() {
    if (viewMode == NAVIGATE_MESH && !currentPos.IsNull()) {
      bool ok;
      size_t n = QInputDialog::getInt(0, tr("Number of polychords to substitute the current one"),
                                   tr("Set the number of polychords to substitute the current one"),
                                   2, 2, std::numeric_limits<int>::max(),
                                   1, &ok);
      if (!ok)
        return;

      polygonEnumerator.resize(0);
      vcg::tri::UpdateFlags<PolyMesh>::Clear(meshToLearnData.meshes.front().mesh);
      vcg::tri::PolychordCollapse<PolyMesh>::SplitPolychord(meshToLearnData.meshes.front().mesh, currentPos, n);
      polygonEnumerator.resize(meshToLearnData.meshes.front().mesh.face.size());
      // update normals
      vcg::tri::UpdateNormal<PolyMesh>::PerPolygonalFaceNormalized(meshToLearnData.meshes.front().mesh);
      // update the bounding box
      vcg::tri::UpdateBounding<PolyMesh>::Box(meshToLearnData.meshes.front().mesh);
      // set face and vertex color
      vcg::tri::UpdateColor<PolyMesh>::PerFaceConstant(meshToLearnData.meshes.front().mesh, LIGHTGRAY);
      currentPos.F()->C() = LIGHTBLUE;
      Coloring::SetVertexColorByValence(meshToLearnData.meshes.front().mesh);
      ui->statusBar->showMessage(tr("Faces: ") + QString::number(meshToLearnData.meshes.front().mesh.FN()) + ".");
      meshToLearnData.meshes.front().polyLines.clear();
      std::list<PolyMesh::CoordType> l;
      l.push_back(currentPos.V()->P());
      l.push_back(currentPos.VFlip()->P());
      vcg::Color4b c = currentPos.IsBorder() || !currentPos.IsManifold() ? RED : GREEN;
      meshToLearnData.meshes.front().polyLines.push_back(PolyLine(c, l));
      ui->centralWidget->clearButtons();
    }
}

/**
 * @brief MainWindow::on_actionFullscreen_triggered
 */
void MainWindow::on_actionFullscreen_triggered() {
  if (isFullScreen()) {
    showNormal();
    ui->actionFullscreen->setChecked(false);
  } else {
    showFullScreen();
    ui->actionFullscreen->setChecked(true);
  }
}

/**
 * @brief MainWindow::on_actionSnapshot_triggered
 */
void MainWindow::on_actionSnapshot_triggered() {
  if (!meshToLearnData.meshes.front().mesh.IsEmpty() ||
      (!patchesData.meshes.empty() && !patchesData.meshes[patchesData.currentMesh].mesh.IsEmpty()) ||
      !patchesData.boundary.sides.empty() ||
      !retopoPatchesData.boundary.sides.empty() ||
      !retopoPatchesData.tMesh.IsEmpty() ||
      (!retopoPatchesData.meshes.empty() && !retopoPatchesData.meshes[retopoPatchesData.currentMesh].mesh.IsEmpty())) {
    QString filename = QFileDialog::getSaveFileName(0, tr("Take a snapshot"),
                                                    QString(), tr("Picture (*.png)"));
    if (!filename.isEmpty())
      ui->centralWidget->takeSnapShot(filename);
  }
}

/**
 * @brief MainWindow::on_actionBase_Patch_toggled
 * @param visible
 */
void MainWindow::on_actionBase_Patch_toggled(bool visible) {
  patchesData.showBaseMesh = visible;
  retopoPatchesData.showBaseMesh = visible;
  ui->centralWidget->clearButtons();
}

/**
 * @brief MainWindow::toStatusBar
 * @param msg
 */
void MainWindow::toStatusBar(const QString &msg) {
  std::lock_guard<std::mutex> lock(mutex);
  emit toStatusBarText(msg);
}

/**
 * @brief MainWindow::on_actionKeep_Polygons_toggled
 * @param keep
 */
void MainWindow::on_actionKeep_Polygons_toggled(bool keep) {
  keepNodesInMemory = keep;
}

/**
 * @brief MainWindow::on_actionLabels_toggled
 * @param showLabels
 */
void MainWindow::on_actionLabels_toggled(bool showLabels) {
  meshToLearnData.showStrings = showLabels;
  patchesData.showStrings = showLabels;
  retopoPatchesData.showStrings = showLabels;
  ui->centralWidget->clearButtons();
}

/**
 * @brief MainWindow::setInputMode
 * @param draw
 */
void MainWindow::setInputMode(bool draw) {
  if (draw)
    ui->centralWidget->setInputMode(GLArea::InputMode::DRAW2DLINES);
  else
    ui->centralWidget->setInputMode(GLArea::InputMode::VIEW);
}

/**
 * @brief MainWindow::activateTrackBall
 * @param a
 */
void MainWindow::activateTrackBall(bool a) {
  ui->centralWidget->setTrackballActive(a);
  ui->centralWidget->resetTrackball();
}

/**
 * @brief MainWindow::on_actionShowVertices
 * @param action
 */
void MainWindow::on_actionShowVertices(QAction *action) {
  if (action == ui->actionNo_vertices)
    meshToLearnData.showVertices = ShowVertices::NONE;
  else if (action == ui->actionAll_vertices)
    meshToLearnData.showVertices = ShowVertices::ALL;
  else
    meshToLearnData.showVertices = ShowVertices::SINGS;
  patchesData.showVertices = meshToLearnData.showVertices;
  retopoPatchesData.showVertices = meshToLearnData.showVertices;
  ui->centralWidget->clearButtons();
}

/**
 * @brief MainWindow::on_actionPreview_toggled
 * @param arg1
 */
void MainWindow::on_actionPreview_toggled(bool showPreview) {
  patchesData.showPreview = showPreview;
  retopoPatchesData.showPreview = showPreview;
  ui->centralWidget->clearButtons();
}

/**
 * @brief MainWindow::on_actionAligned_Vectors_toggled
 * @param showAlignedVectors
 */
void MainWindow::on_actionAligned_Vectors_toggled(bool showAlignedVectors) {
  patchesData.showSampleVectors = showAlignedVectors;
  retopoPatchesData.showSampleVectors = showAlignedVectors;
  ui->centralWidget->clearButtons();
}

/**
 * @brief MainWindow::on_actionAligned_Polychords_toggled
 * @param showAlignedPolychords
 */
void MainWindow::on_actionAligned_Polychords_toggled(bool showAlignedPolychords) {
  patchesData.showPolychords = showAlignedPolychords;
  retopoPatchesData.showPolychords = showAlignedPolychords;
  ui->centralWidget->clearButtons();
}
