#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <SCIP/scip_solver.h>      // because some structs in <SCIP/scip/struct_misc.h> have a member named 'slots' which conflicts with Qt
#include <QMainWindow>
#include <QKeyEvent>
#include "glarea.h"
#include "patchrenderer.h"
#include "patchretopo.h"

namespace Ui {
class MainWindow;
}

class MainWindow : public QMainWindow
#ifndef SINGLE_THREAD
                 , public MT_PolygonEnumerationNotifiable
#endif
                 , public std::enable_shared_from_this<MainWindow> {

  Q_OBJECT

public:
  explicit MainWindow(QWidget *parent = 0);
  ~MainWindow();

#ifndef SINGLE_THREAD
  void notify(NotificationType notificationType,
              num_type nTrees, count_type nPolygons,
              std::chrono::nanoseconds duration);

  void registerForNotifications();
#endif

signals:
  void toStatusBarText(const QString &text);

protected:
  void keyPressEvent(QKeyEvent *e);

private slots:
  void on_actionOpen_triggered();
  void on_actionCollapse_triggered();
  void on_actionSave_triggered();
  void on_actionEnumerate_Polygons_triggered();
  void on_actionNavigate_triggered();
  void on_actionFind_Patches_triggered();
  void on_actionRetopo_Patches_triggered();
  void closeFindPatchesTool();
  void on_actionSplit_polychord_triggered();
  void on_actionFullscreen_triggered();
  void on_actionSnapshot_triggered();
  void on_actionBase_Patch_toggled(bool visible);
  void on_actionShowVertices(QAction *action);
  void on_actionKeep_Polygons_toggled(bool keep);
  void on_actionLabels_toggled(bool showLabels);
  void on_actionPreview_toggled(bool showPreview);
  void on_actionAligned_Vectors_toggled(bool showAlignedVectors);
  void on_actionAligned_Polychords_toggled(bool showAlignedPolychords);
    void toStatusBar(const QString &msg);
  void setInputMode(bool draw);
  void activateTrackBall(bool a);

private:
    // GUI objects
    Ui::MainWindow                       *ui;
    std::shared_ptr<PatchRenderer>        patchRenderer;                  ///< Tool window to explore the learnt patches.
    std::shared_ptr<PatchRetopo>          patchRetopo;                    ///< Tool window to explore the retopologized learnt patches.
    enum ViewMode {
      DEFAULT       = 0,
      NAVIGATE_MESH = 1,
      NAVIGATE_TREE = 2
    };
    int                                   viewMode;                       ///< How to see the input mesh and its enumerated polygons.

    // data objects
    QString                               filename;                       ///< Name of the mesh file.
    RenderData                            meshToLearnData;                ///< Render data on mesh to learn.
    RenderData                            patchesData;                    ///< Render data on patches.
    RenderData                            retopoPatchesData;              ///< Render data on retopo patches.

    // core objects
    typedef vcg::tri::pl::MeshSupport<PolyMesh>               MeshSupport;
    typedef MeshSupport::PosType                              PosType;
    typedef MeshSupport::FaceSet                              FaceSet;
    typedef MeshSupport::FaceSetIterator                      FaceSetIterator;
#ifdef SINGLE_THREAD
    PatchPriorityLookup<PolyMesh>::PatchDBServerPointer       pDBServer;
    PolygonRegister<PolyMesh>                                 polygonRegister;
    PatchLearner<PolyMesh>                                    patchLearner;
    PolygonEnumerator<PolyMesh>                               polygonEnumerator;
    typedef PolygonEnumerator<PolyMesh>::CornersType          CornersType;
    PatchPriorityLookup<PolyMesh>                             patchLookup;
    std::shared_ptr< PatchFlattener<PolyMesh> >               patchGeometrizator;
    PatchPriorityLookup<PolyMesh>                             patchLookupRetopo;
    std::shared_ptr< PatchRetopologizer<PolyMesh,TMesh> >     patchRetopologizer;
#else
    MT_PatchPriorityLookup<PolyMesh>::PatchDBServerPointer    pDBServer;
    TS_PolygonRegister<PolyMesh>                              polygonRegister;
    TS_PatchLearner<PolyMesh>                                 patchLearner;
    MT_PolygonEnumerator<PolyMesh>                            polygonEnumerator;
    typedef MT_PolygonEnumerator<PolyMesh>::CornersType       CornersType;
    MT_PatchPriorityLookup<PolyMesh>                          patchLookup;
    std::shared_ptr< PatchFlattener<PolyMesh> >               patchGeometrizator;
    MT_PatchPriorityLookup<PolyMesh>                          patchLookupRetopo;
    std::shared_ptr< PatchRetopologizer<PolyMesh,TMesh> >     patchRetopologizer;
    std::mutex                                                mutex;
#endif

    // parameters
    std::string                           dbFilename;                     ///< DB filename.
    bool                                  keepNodesInMemory;              ///< true if all polygon nodes must be kept in memory (warning: can require too much memory!)
    num_type                              maxNFaces;                      ///< Threshold on the size of each enumerated polygon.
    std::string                           polygonDBFilename;              ///< Polygon DB filename.
    PosType                               currentPos;                     ///< Pos used to navigate the enumerate polygons.
    n_corners_type                        side;                           ///< Current side to navigate.
#ifndef SINGLE_THREAD
    u_small_int_type                      nThreads;                       ///< The number of threads used to enumerate the polygons.
#endif
};

#endif // MAINWINDOW_H
