#ifndef MESH_UTILITIES_H
#define MESH_UTILITIES_H

#include <vector>
#include <list>
#include <set>
#include <utility>
#include <numeric>
#include <cmath>
#include <eigenlib/Eigen/Dense>
#include <vcg/complex/complex.h>
#include <vcg/complex/algorithms/update/flag.h>
#include <vcg/complex/algorithms/update/selection.h>
#include <vcg/complex/algorithms/update/bounding.h>
#include <vcg/complex/algorithms/update/normal.h>
#include <vcg/simplex/face/pos.h>
#include <vcg/complex/algorithms/polygon_support.h>

#include "learning_quadrangulations.h"
using namespace vcg::tri::pl;

// some colors
static const vcg::Color4b BLACK(vcg::Color4b::Black);
static const vcg::Color4b WHITE(vcg::Color4b::White);
static const vcg::Color4b LIGHTGRAY(vcg::Color4b::LightGray);
static const vcg::Color4b DARKGRAY(vcg::Color4b::DarkGray);
static const vcg::Color4b RED(vcg::Color4b::Red);
static const vcg::Color4b GREEN(vcg::Color4b::Green);
static const vcg::Color4b LIGHTGREEN(0, 153, 153, 255);
static const vcg::Color4b DARKGREEN(vcg::Color4b::DarkGreen);
static const vcg::Color4b BLUE(vcg::Color4b::Blue);
static const vcg::Color4b LIGHTBLUE(vcg::Color4b::LightBlue);
static const vcg::Color4b CYAN(vcg::Color4b::Cyan);
static const vcg::Color4b MAGENTA(vcg::Color4b::Magenta);
static const vcg::Color4b SAND(194, 178, 128, 255);
static const vcg::Color4b DODGERBLUE(30, 144, 255, 255);

class Coloring {
public:
  /**
   * @brief SetVertexColorByValence assigns a color to each vertex based on its valence.
   * @param mesh The input mesh.
   */
  static void SetVertexColorByValence(PolyMesh &mesh) {
    // check input
    if (mesh.IsEmpty())
      return;

    // initialize all vertices to black
    vcg::tri::UpdateColor<PolyMesh>::PerVertexConstant(mesh, DARKGRAY);
    vcg::tri::UpdateFlags<PolyMesh>::VertexClear(mesh);

    // compute vertex valence
    std::vector<int> valence(mesh.vert.size(), 0);
    size_t vInd = 0;
    for (size_t f = 0; f < mesh.face.size(); f++)
      if (!mesh.face[f].IsD()) {
        for (int v = 0; v < mesh.face[f].VN(); v++) {
          vInd = vcg::tri::Index(mesh, mesh.face[f].V(v));
          valence[vInd]++;
          if (vcg::face::IsBorder(mesh.face[f], v)) {
            mesh.vert[vInd].SetB();
            valence[vInd]++;
          }
        }
      }

    // set color
    for (size_t v = 0; v < mesh.vert.size(); v++)
      if (!mesh.vert[v].IsD()) {
        mesh.vert[v].ClearS();
        if ((mesh.vert[v].IsB() && valence[v] != 3) || (!mesh.vert[v].IsB() && valence[v] != 4))
          mesh.vert[v].SetS();
        if (mesh.vert[v].IsB()) {
          if (valence[v] != 3)
            mesh.vert[v].C() = CYAN;
        } else if (valence[v] != 4) {
          if (valence[v] == 2)
            mesh.vert[v].C() = GREEN;
          else if (valence[v] == 3)
            mesh.vert[v].C() = RED;
          else if (valence[v] == 5)
            mesh.vert[v].C() = BLUE;
          else
            mesh.vert[v].C() = MAGENTA;
        }
      }
  }

  /**
   * @brief SetColorScaled scales color such that for min you have BLUE, and for max you have RED.
   * @param c
   * @param min
   * @param max
   * @param v
   */
  template < typename ColorType >
  static void SetColorScaled(vcg::Color4<ColorType> &c, float min, float max, float v, float th) {
      // check min and max
      if (max == min){
          c = vcg::Color4<ColorType>(vcg::Color4<ColorType>::Blue);
          return;
      }
      if (max < min) {
          SetColorScaled(c, max, min, v, th);
          return;
      }

      // check threshold
      assert(th >= min && th <= max);

      // check value
      assert(v >= min && v <= max);

      // generate components
      float R = v <= th ? 0.0 : (v - th) / (max - th);
      float G = v < th ? (v - min) / (th - min) : v > th ? (max - v) / (max - th) : 1.0;
      float B = v < th ? (th - v) / (th - min) : 0.0;
      float A = 1.0;

      // set color
      c.Import(vcg::Color4f(R, G, B, A));
  }

  /**
   * @brief ComputeVertexRadius
   * @param mesh
   * @return
   */
  template < typename MeshType >
  static typename MeshType::ScalarType ComputeVertexRadius(const MeshType &mesh) {
    typename MeshType::ScalarType radius = 0;
    size_t div = 0;
    for (size_t f = 0; f < mesh.face.size(); ++f) {
      if (mesh.face[f].IsD())
        continue;
      for (int v = 0; v < mesh.face[f].VN(); ++v)
        radius += (mesh.face[f].cP1(v) - mesh.face[f].cP(v)).Norm();
      div += mesh.face[f].VN();
    }
    radius /= div * 16;
    return radius;
  }
};


#endif // MESH_UTILITIES_H
