#include <SCIP/scip_solver.h>      // because some structs in <SCIP/scip/struct_misc.h> have a member named 'slots' which conflicts with Qt
#include <cmath>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QGridLayout>
#include <QLayoutItem>
#include <QGroupBox>
#include <QButtonGroup>
#include <QLabel>
#include <QSpinBox>
#include <QCheckBox>
#include <QPushButton>
#include <QRadioButton>
#include <QSvgGenerator>
#include <QPainter>
#include <QFileDialog>
#include <QInputDialog>
#include <QMessageBox>
#include <QTextStream>
#include <QTextLayout>
#include <QStaticText>
#include <QFile>
#include <QElapsedTimer>
#include <QTime>
#include <QVector>
#include <QLineF>
#if QT_VERSION >= QT_VERSION_CHECK(5,0,0)
  #include <QGuiApplication>
#else
  #include <QApplication>
#endif
#include "patchrenderer.h"

QString PatchRenderer::logStr = QString();

/**
 * @brief PatchRenderer::PatchRenderer
 * @param m
 * @param db
 * @param parent
 */
PatchRenderer::PatchRenderer(RenderData *meshData,
#ifdef SINGLE_THREAD
                             PatchPriorityLookup<PolyMesh>::PatchDBServerPointer pServer,
                             PatchPriorityLookup<PolyMesh> *db,
#else
                             MT_PatchPriorityLookup<PolyMesh>::PatchDBServerPointer pServer,
                             MT_PatchPriorityLookup<PolyMesh> *db,
#endif
                             QWidget *parent) : QScrollArea(parent),
                                                patchGeometrizator(new PatchFlattener<PolyMesh>) {
  renderData = meshData;
  pDBServer = pServer;
  pDBlookup = db;
  assert(pDBServer != 0 && renderData != 0 && pDBlookup != 0);
  showQualityColor = false;
  nCorners = 0;
  currentPatch = 0;
  renderData->meshes.resize(9);
  renderData->currentMesh = renderData->meshes.size() / 2;

  QVBoxLayout *layout = new QVBoxLayout;

  QGroupBox *inputBox = new QGroupBox(tr("Boundary Conditions"));
  BCgridLayout = new QGridLayout;
  // number of sides
  QLabel *nCornersLabel = new QLabel(tr("# corners:"));
  BCgridLayout->addWidget(nCornersLabel, 0, 0, Qt::AlignRight);
  nCornersSpinBox = new QSpinBox;
  nCornersSpinBox->setMinimum(0);
  nCornersSpinBox->setMaximum(std::numeric_limits<n_corners_type>::max());
  connect(nCornersSpinBox, SIGNAL(valueChanged(int)), this, SLOT(changeNCorners(int)));
  BCgridLayout->addWidget(nCornersSpinBox, 0, 1);
  // number of edges
  nEdgesPerSide.resize(1);
  nEdgesPerSide[0] = new IntValue(1);
  QLabel *nedgeLabel = new QLabel(tr("Side 1 - # edges:"));
  BCgridLayout->addWidget(nedgeLabel, 1, 0, Qt::AlignRight);
  QSpinBox *nEdgesSpinBox = new QSpinBox;
  nEdgesSpinBox->setMinimum(1);
  nEdgesSpinBox->setMaximum(std::numeric_limits<int>::max());
  connect(nEdgesSpinBox, SIGNAL(valueChanged(int)), nEdgesPerSide[0], SLOT(setValue(int)));
  connect(nEdgesSpinBox, SIGNAL(valueChanged(int)), this, SLOT(changeRenderBoundary()));
  BCgridLayout->addWidget(nEdgesSpinBox, 1, 1);
  inputBox->setLayout(BCgridLayout);
  connect(this, SIGNAL(executing(bool)), inputBox, SLOT(setDisabled(bool)));
  layout->addWidget(inputBox);

  QGroupBox *searchBox = new QGroupBox(tr("Search"));
  QVBoxLayout *searchLayout = new QVBoxLayout;
  QGridLayout *filterLayout = new QGridLayout;
  QLabel *maxSingularitiesLabel = new QLabel(tr("Max # sing.:"));
  filterLayout->addWidget(maxSingularitiesLabel, 0, 0, Qt::AlignRight);
  QSpinBox *maxSingularitiesSpinBox = new QSpinBox;
  maxSingularitiesSpinBox->setMinimum(0);
  maxSingularitiesSpinBox->setMaximum(std::numeric_limits<int>::max());
  maxSingularitiesSpinBox->setValue(std::numeric_limits<int>::max());
  connect(maxSingularitiesSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setMaxSingularities(int)));
  filterLayout->addWidget(maxSingularitiesSpinBox, 0, 1);
  QLabel *minValenceLabel = new QLabel(tr("Min valence:"));
  filterLayout->addWidget(minValenceLabel, 1, 0, Qt::AlignRight);
  UpdatableMinMaxSpinBox *minValenceSpinBox = new UpdatableMinMaxSpinBox;
  minValenceSpinBox->setMinimum(2);
  minValenceSpinBox->setMaximum(std::numeric_limits<valence_type>::max());
  minValenceSpinBox->setValue(minValenceSpinBox->minimum());
  connect(minValenceSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setMinValence(int)));
  filterLayout->addWidget(minValenceSpinBox, 1, 1);
  QLabel *maxValenceLabel = new QLabel(tr("Max valence:"));
  filterLayout->addWidget(maxValenceLabel, 2, 0, Qt::AlignRight);
  UpdatableMinMaxSpinBox *maxValenceSpinBox = new UpdatableMinMaxSpinBox;
  maxValenceSpinBox->setMinimum(2);
  maxValenceSpinBox->setMaximum(std::numeric_limits<valence_type>::max());
  maxValenceSpinBox->setValue(maxValenceSpinBox->maximum());
  connect(maxValenceSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setMaxValence(int)));
  filterLayout->addWidget(maxValenceSpinBox, 2, 1);
  connect(minValenceSpinBox, SIGNAL(valueChanged(int)), maxValenceSpinBox, SLOT(minimumChanged(int)));
  connect(maxValenceSpinBox, SIGNAL(valueChanged(int)), minValenceSpinBox, SLOT(maximumChanged(int)));
  QLabel *maxConcaveLabel = new QLabel(tr("Max # concav.:"));
  filterLayout->addWidget(maxConcaveLabel, 3, 0, Qt::AlignRight);
  QSpinBox *maxConcaveSpinBox = new QSpinBox;
  maxConcaveSpinBox->setMinimum(0);
  maxConcaveSpinBox->setMaximum(std::numeric_limits<n_corners_type>::max());
  maxConcaveSpinBox->setValue(std::numeric_limits<n_corners_type>::max());
  connect(maxConcaveSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setMaxConcavities(int)));
  filterLayout->addWidget(maxConcaveSpinBox, 3, 1);
  QLabel *qualityThresholdLabel = new QLabel(tr("Quality th.:"));
  filterLayout->addWidget(qualityThresholdLabel, 4, 0, Qt::AlignRight);
  QDoubleSpinBox *qualityThresholdSpinBox = new QDoubleSpinBox;
  qualityThresholdSpinBox->setMinimum(0);
  qualityThresholdSpinBox->setMaximum(1);
  qualityThresholdSpinBox->setSingleStep(0.01);
  qualityThresholdSpinBox->setValue(qualityThresholdSpinBox->maximum());
  connect(qualityThresholdSpinBox, SIGNAL(valueChanged(double)), this, SLOT(setQualityThreshold(double)));
  filterLayout->addWidget(qualityThresholdSpinBox, 4, 1);
  QLabel *ilp_thresholdLabel = new QLabel(tr("ILP th.:"));
  filterLayout->addWidget(ilp_thresholdLabel, 5, 0, Qt::AlignRight);
  QSpinBox *ilp_thresholdSpinBox = new QSpinBox;
  ilp_thresholdSpinBox->setToolTip(tr("It's the min side length over which flow constraints are added to the ILP."));
  ilp_thresholdSpinBox->setMinimum(0);
  ilp_thresholdSpinBox->setMaximum(std::numeric_limits<int>::max());
  ilp_thresholdSpinBox->setValue(pDBlookup->getMinSizeThresholdILP());
  connect(ilp_thresholdSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setILPthreshold(int)));
  filterLayout->addWidget(ilp_thresholdSpinBox, 5, 1);
  QLabel *ilp_toleranceLabel = new QLabel(tr("Tolerance:"));
  filterLayout->addWidget(ilp_toleranceLabel, 6, 0, Qt::AlignRight);
  QSpinBox *ilp_toleranceSpinBox = new QSpinBox;
  ilp_toleranceSpinBox->setToolTip(tr("It's the tolerance on edge alignment of flow constraints which are added to the ILP."));
  ilp_toleranceSpinBox->setMinimum(0);
  ilp_toleranceSpinBox->setMaximum(std::numeric_limits<u_small_int_type>::max());
  ilp_toleranceSpinBox->setValue(pDBlookup->getFlowConstraintsILPtolerance());
  connect(ilp_toleranceSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setTolerance(int)));
  filterLayout->addWidget(ilp_toleranceSpinBox, 6, 1);
  searchLayout->addLayout(filterLayout);
  QPushButton *findButton = new QPushButton(tr("Find"));
  findButton->setMinimumWidth(55);
  findButton->setMaximumWidth(70);
  connect(findButton, SIGNAL(clicked()), this, SLOT(findPatches()));
  connect(this, SIGNAL(executing(bool)), findButton, SLOT(setDisabled(bool)));
  searchLayout->addWidget(findButton);
  searchLayout->setAlignment(findButton, Qt::AlignHCenter);
  QPushButton *stopButton = new QPushButton(tr("Stop"));
  stopButton->setDisabled(true);
  stopButton->setMinimumWidth(55);
  stopButton->setMaximumWidth(70);
  connect(stopButton, SIGNAL(clicked()), this, SLOT(stopSearching()));
  connect(this, SIGNAL(executing(bool)), stopButton, SLOT(setEnabled(bool)));
  searchLayout->addWidget(stopButton);
  searchLayout->setAlignment(stopButton, Qt::AlignHCenter);
  searchBox->setLayout(searchLayout);
  layout->addWidget(searchBox);

  QGroupBox *sortPatchBox = new QGroupBox(tr("Sort Patches"));
  connect(this, SIGNAL(executing(bool)), sortPatchBox, SLOT(setDisabled(bool)));
  QVBoxLayout *sortPatchLayout = new QVBoxLayout;
  QCheckBox *showQualityCheckBox = new QCheckBox(tr("Show quad quality"));
  connect(showQualityCheckBox, SIGNAL(toggled(bool)), this, SLOT(showQuality(bool)));
  sortPatchLayout->addWidget(showQualityCheckBox);
  QButtonGroup *sortButtonGroup = new QButtonGroup;
  connect(sortButtonGroup, SIGNAL(buttonClicked(int)), this, SLOT(selectSortCriterion(int)));
  QHBoxLayout *sortButtonsLayout2 = new QHBoxLayout;
  QRadioButton *flowSortButton = new QRadioButton(tr("Flow"));
  connect(this, SIGNAL(executing(bool)), flowSortButton, SLOT(setDisabled(bool)));
  sortButtonGroup->addButton(flowSortButton, 1);
  sortButtonsLayout2->addWidget(flowSortButton);
  QRadioButton *quadSortButton = new QRadioButton(tr("Quad"));
  connect(this, SIGNAL(executing(bool)), quadSortButton, SLOT(setDisabled(bool)));
  sortButtonGroup->addButton(quadSortButton, 2);
  sortButtonsLayout2->addWidget(quadSortButton);
  QRadioButton *singSortButton = new QRadioButton(tr("Sing"));
  connect(this, SIGNAL(executing(bool)), singSortButton, SLOT(setDisabled(bool)));
  sortButtonGroup->addButton(singSortButton, 3);
  sortButtonsLayout2->addWidget(singSortButton);
  flowSortButton->setChecked(true);
  compareCriteria = pDBlookup->getSortCriteria().front();
  sortPatchLayout->addLayout(sortButtonsLayout2);
  drawLinesCheckBox = new QCheckBox(tr("Draw 2D lines as flows."));
  drawLinesCheckBox->setToolTip(tr("Check this to start drawing lines, uncheck to stop."));
  drawLinesCheckBox->setChecked(false);
  connect(drawLinesCheckBox, SIGNAL(toggled(bool)), this, SLOT(drawLinesForSorting(bool)));
  sortPatchLayout->addWidget(drawLinesCheckBox);
  QCheckBox *smoothCheckBox = new QCheckBox(tr("Smooth with flow constraints"));
  smoothCheckBox->setToolTip(tr("Check this to impose flows as geometric constraints."));
  smoothCheckBox->setChecked(patchGeometrizator->getSmoothWithConstraints());
  smoothCheckBox->setEnabled(drawLinesCheckBox->isChecked());
  connect(drawLinesCheckBox, SIGNAL(toggled(bool)), smoothCheckBox, SLOT(setEnabled(bool)));
  connect(smoothCheckBox, SIGNAL(toggled(bool)), this, SLOT(smoothWithFlowConstraints(bool)));
  sortPatchLayout->addWidget(smoothCheckBox);
  QGridLayout *weightsGridLayout = new QGridLayout;
  QLabel *laplacianLabel = new QLabel(tr("Laplacian weight:"));
  laplacianLabel->setEnabled(drawLinesCheckBox->isChecked());
  connect(drawLinesCheckBox, SIGNAL(toggled(bool)), laplacianLabel, SLOT(setEnabled(bool)));
  weightsGridLayout->addWidget(laplacianLabel, 0, 0, Qt::AlignRight);
  QDoubleSpinBox *laplacianSpinBox = new QDoubleSpinBox;
  laplacianSpinBox->setMinimum(0.01);
  laplacianSpinBox->setMaximum(0.99);
  laplacianSpinBox->setValue(patchGeometrizator->getLaplacianWeight());
  laplacianSpinBox->setDecimals(2);
  laplacianSpinBox->setSingleStep(0.01);
  laplacianSpinBox->setEnabled(drawLinesCheckBox->isChecked());
  connect(laplacianSpinBox, SIGNAL(valueChanged(double)), this, SLOT(setLaplacianWeight(double)));
  connect(drawLinesCheckBox, SIGNAL(toggled(bool)), laplacianSpinBox, SLOT(setEnabled(bool)));
  weightsGridLayout->addWidget(laplacianSpinBox, 0, 1);
  sortPatchLayout->addLayout(weightsGridLayout);
  QPushButton *clearLinesButton = new QPushButton(tr("Clear"));
  clearLinesButton->setEnabled(drawLinesCheckBox->isChecked());
  clearLinesButton->setMinimumWidth(55);
  clearLinesButton->setMaximumWidth(70);
  connect(drawLinesCheckBox, SIGNAL(toggled(bool)), clearLinesButton, SLOT(setEnabled(bool)));
  connect(clearLinesButton, SIGNAL(clicked()), this, SLOT(deleteDrawnLines()));
  sortPatchLayout->addWidget(clearLinesButton);
  sortPatchLayout->setAlignment(clearLinesButton, Qt::AlignHCenter);
  QPushButton *sortPatchButton = new QPushButton(tr("Sort"));
  sortPatchButton->setEnabled(drawLinesCheckBox->isChecked());
  sortPatchButton->setMinimumWidth(55);
  sortPatchButton->setMaximumWidth(70);
  connect(drawLinesCheckBox, SIGNAL(toggled(bool)), sortPatchButton, SLOT(setEnabled(bool)));
  connect(sortPatchButton, SIGNAL(clicked()), this, SLOT(sortPatches()));
  sortPatchLayout->addWidget(sortPatchButton);
  sortPatchLayout->setAlignment(sortPatchButton, Qt::AlignHCenter);
  sortPatchBox->setLayout(sortPatchLayout);
  layout->addWidget(sortPatchBox);

  QGroupBox *selectPatchBox = new QGroupBox(tr("Select Patch"));
  connect(this, SIGNAL(enable(bool)), selectPatchBox, SLOT(setEnabled(bool)));
  QVBoxLayout *selectPatchLayout = new QVBoxLayout;
  UpdatableMinMaxSpinBox *selectPatchSpinBox = new UpdatableMinMaxSpinBox;
  selectPatchSpinBox->setMinimum(0);
  selectPatchSpinBox->setMaximum(0);
  connect(this, SIGNAL(setMin(int)), selectPatchSpinBox, SLOT(minimumChanged(int)));
  connect(this, SIGNAL(setMax(int)), selectPatchSpinBox, SLOT(maximumChanged(int)));
  connect(this, SIGNAL(setVal(int)), selectPatchSpinBox, SLOT(setValue(int)));
  connect(selectPatchSpinBox, SIGNAL(valueChanged(int)), this, SLOT(selectPatch(int)));
  connect(this, SIGNAL(patchesChangedAt(int)), this, SLOT(selectPatch(int)));
  selectPatchLayout->addWidget(selectPatchSpinBox);

  patchInfoTextEdit = new QTextEdit(tr("No patches."));
  patchInfoTextEdit->setMinimumHeight(100);
  QFont font = patchInfoTextEdit->currentFont();
  font.setPointSize(10);
  patchInfoTextEdit->setFont(font);
  patchInfoTextEdit->setReadOnly(true);
  selectPatchLayout->addWidget(patchInfoTextEdit);
  selectPatchBox->setLayout(selectPatchLayout);
  selectPatchBox->setDisabled(true);
  layout->addWidget(selectPatchBox);

  QGroupBox *provenanceBox = new QGroupBox(tr("Provenance"));
  connect(this, SIGNAL(enable(bool)), provenanceBox, SLOT(setEnabled(bool)));
  QVBoxLayout *provenanceLayout = new QVBoxLayout;
  showProvenanceCheckBox = new QCheckBox(tr("Show provenance"));
  showProvenanceCheckBox->setToolTip(tr("Check this to show where the current template patch has been found."));
  showProvenanceCheckBox->setChecked(false);
  connect(showProvenanceCheckBox, SIGNAL(toggled(bool)), this, SLOT(showProvenance(bool)));
  provenanceLayout->addWidget(showProvenanceCheckBox);
  QPushButton *nextProvenanceButton = new QPushButton(tr("Next"));
  nextProvenanceButton->setEnabled(showProvenanceCheckBox->isChecked());
  nextProvenanceButton->setMinimumWidth(55);
  nextProvenanceButton->setMaximumWidth(70);
  connect(showProvenanceCheckBox, SIGNAL(toggled(bool)), nextProvenanceButton, SLOT(setEnabled(bool)));
  connect(nextProvenanceButton, SIGNAL(clicked()), this, SLOT(nextProvenance()));
  provenanceLayout->addWidget(nextProvenanceButton);
  provenanceLayout->setAlignment(nextProvenanceButton, Qt::AlignHCenter);
  provenanceBox->setLayout(provenanceLayout);
  provenanceBox->setEnabled(false);
  layout->addWidget(provenanceBox);

  QGroupBox *exportBox = new QGroupBox(tr("Export vector graphics"));
  QVBoxLayout *exportLayout = new QVBoxLayout;
  QPushButton *currentSVGButton = new QPushButton(tr("SVG it"));
  currentSVGButton->setMinimumWidth(55);
  currentSVGButton->setMaximumWidth(70);
  connect(currentSVGButton, SIGNAL(clicked()), this, SLOT(currentPatchToSVG()));
  currentSVGButton->setEnabled(false);
  connect(this, SIGNAL(enable(bool)), currentSVGButton, SLOT(setEnabled(bool)));
  exportLayout->addWidget(currentSVGButton);
  exportLayout->setAlignment(currentSVGButton, Qt::AlignHCenter);
  QPushButton *allSVGButton = new QPushButton(tr("SVG templates"));
  allSVGButton->setMinimumWidth(55);
  allSVGButton->setMaximumWidth(120);
  connect(allSVGButton, SIGNAL(clicked()), this, SLOT(allPatchesToSVG()));
  exportLayout->addWidget(allSVGButton);
  exportLayout->setAlignment(allSVGButton, Qt::AlignHCenter);
  exportBox->setLayout(exportLayout);
  connect(this, SIGNAL(executing(bool)), exportBox, SLOT(setDisabled(bool)));
  layout->addWidget(exportBox);

  QWidget *widget = new QWidget;
  widget->setLayout(layout);
  layout->setSizeConstraint(QLayout::SetMinAndMaxSize);
  setWidget(widget);
  setWidgetResizable(true);
  widget->setMinimumSize(50, 200);
  setMinimumSize(50, 200);
}

/**
 * @brief PatchRenderer::~PatchRenderer
 */
PatchRenderer::~PatchRenderer() {
#ifndef SINGLE_THREAD
  std::shared_ptr<MT_PatchPriorityLookupNotifiable> shared_this =
      std::dynamic_pointer_cast<MT_PatchPriorityLookupNotifiable>(shared_from_this());
  pDBlookup->deregisterNotifiable(shared_this);
#endif
}

#ifndef SINGLE_THREAD
/**
 * @brief PatchRenderer::registerForNotifications
 */
void PatchRenderer::registerForNotifications() {
  std::shared_ptr<MT_PatchPriorityLookupNotifiable> shared_this =
      std::dynamic_pointer_cast<MT_PatchPriorityLookupNotifiable>(shared_from_this());
  pDBlookup->registerNotifiable(shared_this);
}
#endif

/**
 * @brief PatchRenderer::changeNCorners
 * @param n
 */
void PatchRenderer::changeNCorners(int n)
{
  int i = 0;
  QWidget *w = 0;
  if (n < (int)nCorners) {
    for (i = nCorners; i > n; i--) {
      if (i <= 1)
        continue;
      w = BCgridLayout->itemAtPosition(i, 0)->widget();
      delete w;
      w = BCgridLayout->itemAtPosition(i, 1)->widget();
      delete w;
      delete nEdgesPerSide[i-1];
    }
    BCgridLayout->parentWidget()->resize(BCgridLayout->parentWidget()->sizeHint());
    widget()->resize(widget()->sizeHint());

    if (n != 0)
      nEdgesPerSide.resize(n);
    else
      nEdgesPerSide.resize(1);

  } else if (n > 1) {

    nEdgesPerSide.resize(n);

    for (i = nCorners > 0 ? nCorners : 1; i < n; i++) {
      QLabel *nedgeLabel = new QLabel(tr("Side ") + QString::number(i+1) + tr(" - # edges:"));
      BCgridLayout->addWidget(nedgeLabel, i+1, 0, Qt::AlignRight);
      nEdgesPerSide[i] = new IntValue(1);
      QSpinBox *nEdgesSpinBox = new QSpinBox;
      nEdgesSpinBox->setMinimum(1);
      nEdgesSpinBox->setMaximum(std::numeric_limits<int>::max());
      connect(nEdgesSpinBox, SIGNAL(valueChanged(int)), nEdgesPerSide[i], SLOT(setValue(int)));
      connect(nEdgesSpinBox, SIGNAL(valueChanged(int)), this, SLOT(changeRenderBoundary()));
      BCgridLayout->addWidget(nEdgesSpinBox, i+1, 1);
      QWidget::setTabOrder(BCgridLayout->itemAtPosition(i, 1)->widget(), nEdgesSpinBox);
    }
  }

  nCorners = n;

  changeRenderBoundary();
}

/**
 * @brief PatchRenderer::changeRenderBoundary
 */
void PatchRenderer::changeRenderBoundary() {
  renderData->baseMesh.clear();
  for (size_t i = 0; i < renderData->meshes.size(); ++i)
    renderData->meshes[i].clear();
  renderData->currentMesh = renderData->meshes.size() / 2;
  renderData->boundary.sides.clear();
  renderData->boundary.colors.clear();
  renderData->flows.clear();
  renderData->flowsColors.clear();
  renderData->nFlows = 0;
  emit changed();
  // build the boundary
  std::vector<num_type> boundary(nEdgesPerSide.size());
  for (size_t i = 0; i < boundary.size(); ++i)
    boundary[i] = nEdgesPerSide[i]->getValue();

  num_type sum = std::accumulate(boundary.begin(), boundary.end(), 0);
  if (sum < 4)
    return;
  if (sum % 2 != 0)
    return;

  PatchFlattener<PolyMesh>::BoundaryToPolygon(boundary, renderData->boundary.sides, patchGeometrizator->getRadius());
  renderData->boundary.colors.resize(boundary.size());
  for (size_t side = 0; side < renderData->boundary.colors.size(); ++side)
    renderData->boundary.colors[side] = vcg::Color4b::Scatter(renderData->boundary.colors.size(), side);

  topologyProvenance.clear();
  meshProvenance.clear();
  cornersProvenance.clear();
  renderData->provMesh.mesh.Clear();
  showProvenanceCheckBox->setChecked(false);

  emit changed();
}

/**
 * @brief PatchRenderer::closeEvent
 * @param e
 */
void PatchRenderer::closeEvent(QCloseEvent *e) {
  if (drawLinesCheckBox->isChecked()) {
    drawLinesCheckBox->setChecked(false);
    drawLinesForSorting(false);
  }
  showProvenanceCheckBox->setChecked(false);
  QScrollArea::closeEvent(e);
  emit closed();
}

/**
 * @brief PatchRenderer::reset
 */
void PatchRenderer::reset() {
  changeNCorners(0);
  nCornersSpinBox->setValue(0);
  nEdgesPerSide[0]->setValue(1);
  QSpinBox *nEdgesSpinBox = (QSpinBox *)BCgridLayout->itemAtPosition(1, 1)->widget();
  nEdgesSpinBox->setValue(1);
  renderData->baseMesh.clear();
  for (size_t i = 0; i < renderData->meshes.size(); ++i)
    renderData->meshes[i].clear();
  renderData->currentMesh = renderData->meshes.size() / 2;
  renderData->boundary.sides.clear();
  renderData->boundary.colors.clear();
  renderData->flows.clear();
  renderData->flowsColors.clear();
  renderData->nFlows = 0;
  description.clear();
  baseDescription.clear();
  logStr.clear();
  msg.clear();
  patchInfoTextEdit->setText(logStr);
  polygonDB.close();
  topologyProvenance.clear();
  meshProvenance.clear();
  cornersProvenance.clear();
  renderData->provMesh.mesh.Clear();
  emit toStatus(msg);
}

/**
 * @brief PatchRenderer::setPolygonDBFilename
 * @param filename
 */
void PatchRenderer::setPolygonDBFilename(const string &filename) {
  polygonDB.close();
  polygonDB.setFilename(filename);
  polygonDB.initialize();
}

/**
 * @brief PatchRenderer::setMaxSingularities
 * @param maxS
 */
void PatchRenderer::setMaxSingularities(int maxS) {
  if (pDBlookup && maxS >= 0)
    pDBlookup->setMaxSingularities(maxS);
}

/**
 * @brief PatchRenderer::setMinValence
 * @param minV
 */
void PatchRenderer::setMinValence(int minV) {
  if (pDBlookup && minV >= 2 && minV <= std::numeric_limits<valence_type>::max())
    pDBlookup->setMinValence(minV);
}

/**
 * @brief PatchRenderer::setMaxValence
 * @param maxV
 */
void PatchRenderer::setMaxValence(int maxV) {
  if (pDBlookup && maxV >= 2 && maxV <= std::numeric_limits<valence_type>::max())
    pDBlookup->setMaxValence(maxV);
}

/**
 * @brief PatchRenderer::setMaxConcavities
 * @param maxC
 */
void PatchRenderer::setMaxConcavities(int maxC) {
  if (pDBlookup && maxC >= 0 && maxC <= std::numeric_limits<n_corners_type>::max())
    pDBlookup->setMaxConcavities(maxC);
}

/**
 * @brief PatchRenderer::setQualityThreshold
 * @param th
 */
void PatchRenderer::setQualityThreshold(double th) {
  if (pDBlookup && th >= 0 && th <= 1) {
    patchGeometrizator->setQualityThreshold(th);
    pDBlookup->clear();
  }
}

/**
 * @brief PatchRenderer::setILPthreshold
 * @param th
 */
void PatchRenderer::setILPthreshold(int th) {
  if (pDBlookup && th >= 0)
    pDBlookup->setMinSizeThresholdILP(th);
}

/**
 * @brief PatchRenderer::setTolerance
 * @param tolerance
 */
void PatchRenderer::setTolerance(int tolerance) {
  if (pDBlookup && tolerance >= 0)
    pDBlookup->setFlowConstraintsILPtolerance(tolerance);
}

/**
 * @brief PatchRenderer::findPatches
 */
void PatchRenderer::findPatches() {
  // build the boundary
  std::vector<num_type> boundary(nEdgesPerSide.size());
  for (size_t i = 0; i < boundary.size(); ++i)
    boundary[i] = nEdgesPerSide[i]->getValue();

  num_type sum = std::accumulate(boundary.begin(), boundary.end(), 0);
  if (sum < 4) {
    QMessageBox::critical(0, tr("Wrong boundary conditions"), tr("Boundary edges must be at least 4!"));
    return;
  }
  if (sum % 2 != 0) {
    QMessageBox::critical(0, tr("Wrong boundary conditions"), tr("Odd boundary conditions!"));
    return;
  }

  // update GUI
  topologyProvenance.clear();
  meshProvenance.clear();
  cornersProvenance.clear();
  renderData->provMesh.mesh.Clear();
  emit setMin(0);
  emit setVal(0);
  emit setMax(0);
  emit enable(false);
  renderData->baseMesh.clear();
  for (size_t i = 0; i < renderData->meshes.size(); ++i)
    renderData->meshes[i].clear();
  emit changed();
  emit executing(true);
  msg = "Searching...";
  emit toStatus(msg);

  // search
  currentPatch = -1;
  elapsedFromPreviousUIupdated = std::chrono::nanoseconds::zero();
  pDBlookup->findPatchesFromBoundaryLength(nCorners, boundary, renderData->flows, patchGeometrizator);
#ifdef SINGLE_THREAD
  count_type nPatches = pDBlookup->numberOfPatches();
  if (nPatches > 0) {
    emit setMin(1);
    emit setMax(nPatches);
    emit enable(true);
    currentPatch = 0;
    emit patchesChangedAt(currentPatch + 1);
    emit executing(false);
    msg += " Expanded: " + QString::number(nPatches) + ". Complete!";
    emit toStatus(msg);
  }
#endif
}

/**
 * @brief PatchRenderer::stopSearching
 */
void PatchRenderer::stopSearching() {
  if (pDBlookup)
    pDBlookup->stop();
}

/**
 * @brief PatchRenderer::showQuality
 * @param show
 */
void PatchRenderer::showQuality(bool show) {
  showQualityColor = show;
  for (size_t i = 0; i < renderData->meshes.size(); ++i) {
    if (showQualityColor)
      vcg::tri::UpdateColor<PolyMesh>::PerFaceQualityRamp(renderData->meshes[i].mesh, 0, 1);
    else
      vcg::tri::UpdateColor<PolyMesh>::PerFaceConstant(renderData->meshes[i].mesh, LIGHTGRAY);
  }
  emit changed();
}

/**
 * @brief PatchRenderer::smoothWithFlowConstraints
 * @param smooth
 */
void PatchRenderer::smoothWithFlowConstraints(bool smooth) {
  patchGeometrizator->setSmoothWithConstraints(smooth);
}

/**
 * @brief PatchRenderer::drawLinesForSorting
 * @param draw
 */
void PatchRenderer::drawLinesForSorting(bool draw) {
  if (draw)
    showProvenanceCheckBox->setChecked(false);
  emit drawLines(draw);
}

/**
 * @brief PatchRenderer::setLaplacianWeight
 * @param weight
 */
void PatchRenderer::setLaplacianWeight(double weight) {
  patchGeometrizator->setLaplacianWeight(weight);
}

/**
 * @brief PatchRenderer::deleteDrawnLines
 */
void PatchRenderer::deleteDrawnLines() {
  renderData->flows.clear();
  renderData->flowsColors.clear();
  renderData->nFlows = 0;
  for (size_t i = 0; i < renderData->meshes.size(); ++i) {
    renderData->meshes[i].polychordMap.clear();
  }
  emit changed();
}

/**
 * @brief PatchRenderer::sortPatches
 */
void PatchRenderer::sortPatches() {
  renderData->baseMesh.clear();
  for (size_t i = 0; i < renderData->meshes.size(); ++i)
    renderData->meshes[i].clear();
  // update GUI
  topologyProvenance.clear();
  meshProvenance.clear();
  cornersProvenance.clear();
  renderData->provMesh.mesh.Clear();
  emit setMin(0);
  emit setVal(0);
  emit setMax(0);
  emit enable(false);
  emit executing(true);
  msg = "Sorting...";
  emit toStatus(msg);
  elapsedFromPreviousUIupdated = std::chrono::nanoseconds::zero();
  pDBlookup->sortPatches(renderData->flows, patchGeometrizator);
#ifdef SINGLE_THREAD
  count_type nPatches = pDBlookup->numberOfPatches();
  if (nPatches > 0) {
    emit setMin(1);
    emit setMax(nPatches);
    emit enable(true);
    emit executing(false);
    msg += " Expanded: " + QString::number(nPatches) + ". Complete!";
    emit toStatus(msg);
  }
#endif
}

#ifndef SINGLE_THREAD
void PatchRenderer::notify(NotificationType notificationType,
                           count_type nTemplatePatches, count_type nPatches,
                           std::chrono::nanoseconds duration) {
  std::string durationStr;
  TimeElements timeElements;
  TimeSupport::SplitTimeElements(duration, timeElements);
  TimeSupport::TimeElementsToString(timeElements, durationStr);
  bool emitSignal = false;
  msg = "Elapsed: ";
  msg += durationStr.c_str();
  msg += " Rotated templates: " + QString::number(nTemplatePatches) + '.';

  mutexForTime.lock();
  if (duration - elapsedFromPreviousUIupdated >= std::chrono::milliseconds(100))
    emitSignal = true;
  mutexForTime.unlock();

  switch (notificationType) {
  case STRUCTURE_FILTER:
    msg += " Filtering by structure";
    break;
  case FLOW_FILTER:
    msg += " Filtering by flows";
    break;
  case FLUSHING:
    msg += " Flushing the buffer: " + QString::number(nPatches) + " previously computed patches to delete";
    break;
  case ILP_FILTER:
    msg += " Expanded: " + QString::number(nPatches) + ". Filtering by ILP";
    if (nPatches > 0) {
      mutexForTime.lock();
      if (emitSignal) {
        elapsedFromPreviousUIupdated = duration;
        if (currentPatch < 0)
          currentPatch = 0;
        emit setMax(nPatches);
        emit patchesChangedAt(currentPatch + 1);
        emit setMin(1);
        emit enable(true);
      }
      mutexForTime.unlock();
    }
    break;
  case SORT:
    msg += " Sorted " + QString::number(nPatches) + " patches. Sorting";
    if (nPatches > 0) {
      mutexForTime.lock();
      if (emitSignal) {
        elapsedFromPreviousUIupdated = duration;
        if (currentPatch < 0)
          currentPatch = 0;
        emit setMax(nPatches);
        emit patchesChangedAt(currentPatch + 1);
        emit setMin(1);
        emit enable(true);
      }
      mutexForTime.unlock();
    }
    break;
  case COMPLETE:
    msg += " Found: " + QString::number(nPatches) + ". Complete!";
    emitSignal = true;
    if (nPatches > 0) {
      if (currentPatch < 0)
        currentPatch = 0;
      emit setMax(nPatches);
      emit patchesChangedAt(currentPatch + 1);
      emit setMin(1);
      emit enable(true);
    }
    emit executing(false);
    break;
  default:
    break;
  }

  if (notificationType != EMPTY && notificationType != COMPLETE)
    for (int i = 0; i <= timeElements.s.count() % 3; ++i)
      msg += '.';
  if (emitSignal)
    emit toStatus(msg);
}
#endif

/**
 * @brief PatchRenderer::printBoundaryIDs
 * @param startPos
 * @param mesh
 * @param bEq
 * @param bVars
 */
void PatchRenderer::printBoundaryIDs(const vcg::face::Pos<PolyFace> &startPos, const std::string &bEq,
                                     const std::vector<num_type> &bVars,
                                     std::list<std::pair<var_label_type, CoordType> > &labels) {
  labels.clear();
  // print labels
  PolyMesh::CoordType v1(0,0,0), v2(0,0,0);
  std::stringstream stream;
  size_t sInd = 0, vInd = 0;
  stream << bEq.substr(sInd + 1, 2 * sizeof(var_label_type)) + std::string(" ");
  stream >> std::hex >> vInd;
  var_label_type count = 0;
  vcg::face::Pos<PolyFace> runPos = startPos;
  do {
    // find location
    v1 = runPos.VFlip()->P() - runPos.V()->P();
    v2[0] = v1[1];
    v2[1] = -v1[0];
    v1 = (runPos.V()->P() + runPos.VFlip()->P()) / 2.0 + v2 * 0.2;
    // build the label
    labels.push_back(std::pair<var_label_type, PolyMesh::CoordType>(vInd, v1));
    // go to next edge
    runPos.FlipV();
    runPos.FlipE();
    if (!runPos.IsBorder()) {
      runPos.FlipF();
      runPos.FlipE();
      if (!runPos.IsBorder()) {   // concave corner
        while (!runPos.IsBorder()) {
          runPos.FlipF();
          runPos.FlipE();
        }
        ++sInd;
      }
    } else {    // convex corner
      // skip SEPARATOR
      ++sInd;
    }
    // next label
    ++count;
    if (count >= bVars[vInd] && runPos != startPos) {
      count = 0;
      sInd += 1/*VARLABEL*/ + 2 * sizeof(var_label_type)/*variable index*/;
      stream.str(bEq.substr(sInd + 1, 2 * sizeof(var_label_type)) + std::string(" "));
      stream >> std::hex >> vInd;
    }
  } while (runPos != startPos);
}

/**
 * @brief PatchRenderer::selectPatch
 * @param i
 */
void PatchRenderer::selectPatch(int i) {
  std::string tmpStr;
  if (i > 0 && (count_type)i <= pDBlookup->numberOfPatches()) {
    currentPatch = i-1;

    // expand patches
    if (renderData->meshes.empty())
      renderData->meshes.resize(1);
    Patch<PolyMesh> emptyPatch;
    std::deque< Patch<PolyMesh> > patches;
    count_type startIndex = std::max(currentPatch - (int)renderData->currentMesh, 0);
    count_type offset = renderData->meshes.size() - (renderData->currentMesh - (currentPatch - startIndex));
    pDBlookup->at(startIndex, offset, patches);
    for (int nEmptyBefore = currentPatch - (int)renderData->currentMesh; nEmptyBefore < 0; ++nEmptyBefore)
      patches.push_front(emptyPatch);
    while (patches.size() < renderData->meshes.size())
      patches.push_back(emptyPatch);
    for (size_t j = 0; j < renderData->meshes.size(); ++j)
      patchToMeshData(patches[j], renderData->meshes[j], showQualityColor);

    // description
    description.clear();
    description += "Topology: " + QString::fromStdString(patches.at(renderData->currentMesh).getTopologyStr()) + '\n';
    patches.at(renderData->currentMesh).getBoundaryIDsStr(tmpStr);
    description += "BoundaryIDs: " + QString::fromStdString(tmpStr) + '\n';
    patches.at(renderData->currentMesh).getBoundaryVarStr(tmpStr);
    description += "BoundaryVars: " + QString::fromStdString(tmpStr) + '\n';
    patches.at(renderData->currentMesh).getBoundaryIDsEndsStr(tmpStr);
    description += "BoundaryIDsEnds: " + QString::fromStdString(tmpStr) + '\n';
    patches.at(renderData->currentMesh).getBoundaryLenStr(tmpStr);
    description += "BoundaryLens: " + QString::fromStdString(tmpStr) + '\n';
    description += "# singularities: " + QString::number(patches.at(renderData->currentMesh).getNumberOfSingularities()) + '\n';
    description += "Singularities: " + QString::fromStdString(patches.at(renderData->currentMesh).getSingularityStr()) + '\n';
    description += "# corners: " + QString::number(patches.at(renderData->currentMesh).getNumberOfCorners()) + '\n';
    description += "# faces: " + QString::number(patches.at(renderData->currentMesh).getNumberOfFaces()) + '\n';
    description += "# vertices: " + QString::number(patches.at(renderData->currentMesh).getNumberOfVertices()) + '\n';
    description += "# occurrences: " + QString::number(patches.at(renderData->currentMesh).getNumberOfOccurrences()) + '\n';
    description += "Perimeter (reconstructed patch): " + QString::number(patches.at(renderData->currentMesh).getPerimeter()) + '\n';
    description += "# rotations (reconstructed patch): " + QString::number(patches.at(renderData->currentMesh).getNumberOfRotations()) + '\n';
    description += "# faces (reconstructed patch): " + QString::number(renderData->meshes[renderData->currentMesh].mesh.FN()) + '\n';
    description += "# vertices (reconstructed patch): " + QString::number(renderData->meshes[renderData->currentMesh].mesh.VN());

    // base mesh
    Patch<PolyMesh> basePatch = patches.at(renderData->currentMesh);
    basePatch.getPolychordMap().clear();
    std::string bIDs;
    basePatch.getBoundaryIDsStr(bIDs);
    basePatch.setBoundaryFromStr(bIDs);
    basePatch.buildExpandedTopologyStr();
    PatchFlattener<PolyMesh>::PatchFlatteningBySmoothing(basePatch.getMesh(), basePatch.getStartCorner(),
                                                         basePatch.getBoundaryLenVec(), patchGeometrizator->getRadius());
    vcg::tri::UpdateBounding<PolyMesh>::Box(basePatch.getMesh());
    patchToMeshData(basePatch, renderData->baseMesh);
    renderData->baseMesh.vertexRadius = Coloring::ComputeVertexRadius(renderData->baseMesh.mesh);
    // description
    baseDescription.clear();
    baseDescription += "Topology: " + QString::fromStdString(basePatch.getTopologyStr()) + '\n';
    basePatch.getBoundaryIDsStr(tmpStr);
    baseDescription += "BoundaryIDs: " + QString::fromStdString(tmpStr) + '\n';
    basePatch.getBoundaryVarStr(tmpStr);
    baseDescription += "BoundaryVars: " + QString::fromStdString(tmpStr) + '\n';
    basePatch.getBoundaryIDsEndsStr(tmpStr);
    baseDescription += "BoundaryIDsEnds: " + QString::fromStdString(tmpStr) + '\n';
    basePatch.getBoundaryLenStr(tmpStr);
    baseDescription += "BoundaryLens: " + QString::fromStdString(tmpStr) + '\n';
    baseDescription += "Singularities: " + QString::fromStdString(basePatch.getSingularityStr()) + '\n';
    baseDescription += "Perimeter: " + QString::number(basePatch.getPerimeter()) + '\n';
    baseDescription += "# corners: " + QString::number(basePatch.getNumberOfCorners()) + '\n';
    baseDescription += "# faces: " + QString::number(basePatch.getNumberOfFaces()) + '\n';
    baseDescription += "# vertices: " + QString::number(basePatch.getNumberOfVertices()) + '\n';
    baseDescription += "# rotations: " + QString::number(basePatch.getNumberOfRotations()) + '\n';
    baseDescription += "# occurrences: " + QString::number(basePatch.getNumberOfOccurrences());

    if (basePatch.getTopologyStr() != topologyProvenance) {
      topologyProvenance = basePatch.getTopologyStr();
      if (polygonDB.isInitialized()) {
        polygonDB.querySelect(topologyProvenance);
        if (!renderData->provMesh.mesh.IsEmpty())
          nextProvenance();
      }
    }

    logStr = description;
    patchInfoTextEdit->setText(logStr);
  } else {
    currentPatch = -1;
    renderData->baseMesh.clear();
    for (size_t i = 0; i < renderData->meshes.size(); ++i)
      renderData->meshes[i].clear();
    description.clear();
    baseDescription.clear();
    logStr = tr("No patches found.");
    patchInfoTextEdit->setText(logStr);
    topologyProvenance.clear();
    meshProvenance.clear();
    cornersProvenance.clear();
    renderData->provMesh.mesh.Clear();
  }

  emit changed();
}

/**
 * @brief PatchRenderer::patchToMeshData
 * @param patch
 * @param meshData
 */
void PatchRenderer::patchToMeshData(const Patch<PolyMesh> &patch, MeshData &meshData, const bool qualityColor) {
  meshData.clear();
  if (patch.isNull())
    return;
  // copy mesh
  vcg::face::Pos<PolyFace> startPos;
  patch.getMeshAndStartCorner(meshData.mesh, startPos);
  patch.getPolychordMap(meshData.mesh, meshData.polychordMap);
  // initialize mark
  vcg::tri::InitFaceIMark(meshData.mesh);
  vcg::tri::InitVertexIMark(meshData.mesh);
  vcg::tri::IMark(meshData.mesh) = 0;
  // set vertex radius
  meshData.vertexRadius = renderData->vertexRadius;
  // colorize faces
  if (qualityColor)
    vcg::tri::UpdateColor<PolyMesh>::PerFaceQualityRamp(meshData.mesh, 0, 1);
  else
    vcg::tri::UpdateColor<PolyMesh>::PerFaceConstant(meshData.mesh, LIGHTGRAY);
  // colorize vertices
  Coloring::SetVertexColorByValence(meshData.mesh);
  // colorize the boundary
  vcg::face::Pos<PolyFace> runPos = startPos;
  std::list<CoordType> l;
  n_corners_type range = patch.getNumberOfCorners();
  if (range == 0)
    range = 1;
  n_corners_type side = 0;
  vcg::Color4b c = vcg::Color4b::Scatter(range, side);
  do {
    l.push_back(runPos.V()->P());
    runPos.FlipV();
    runPos.FlipE();
    if (!runPos.IsBorder()) {
      runPos.FlipF();
      runPos.FlipE();
      if (!runPos.IsBorder()) {   // concave corner
        while (!runPos.IsBorder()) {
          runPos.FlipF();
          runPos.FlipE();
        }
        l.push_back(runPos.V()->P());
        c = vcg::Color4b::Scatter(range, side);
        meshData.polyLines.push_back(PolyLine(c, l));
        l.clear();
        ++side;
      } else {
        if (runPos == startPos) {
          l.push_back(runPos.V()->P());
        }
      }
    } else {    // convex corner
      l.push_back(runPos.V()->P());
      c = vcg::Color4b::Scatter(range, side);
      meshData.polyLines.push_back(PolyLine(c, l));
      l.clear();
      ++side;
    }
  } while (runPos != startPos);
  if (patch.getNumberOfCorners() == 0)
    meshData.polyLines.push_back(PolyLine(c, l));
  // print labels
  std::string bEqTmp;
  patch.getBoundaryIDsStr(bEqTmp);
  size_t ind = 0;
  for (n_corners_type r = 0; r < patch.getNumberOfRotations(); ++r) {
    while (bEqTmp[ind] != SEPARATOR)
      ++ind;
    ++ind;
  }
  std::string bEq(bEqTmp);
  for (size_t i = 0; i < bEq.length(); ++i)
    if (ind >= bEqTmp.length()) {
      bEq[i] = SEPARATOR;
      ind = 0;
    } else {
      bEq[i] = bEqTmp[ind];
      ++ind;
    }
  std::list< std::pair<var_label_type, CoordType> > labelPosList;
  printBoundaryIDs(startPos, bEq, patch.getBoundaryVarVec(), labelPosList);
  std::list< std::pair<var_label_type, CoordType> >::iterator lIt;
  std::stringstream stream;
  std::string str;
  for (lIt = labelPosList.begin(); lIt != labelPosList.end(); ++lIt) {
    str.clear();
    stream.str(str);
    stream << VARLABEL << "<sub>" << std::hex << std::noshowbase << lIt->first << "</sub>";
    meshData.strings.push_back(Text(DARKGRAY, lIt->second, stream.str()));
  }
  switch (compareCriteria) {
  case CompareCriteria::FLOW_DISTANCE:
    meshData.value = patch.getFlowDistance();
    break;
  case CompareCriteria::QUADS_QUALITY:
    meshData.value = patch.getQuadsQuality();
    break;
  case CompareCriteria::SINGULARITYNUM:
    meshData.value = patch.getNumberOfSingularities();
    break;
  default:
    meshData.value = 0;
    break;
  }
}

/**
 * @brief PatchRenderer::patchToSVG
 * @param mesh
 * @param svgFilename
 * @param description
 * @param lineStrips
 */
void PatchRenderer::patchToSVG(MeshData &meshData, const QString &svgFilename, const QString &description,
                               const bool showFlows, const bool showPolychords, const bool showSamples) {
  if (meshData.mesh.IsEmpty())
    return;

#if QT_VERSION >= QT_VERSION_CHECK(5,0,0)
  QFont font = QGuiApplication::font();
#else
  QFont font = QApplication::font();
#endif
  font.setFamily("Palatino");
  font.setItalic(true);
  font.setPointSize(35);

  ScalarType minW, maxW;
  FlowMapConstIterator flowMapIt;
  RenderData::FlowColorsMap::const_iterator flowColorsMapIt;
  PolychordMapConstIterator polychordMapIt;
  QVector<QPointF> lineStrip;
  QLineF line;
  vcg::Color4b c;

  ScalarType minX = meshData.mesh.bbox.min[0];
  ScalarType minY = meshData.mesh.bbox.min[1];
  ScalarType maxX = meshData.mesh.bbox.max[0];
  ScalarType maxY = meshData.mesh.bbox.max[1];
  ScalarType w = maxX - minX;
  ScalarType h = maxY - minY;
  ScalarType dw = 20;
  ScalarType dh = 20;
  ScalarType dwlL = 0;
  ScalarType dwrL = 0;
  ScalarType dhtL = 0;
  ScalarType dhbL = 0;
  if (renderData->showStrings) {
    QPointF p;
    std::list<Text>::const_iterator lIt = meshData.strings.begin();
    for (; lIt != meshData.strings.end(); ++lIt) {
      QStaticText t(QString(lIt->text.c_str()));
      p.setX(lIt->point[0] - minX);
      p.setY(maxY - lIt->point[1]);
      if (t.size().width()/2 - p.x() > dwlL)
        dwlL = t.size().width()/2 - p.x();
      if (p.x() + t.size().width()/2 - w > dwrL)
        dwrL = p.x() + t.size().width()/2 - w;
      if (p.y() + t.size().height()/2 - h > dhbL)
        dhbL = p.y() + t.size().height()/2 - h;
      if (t.size().height()/2 - p.y() > dhtL)
        dhtL = t.size().height()/2 - p.y();
    }
  }

  QSvgGenerator svgGenerator;
  svgGenerator.setFileName(svgFilename);
  svgGenerator.setDescription(description);
  svgGenerator.setTitle(svgFilename);
  svgGenerator.setSize(QSize(w + dwlL + dwrL + dw, h + dhtL + dhbL + dh));

  QPainter painter;
  painter.begin(&svgGenerator);
  painter.setFont(font);
  painter.setBackgroundMode(Qt::TransparentMode);
  painter.setPen(QPen(QBrush(QColor(DARKGRAY[0], DARKGRAY[1], DARKGRAY[2], DARKGRAY[3])), 2, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
  QPolygonF polygon;
  for (size_t f = 0; f < meshData.mesh.face.size(); ++f)
    if (!meshData.mesh.face[f].IsD()) {
      painter.setBrush(QBrush(QColor(meshData.mesh.face[f].cC()[0], meshData.mesh.face[f].cC()[1],
                                     meshData.mesh.face[f].cC()[2], meshData.mesh.face[f].cC()[3])));
      polygon.clear();
      for (int v = 0; v <meshData.mesh.face[f].VN(); ++v)
        polygon << QPointF(meshData.mesh.face[f].V(v)->P()[0] - minX + dwlL + dw/2,
                           maxY - meshData.mesh.face[f].V(v)->P()[1] + dhtL + dh/2);
      painter.drawPolygon(polygon);
    }
  // draw other lines
  if (!meshData.polyLines.empty()) {
    std::list<CoordType>::const_iterator verticesIt;
    std::list<PolyLine>::const_iterator stripsIt = meshData.polyLines.begin();
    for (; stripsIt != meshData.polyLines.end(); ++stripsIt) {
      // set the color
      painter.setPen(QPen(QBrush(QColor(stripsIt->color[0], stripsIt->color[1], stripsIt->color[2], stripsIt->color[3])), 2, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
      // draw the lines
      polygon.clear();
      verticesIt = stripsIt->points.begin();
      for (; verticesIt != stripsIt->points.end(); ++verticesIt) {
        polygon << QPointF((*verticesIt)[0] - minX + dwlL + dw/2, maxY - (*verticesIt)[1] + dhtL + dh/2);
      }
      painter.drawPolyline(polygon);
    }
  }
  // draw vertices
  if (renderData->showVertices == ShowVertices::ALL || renderData->showVertices == ShowVertices::SINGS) {
    for (size_t v = 0; v < meshData.mesh.vert.size(); ++v) {
      if (meshData.mesh.vert[v].IsD())
        continue;
      if (renderData->showVertices == ShowVertices::SINGS && !meshData.mesh.vert[v].IsS())
        continue;
      // set the color
      painter.setPen(QPen(QBrush(QColor(meshData.mesh.vert[v].cC()[0], meshData.mesh.vert[v].cC()[1], meshData.mesh.vert[v].cC()[2])),
                          10, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
      // draw the point
      painter.drawPoint(QPointF(meshData.mesh.vert[v].P()[0] - minX + dwlL + dw/2, maxY - meshData.mesh.vert[v].P()[1] + dhtL + dh/2));
    }
  }

  // drawn lines and polychords
  if (showFlows) {
    if (showSamples) {
      // crossed lines
      minW = std::numeric_limits<ScalarType>::max();
      maxW = 0.0;
      for (polychordMapIt = meshData.polychordMap.begin(); polychordMapIt != meshData.polychordMap.end(); ++polychordMapIt) {
        const PolychordContainer &polychordContainer = polychordMapIt->second;
        for (size_t i = 0; i < polychordContainer.size(); ++i) {
          const PolychordPoints &polychordPoints = polychordContainer[i];
          for (size_t p = 0; p < polychordPoints.distances.size(); ++p) {
            if (polychordPoints.distances[p] < minW)
              minW = polychordPoints.distances[p];
            if (polychordPoints.distances[p] > maxW)
              maxW = polychordPoints.distances[p];
          }
        }
      }
    }
    for (flowMapIt = renderData->flows.begin(); flowMapIt != renderData->flows.end(); ++flowMapIt) {
      flowColorsMapIt = renderData->flowsColors.find(flowMapIt->first);
      assert(flowColorsMapIt != renderData->flowsColors.end());
      const FlowContainer &flowContainer = flowMapIt->second;
      const std::deque<vcg::Color4b> &flowColors = flowColorsMapIt->second;
      assert(flowContainer.size() == flowColors.size());
      polychordMapIt = meshData.polychordMap.find(flowMapIt->first);
      for (size_t i = 0; i < flowContainer.size() && i < flowColors.size(); ++i) {
        // set the color
        painter.setPen(QPen(QBrush(QColor(flowColors[i][0], flowColors[i][1], flowColors[i][2])),
                            4, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
        // draw flows
        const FlowPoints &flowPoints = flowContainer[i];
        lineStrip.clear();
        for (size_t p = 0; p < flowContainer[i].points.size(); ++p) {
          lineStrip.push_back(QPointF(flowPoints.points[p][0] - minX + dwlL + dw/2, maxY - flowPoints.points[p][1] + dhtL + dh/2));
          lineStrip.push_back(QPointF(flowPoints.points[p][0] - minX + dwlL + dw/2, maxY - flowPoints.points[p][1] + dhtL + dh/2));
        }
        lineStrip.pop_front();
        lineStrip.pop_back();
        painter.drawLines(lineStrip);
        // draw polychords
        if (showPolychords && polychordMapIt != meshData.polychordMap.end() && i < polychordMapIt->second.size()) {
          const PolychordPoints &polychordPoints = polychordMapIt->second[i];
          // set the color
          painter.setPen(QPen(QBrush(QColor(flowColors[i][0], flowColors[i][1], flowColors[i][2])),
                              3, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
          lineStrip.clear();
          for (size_t p = 0; p < polychordPoints.points.size(); ++p) {
            lineStrip.push_back(QPointF(polychordPoints.points[p][0] - minX + dwlL + dw/2, maxY - polychordPoints.points[p][1] + dhtL + dh/2));
            lineStrip.push_back(QPointF(polychordPoints.points[p][0] - minX + dwlL + dw/2, maxY - polychordPoints.points[p][1] + dhtL + dh/2));
          }
          lineStrip.pop_front();
          lineStrip.pop_back();
          painter.drawLines(lineStrip);
        }
        // draw sample vectors
        if (showSamples && polychordMapIt != meshData.polychordMap.end() && i < polychordMapIt->second.size()) {
          const PolychordPoints &polychordPoints = polychordMapIt->second[i];
          lineStrip.clear();
          for (size_t s = 0; s < polychordPoints.samplePoints.size() && s < flowPoints.samplePoints.size(); ++s) {
            Coloring::SetColorScaled(c, minW, maxW, polychordPoints.distances[s], (minW + maxW) * 0.5);
            // set the color
            painter.setPen(QPen(QBrush(QColor(c[0], c[1], c[2])), 2, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
            line.setP1(QPointF(flowPoints.samplePoints[s][0] - minX + dwlL + dw/2, maxY - flowPoints.samplePoints[s][1] + dhtL + dh/2));
            line.setP2(QPointF(polychordPoints.samplePoints[s][0] - minX + dwlL + dw/2, maxY - polychordPoints.samplePoints[s][1] + dhtL + dh/2));
            painter.drawLine(line);
          }
        }
      }
    }
  }

  // print labels
  if (renderData->showStrings) {
    painter.setPen(QPen(QBrush(QColor(DARKGRAY[0], DARKGRAY[1], DARKGRAY[2], DARKGRAY[3])), 2, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
    std::list<Text>::const_iterator lIt = meshData.strings.begin();
    QString str;
    for (; lIt != meshData.strings.end(); ++lIt) {
      str = lIt->text.c_str();
      QStaticText t(str);
      t.setTextFormat(Qt::RichText);
      t.prepare(QTransform(), painter.font());
      QPointF point(lIt->point[0] - minX + dwlL + dw/2 - t.size().width()/2, maxY - lIt->point[1] + dhtL + dh/2 - t.size().height()/2);
      painter.drawStaticText(point, t);
    }
  }
  painter.end();
}

/**
 * @brief PatchRenderer::currentPatchToSVG
 */
void PatchRenderer::currentPatchToSVG() {
  if (renderData->currentMesh >= renderData->meshes.size() || renderData->meshes[renderData->currentMesh].mesh.IsEmpty())
    return;
  QString svgFilename = QFileDialog::getSaveFileName(0, tr("Save SVG ..."), QString(), tr("SVG files (*.svg)"));
  if (!svgFilename.isEmpty()) {
    int pos = svgFilename.lastIndexOf(QChar('.'));
    // save patch
    patchToSVG(renderData->meshes[renderData->currentMesh], svgFilename, description, true, renderData->showPolychords, renderData->showSampleVectors);
    // save base patch
    QString svgBaseFilename = svgFilename;
    svgBaseFilename.insert(pos, QString("_base"));
    patchToSVG(renderData->baseMesh, svgBaseFilename, baseDescription);
    // save log file
    QString svgLogFilename = svgFilename;
    svgLogFilename.replace(pos, svgFilename.length() - pos, QString("_log.txt"));
    QFile logFile(svgLogFilename);
    if (!logFile.open(QFile::WriteOnly | QFile::Text | QFile::Truncate)) {
      std::cerr << "Error. Enable to open file '" << logFile.fileName().toStdString() << "'." << std::endl;
      return;
    }
    QTextStream qFileStream(&logFile);
    qFileStream << svgFilename.right(svgFilename.length() - svgFilename.lastIndexOf(QChar('/')) - 1) << ':' << endl;
    qFileStream << description << endl << endl;
    qFileStream << svgBaseFilename.right(svgBaseFilename.length() - svgBaseFilename.lastIndexOf(QChar('/')) - 1) << ':' << endl;
    qFileStream << baseDescription << endl;
    logFile.close();
  }
}

/**
 * @brief PatchRenderer::allPatchesToSVG
 */
void PatchRenderer::allPatchesToSVG() {
  bool ok;
  int nCorners = QInputDialog::getInt(0, tr("How many corners"), tr("Asking for patches with how many corners?"),
                                      -1, -1, std::numeric_limits<int>::max(), 1, &ok);
  if (!ok)
    return;
  int limitI = QInputDialog::getInt(0, tr("Limit"), tr("How many patches?"),
                                    -1, -1, std::numeric_limits<int>::max(), 1, &ok);
  if (!ok)
    return;
  bool fromEnd;
  QMessageBox::StandardButton button = QMessageBox::question(0, tr("Direction"),
                                                             tr("Less frequent patches instead of the most frequent ones?"),
                                                             QMessageBox::StandardButtons(QMessageBox::Yes | QMessageBox::No),
                                                             QMessageBox::No);
  if (button == QMessageBox::Yes)
    fromEnd = true;
  else if (button == QMessageBox::No)
    fromEnd = false;
  else
    return;

  QString svgDescription, svgFilename;
  std::stringstream stream;
  std::string tmpStr;
  MeshData meshData;
  size_t tot = 0, len = 0, limit = std::numeric_limits<size_t>::max();
  if (limitI > 0)
    limit = limitI;

  QString svgDir = QFileDialog::getExistingDirectory(0, tr("Select a directory where storing SVG files"));
  if (svgDir.isEmpty())
    return;
  svgDir += '/';

  MT_PatchPriorityLookup<PolyMesh>::PatchContainer patches;
  MT_PatchPriorityLookup<PolyMesh>::PatchContainer::iterator pos;
  FlowMap tmpFlowMap;
  std::cout << "Sorting all templates...";
  std::cout.flush();
  if (nCorners < 0) {  
    for (n_corners_type c = 0; c <= pDBServer->maxNumberOfCorners(); ++c)
      for (count_type p = 0; p < pDBServer->numberOfCachedPatches(c); ++p)
        if (pDBServer->at(c, p).getNumberOfConcaveCorners() <= pDBlookup->getMaxConcavities()) {
          pos = std::lower_bound(patches.begin(), patches.end(), pDBServer->at(c, p));
          pos = patches.insert(pos, pDBServer->at(c, p));
          patchGeometrizator->geometrizePatch(*pos, tmpFlowMap);
        }
  } else {
    for (count_type p = 0; p < pDBServer->numberOfCachedPatches(nCorners); ++p)
      if (pDBServer->at(nCorners, p).getNumberOfConcaveCorners() <= pDBlookup->getMaxConcavities()) {
        pos = std::lower_bound(patches.begin(), patches.end(), pDBServer->at(nCorners, p));
        pos = patches.insert(pos, pDBServer->at(nCorners, p));
        patchGeometrizator->geometrizePatch(*pos, tmpFlowMap);
      }
  }
  tot = patches.size();
  stream << std::min<size_t>(tot, limit);
  len = stream.str().length();
  std::cout << "done." << tot << " patches sorted. Now store SVG:" << std::endl;

  for (size_t i = 0; i < tot && i < limit; ++i) {
    if (fromEnd) {
      // description
      svgDescription.clear();
      svgDescription += "Topology: " + QString::fromStdString(patches.at(tot-i-1).getTopologyStr()) + '\n';
      patches.at(tot-i-1).getBoundaryIDsStr(tmpStr);
      svgDescription += "BoundaryIDs: " + QString::fromStdString(tmpStr) + '\n';
      patches.at(tot-i-1).getBoundaryVarStr(tmpStr);
      svgDescription += "BoundaryVars: " + QString::fromStdString(tmpStr) + '\n';
      patches.at(tot-i-1).getBoundaryIDsEndsStr(tmpStr);
      svgDescription += "BoundaryIDsEnds: " + QString::fromStdString(tmpStr) + '\n';
      patches.at(tot-i-1).getBoundaryLenStr(tmpStr);
      svgDescription += "BoundaryLens: " + QString::fromStdString(tmpStr) + '\n';
      svgDescription += "# singularities: " + QString::number(patches.at(tot-i-1).getNumberOfSingularities()) + '\n';
      svgDescription += "Singularities: " + QString::fromStdString(patches.at(tot-i-1).getSingularityStr()) + '\n';
      svgDescription += "# corners: " + QString::number(patches.at(tot-i-1).getNumberOfCorners()) + '\n';
      svgDescription += "# faces: " + QString::number(patches.at(tot-i-1).getNumberOfFaces()) + '\n';
      svgDescription += "# vertices: " + QString::number(patches.at(tot-i-1).getNumberOfVertices()) + '\n';
      svgDescription += "# occurrences: " + QString::number(patches.at(tot-i-1).getNumberOfOccurrences()) + '\n';
      // mesh data
      patchToMeshData(patches.at(tot-i-1), meshData);
    } else {
      // description
      svgDescription.clear();
      svgDescription += "Topology: " + QString::fromStdString(patches.at(i).getTopologyStr()) + '\n';
      patches.at(i).getBoundaryIDsStr(tmpStr);
      svgDescription += "BoundaryIDs: " + QString::fromStdString(tmpStr) + '\n';
      patches.at(i).getBoundaryVarStr(tmpStr);
      svgDescription += "BoundaryVars: " + QString::fromStdString(tmpStr) + '\n';
      patches.at(i).getBoundaryIDsEndsStr(tmpStr);
      svgDescription += "BoundaryIDsEnds: " + QString::fromStdString(tmpStr) + '\n';
      patches.at(i).getBoundaryLenStr(tmpStr);
      svgDescription += "BoundaryLens: " + QString::fromStdString(tmpStr) + '\n';
      svgDescription += "# singularities: " + QString::number(patches.at(i).getNumberOfSingularities()) + '\n';
      svgDescription += "Singularities: " + QString::fromStdString(patches.at(i).getSingularityStr()) + '\n';
      svgDescription += "# corners: " + QString::number(patches.at(i).getNumberOfCorners()) + '\n';
      svgDescription += "# faces: " + QString::number(patches.at(i).getNumberOfFaces()) + '\n';
      svgDescription += "# vertices: " + QString::number(patches.at(i).getNumberOfVertices()) + '\n';
      svgDescription += "# occurrences: " + QString::number(patches.at(i).getNumberOfOccurrences()) + '\n';
      // mesh data
      patchToMeshData(patches.at(i), meshData);
    }
    // svg filename
    tmpStr.clear();
    stream.str(tmpStr);
    stream << std::setw(len) << std::setfill('0') << std::dec << std::noshowbase;
    stream << i;
    stream << ".svg";
    svgFilename = svgDir + stream.str().c_str();
    // store svg
    patchToSVG(meshData, svgFilename, svgDescription);
    std::cout << "Store file: " << stream.str() << std::endl;
    // store log
    QString svgLogFilename = svgFilename;
    int lastDotPos = svgFilename.lastIndexOf(QChar('.'));
    svgLogFilename.replace(lastDotPos, svgFilename.length() - lastDotPos, QString("_log.txt"));
    QFile logFile(svgLogFilename);
    if (!logFile.open(QFile::WriteOnly | QFile::Text | QFile::Truncate)) {
      std::cerr << "Error. Enable to open file '" << logFile.fileName().toStdString() << "'." << std::endl;
      return;
    }
    QTextStream qFileStream(&logFile);
    qFileStream << svgFilename.right(svgFilename.length() - svgFilename.lastIndexOf(QChar('/')) - 1) << ':' << endl;
    qFileStream << svgDescription << endl;
    logFile.close();
  }
  std::cout << "All SVG stored." << std::endl;
}

/**
 * @brief PatchRenderer::showProvenance
 * @param show
 */
void PatchRenderer::showProvenance(bool show) {
  if (show)
    drawLinesCheckBox->setChecked(false);
  renderData->showProvMesh = show;
  emit activateTrackball(show);
  emit changed();
}

/**
 * @brief PatchRenderer::nextProvenance
 */
void PatchRenderer::nextProvenance() {
  std::deque<PosType> corners;
  FaceSet faces;
  if (!renderData->provMesh.mesh.IsEmpty()) {
    PolygonDBInterface::PolygonCornersFromString(renderData->provMesh.mesh, corners, cornersProvenance);
    MeshSupport<PolyMesh>::MarkSubmeshFromBoundary(renderData->provMesh.mesh, corners, faces);
    for (FaceSetIterator faceIt = faces.begin(); faceIt != faces.end(); ++faceIt)
      renderData->provMesh.mesh.face[*faceIt].C() = LIGHTGRAY;
  }
  std::string meshFilename, polygonStr;
  if (polygonDB.isInitialized()) {
    bool found = polygonDB.getNextRow(meshFilename, polygonStr);
    if (found) {
      if (meshFilename != meshProvenance) {
        cornersProvenance.clear();
        meshProvenance = meshFilename;
        int err = MeshSupport<PolyMesh>::OpenMesh(renderData->provMesh.mesh, meshProvenance.c_str());
        if (err != 0) {
          QString newMeshFilename = QFileDialog::getOpenFileName(0, tr("Mesh ") + meshProvenance.c_str() + tr(" not found. Select new path"),
                                                                 QString(), tr("Mesh (*.ply *.off *.obj *.vmi)"));
          err = MeshSupport<PolyMesh>::OpenMesh(renderData->provMesh.mesh, newMeshFilename.toStdString().c_str());
        }
        if (err != 0)
          renderData->provMesh.mesh.Clear();
        vcg::tri::UpdateColor<PolyMesh>::PerFaceConstant(renderData->provMesh.mesh, LIGHTGRAY);
        Coloring::SetVertexColorByValence(renderData->provMesh.mesh);
        renderData->provMesh.vertexRadius = Coloring::ComputeVertexRadius(renderData->provMesh.mesh);
      }
      if (polygonStr != cornersProvenance) {
        cornersProvenance = polygonStr;
        PolygonDBInterface::PolygonCornersFromString(renderData->provMesh.mesh, corners, cornersProvenance);
        MeshSupport<PolyMesh>::MarkSubmeshFromBoundary(renderData->provMesh.mesh, corners, faces);
        for (FaceSetIterator faceIt = faces.begin(); faceIt != faces.end(); ++faceIt)
          renderData->provMesh.mesh.face[*faceIt].C() = DODGERBLUE;
        // colorize the boundary
        renderData->provMesh.polyLines.clear();
        vcg::face::Pos<PolyFace> runPos = corners.front();
        std::list<CoordType> l;
        n_corners_type range = corners.size();
        n_corners_type side = 0;
        vcg::Color4b c = vcg::Color4b::Scatter(range, side);
        do {
          l.push_back(runPos.V()->P());
          runPos.FlipV();
          runPos.FlipE();
          if (!runPos.IsBorder() && vcg::tri::IsMarked(renderData->provMesh.mesh, runPos.FFlip())) {
            runPos.FlipF();
            runPos.FlipE();
            if (!runPos.IsBorder() && vcg::tri::IsMarked(renderData->provMesh.mesh, runPos.FFlip())) {   // concave corner
              while (!runPos.IsBorder() && vcg::tri::IsMarked(renderData->provMesh.mesh, runPos.FFlip())) {
                runPos.FlipF();
                runPos.FlipE();
              }
              l.push_back(runPos.V()->P());
              c = vcg::Color4b::Scatter(range, side);
              renderData->provMesh.polyLines.push_back(PolyLine(c, l));
              l.clear();
              ++side;
            } else {
              if (runPos == corners.front()) {
                l.push_back(runPos.V()->P());
              }
            }
          } else {    // convex corner
            l.push_back(runPos.V()->P());
            c = vcg::Color4b::Scatter(range, side);
            renderData->provMesh.polyLines.push_back(PolyLine(c, l));
            l.clear();
            ++side;
          }
        } while (runPos != corners.front());
        if (nCorners == 0)
          renderData->provMesh.polyLines.push_back(PolyLine(c, l));
      }
    } else {
      renderData->provMesh.mesh.Clear();
      renderData->provMesh.polyLines.clear();
      meshProvenance.clear();
      polygonDB.querySelect(topologyProvenance);
    }
  } else {
    renderData->provMesh.mesh.Clear();
    renderData->provMesh.polyLines.clear();
  }
  emit changed();
}

/**
 * @brief PatchRenderer::selectSortCriterion
 * @param c
 */
void PatchRenderer::selectSortCriterion(int c) {
  std::list<CompareCriteria> criteria;
  switch (c) {
  case 1:
    criteria.push_back(CompareCriteria::FLOW_DISTANCE);
    criteria.push_back(CompareCriteria::SINGULARITYNUM);
    criteria.push_back(CompareCriteria::SINGULARITYTYPES);
    criteria.push_back(CompareCriteria::QUADS_QUALITY);
    break;
  case 2:
    criteria.push_back(CompareCriteria::QUADS_QUALITY);
    criteria.push_back(CompareCriteria::SINGULARITYNUM);
    criteria.push_back(CompareCriteria::SINGULARITYTYPES);
    criteria.push_back(CompareCriteria::FLOW_DISTANCE);
    break;
  case 3:
    criteria.push_back(CompareCriteria::SINGULARITYNUM);
    criteria.push_back(CompareCriteria::SINGULARITYTYPES);
    criteria.push_back(CompareCriteria::FLOW_DISTANCE);
    criteria.push_back(CompareCriteria::QUADS_QUALITY);
    break;
  default:
    break;
  }
  compareCriteria = criteria.front();
  pDBlookup->setSortCriteria(std::move(criteria));
}
