#ifndef PATCHRENDERER_H
#define PATCHRENDERER_H

#include <SCIP/scip_solver.h>      // because some structs in <SCIP/scip/struct_misc.h> have a member named 'slots' which conflicts with Qt
#include <memory>
#include <QList>
#include <QVector>
#include <QScrollArea>
#include <QGridLayout>
#include <QCheckBox>
#include <QSpinBox>
#include <QComboBox>
#include <QTextEdit>
#include <QFile>
#include "spinbox.h"
#include "render_data.h"
using namespace vcg::tri::pl;

class IntValue : public QObject {
  Q_OBJECT
public:
  IntValue(const int init) : value(init) { }

  inline int getValue() const {
    return value;
  }

public slots:
  inline void setValue(const int val) {
    value = val;
  }

private:
  int value;
};


class PatchRenderer : public QScrollArea
#ifndef SINGLE_THREAD
                    , public MT_PatchPriorityLookupNotifiable
#endif
                    , public std::enable_shared_from_this<PatchRenderer> {

  Q_OBJECT

public:
  explicit PatchRenderer(RenderData *meshData,
#ifdef SINGLE_THREAD
                         PatchPriorityLookup<PolyMesh>::PatchDBServerPointer pServer,
                         PatchPriorityLookup<PolyMesh> *db,
#else
                         MT_PatchPriorityLookup<PolyMesh>::PatchDBServerPointer pServer,
                         MT_PatchPriorityLookup<PolyMesh> *db,
#endif
                         QWidget *parent = 0);

#ifndef SINGLE_THREAD
  void notify(NotificationType notificationType,
              count_type nTemplatePatches, count_type nPatches,
              std::chrono::nanoseconds duration);

  void registerForNotifications();
#endif

  ~PatchRenderer();

signals:
  void changed();
  void closed();
  void log(QString string);
  void setMin(int min);
  void setMax(int max);
  void setVal(int val);
  void patchesChangedAt(int index);
  void enable(bool en);
  void executing(bool ex);
  void toStatus(const QString &string);
  void drawLines(bool draw);
  void activateTrackball(bool a);

public slots:
  void reset();
  void setPolygonDBFilename(const std::string &filename);

protected:
  void closeEvent(QCloseEvent *e);

private slots:
  void changeNCorners(int n);
  void changeRenderBoundary();
  void setMaxSingularities(int maxS);
  void setMinValence(int minV);
  void setMaxValence(int maxV);
  void setMaxConcavities(int maxC);
  void setQualityThreshold(double th);
  void setILPthreshold(int th);
  void setTolerance(int tolerance);
  void findPatches();
  void stopSearching();
  void selectPatch(int i);
  void currentPatchToSVG();
  void allPatchesToSVG();
  void showQuality(bool show);
  void smoothWithFlowConstraints(bool smooth);
  void drawLinesForSorting(bool draw);
  void setLaplacianWeight(double weight);
  void deleteDrawnLines();
  void sortPatches();
  void showProvenance(bool show);
  void nextProvenance();
  void selectSortCriterion(int c);

private:
  typedef PolychordSupport<PolyMesh>                       PolychordSupport;
  typedef PolychordSupport::ScalarType                     ScalarType;
  typedef PolychordSupport::ScalarContainer                ScalarContainer;
  typedef PolychordSupport::PointType                      PointType;
  typedef PolychordSupport::PointContainer                 PointContainer;
  typedef PolychordSupport::PointConstIterator             PointConstIterator;
  typedef PolychordSupport::FlowPoints                     FlowPoints;
  typedef PolychordSupport::FlowContainer                  FlowContainer;
  typedef PolychordSupport::FlowContainerConstIterator     FlowContainerConstIterator;
  typedef PolychordSupport::FlowMap                        FlowMap;
  typedef PolychordSupport::FlowMapIterator                FlowMapIterator;
  typedef PolychordSupport::FlowMapConstIterator           FlowMapConstIterator;
  typedef PolychordSupport::PolychordPoints                PolychordPoints;
  typedef PolychordSupport::PolychordContainer             PolychordContainer;
  typedef PolychordSupport::PolychordMap                   PolychordMap;
  typedef PolychordSupport::PolychordMapIterator           PolychordMapIterator;
  typedef PolychordSupport::PolychordMapConstIterator      PolychordMapConstIterator;

  void patchToMeshData(const Patch<PolyMesh> &patch, MeshData &meshData, const bool qualityColor = false);
  void printBoundaryIDs(const vcg::face::Pos<PolyFace> &startPos, const std::string &bEq, const std::vector<num_type> &bVars,
                        std::list< std::pair<var_label_type, CoordType> > &labels);
  void patchToSVG(MeshData &meshData, const QString &svgFilename, const QString &description,
                  const bool showFlows = false, const bool showPolychords = false, const bool showSamples = false);

  // GUI elements
  QGridLayout                                    *BCgridLayout;
  QSpinBox                                       *nCornersSpinBox;
  QVector<IntValue *>                             nEdgesPerSide;
  QGridLayout                                    *pvGridLayout;
  QCheckBox                                      *drawLinesCheckBox;
  QCheckBox                                      *showProvenanceCheckBox;
  QTextEdit                                      *patchInfoTextEdit;
  static QString                                  logStr;
  QString                                         description;
  QString                                         baseDescription;
  QString                                         msg;
  bool                                            showQualityColor;

  // data objects
  RenderData                                     *renderData;
#ifdef SINGLE_THREAD
  typedef PatchPriorityLookup<PolyMesh>::PatchDBServerPointer       PatchDBServerPointer;
  typedef PatchPriorityLookup<PolyMesh>::CompareCriteria            CompareCriteria;
  PatchDBServerPointer                            pDBServer;
  PatchPriorityLookup<PolyMesh>                  *pDBlookup;
#else
  typedef MT_PatchPriorityLookup<PolyMesh>::PatchDBServerPointer    PatchDBServerPointer;
  typedef MT_PatchPriorityLookup<PolyMesh>::CompareCriteria         CompareCriteria;
  PatchDBServerPointer                            pDBServer;
  MT_PatchPriorityLookup<PolyMesh>               *pDBlookup;
  std::chrono::nanoseconds                        elapsedFromPreviousUIupdated;
  std::mutex                                      mutexForTime;
#endif
  std::shared_ptr<PatchFlattener<PolyMesh>>       patchGeometrizator;
  CompareCriteria                                 compareCriteria;

  // provenance
  typedef MeshSupport<PolyMesh>::PosType          PosType;
  typedef MeshSupport<PolyMesh>::FaceSet          FaceSet;
  typedef MeshSupport<PolyMesh>::FaceSetIterator  FaceSetIterator;
  PolygonDBInterface                              polygonDB;
  std::string                                     topologyProvenance;
  std::string                                     meshProvenance;
  std::string                                     cornersProvenance;

  // parameters
  n_corners_type                                  nCorners;
  int                                             currentPatch;
};

#endif // PATCHRENDERER_H
