#include <SCIP/scip_solver.h>      // because some structs in <SCIP/scip/struct_misc.h> have a member named 'slots' which conflicts with Qt
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QGridLayout>
#include <QLayoutItem>
#include <QGroupBox>
#include <QButtonGroup>
#include <QLabel>
#include <QSpinBox>
#include <QCheckBox>
#include <QRadioButton>
#include <QPushButton>
#include <QFileDialog>
#include <QInputDialog>
#include <QMessageBox>
#include <QTextStream>
#include <QFile>
#if QT_VERSION >= QT_VERSION_CHECK(5,0,0)
  #include <QGuiApplication>
#else
  #include <QApplication>
#endif
#include "spinbox.h"
#include "patchretopo.h"

QString PatchRetopo::logStr = QString();

/**
 * @brief PatchRetopo::PatchRetopo
 * @param meshData
 * @param db
 * @param parent
 */
PatchRetopo::PatchRetopo(RenderData *meshData,
#ifdef SINGLE_THREAD
                         PatchPriorityLookup<PolyMesh> *db,
#else
                         MT_PatchPriorityLookup<PolyMesh> *db,
#endif
                         QWidget *parent) : QScrollArea(parent),
                                            patchGeometrizator(new PatchRetopologizer) {
  renderData = meshData;
  pDBlookup = db;
  assert(renderData != 0 && pDBlookup != 0);
  showQualityColor = false;
  patchGeometrizator->getNumberOfCorners() = 0;
  currentPatch = 0;
  renderData->meshes.resize(9);
  renderData->currentMesh = renderData->meshes.size() / 2;

  QVBoxLayout *layout = new QVBoxLayout;

  QGroupBox *inputBox = new QGroupBox(tr("Input"));
  QVBoxLayout *inputLayout = new QVBoxLayout;
  QCheckBox *showTriCheckBox = new QCheckBox(tr("Show TriMesh"));
  showTriCheckBox->setChecked(renderData->showTriMesh);
  connect(showTriCheckBox, SIGNAL(toggled(bool)), this, SLOT(showTriMeshSlot(bool)));
  inputLayout->addWidget(showTriCheckBox);
  QPushButton *loadButton = new QPushButton(tr("Load"));
  loadButton->setMinimumWidth(55);
  loadButton->setMaximumWidth(70);
  connect(this, SIGNAL(executing(bool)), loadButton, SLOT(setDisabled(bool)));
  connect(loadButton, SIGNAL(clicked()), this, SLOT(loadXML()));
  inputLayout->addWidget(loadButton);
  inputLayout->setAlignment(loadButton, Qt::AlignHCenter);
  QPushButton *clearButton = new QPushButton(tr("Clear"));
  clearButton->setMinimumWidth(55);
  clearButton->setMaximumWidth(70);
  connect(this, SIGNAL(executing(bool)), clearButton, SLOT(setDisabled(bool)));
  connect(clearButton, SIGNAL(clicked()), this, SLOT(clearInput()));
  inputLayout->addWidget(clearButton);
  inputLayout->setAlignment(clearButton, Qt::AlignHCenter);
  inputBox->setLayout(inputLayout);
  layout->addWidget(inputBox);

  QGroupBox *searchBox = new QGroupBox(tr("Search"));
  QVBoxLayout *searchLayout = new QVBoxLayout;
  QCheckBox *showQualityCheckBox = new QCheckBox(tr("Show quad quality"));
  connect(showQualityCheckBox, SIGNAL(toggled(bool)), this, SLOT(showQuality(bool)));
  searchLayout->addWidget(showQualityCheckBox);
  QButtonGroup *sortButtonGroup = new QButtonGroup;
  connect(sortButtonGroup, SIGNAL(buttonClicked(int)), this, SLOT(selectSortCriterion(int)));
  QHBoxLayout *sortButtonsLayout2 = new QHBoxLayout;
  QRadioButton *flowSortButton = new QRadioButton(tr("Flow"));
  connect(this, SIGNAL(executing(bool)), flowSortButton, SLOT(setDisabled(bool)));
  sortButtonGroup->addButton(flowSortButton, 1);
  sortButtonsLayout2->addWidget(flowSortButton);
  QRadioButton *quadSortButton = new QRadioButton(tr("Quad"));
  connect(this, SIGNAL(executing(bool)), quadSortButton, SLOT(setDisabled(bool)));
  sortButtonGroup->addButton(quadSortButton, 2);
  sortButtonsLayout2->addWidget(quadSortButton);
  QRadioButton *singSortButton = new QRadioButton(tr("Sing"));
  connect(this, SIGNAL(executing(bool)), singSortButton, SLOT(setDisabled(bool)));
  sortButtonGroup->addButton(singSortButton, 3);
  sortButtonsLayout2->addWidget(singSortButton);
  flowSortButton->setChecked(true);
  compareCriteria = pDBlookup->getSortCriteria().front();
  searchLayout->addLayout(sortButtonsLayout2);
  QCheckBox *smoothCheckBox = new QCheckBox(tr("Smooth with flow constraints"));
  smoothCheckBox->setToolTip(tr("Check this to impose flows as geometric constraints."));
  smoothCheckBox->setChecked(patchGeometrizator->getSmoothWithConstraints());
  connect(smoothCheckBox, SIGNAL(toggled(bool)), this, SLOT(smoothWithFlowConstraints(bool)));
  connect(this, SIGNAL(executing(bool)), smoothCheckBox, SLOT(setDisabled(bool)));
  searchLayout->addWidget(smoothCheckBox);
  QGridLayout *parametersLayout = new QGridLayout;
  QLabel *maxSingularitiesLabel = new QLabel(tr("Max # sing.:"));
  parametersLayout->addWidget(maxSingularitiesLabel, 0, 0, Qt::AlignRight);
  QSpinBox *maxSingularitiesSpinBox = new QSpinBox;
  maxSingularitiesSpinBox->setMinimum(0);
  maxSingularitiesSpinBox->setMaximum(std::numeric_limits<int>::max());
  maxSingularitiesSpinBox->setValue(std::numeric_limits<int>::max());
  connect(maxSingularitiesSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setMaxSingularities(int)));
  parametersLayout->addWidget(maxSingularitiesSpinBox, 0, 1);
  QLabel *minValenceLabel = new QLabel(tr("Min valence:"));
  parametersLayout->addWidget(minValenceLabel, 1, 0, Qt::AlignRight);
  UpdatableMinMaxSpinBox *minValenceSpinBox = new UpdatableMinMaxSpinBox;
  minValenceSpinBox->setMinimum(2);
  minValenceSpinBox->setMaximum(std::numeric_limits<valence_type>::max());
  minValenceSpinBox->setValue(minValenceSpinBox->minimum());
  connect(minValenceSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setMinValence(int)));
  parametersLayout->addWidget(minValenceSpinBox, 1, 1);
  QLabel *maxValenceLabel = new QLabel(tr("Max valence:"));
  parametersLayout->addWidget(maxValenceLabel, 2, 0, Qt::AlignRight);
  UpdatableMinMaxSpinBox *maxValenceSpinBox = new UpdatableMinMaxSpinBox;
  maxValenceSpinBox->setMinimum(2);
  maxValenceSpinBox->setMaximum(std::numeric_limits<valence_type>::max());
  maxValenceSpinBox->setValue(maxValenceSpinBox->maximum());
  connect(maxValenceSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setMaxValence(int)));
  parametersLayout->addWidget(maxValenceSpinBox, 2, 1);
  connect(minValenceSpinBox, SIGNAL(valueChanged(int)), maxValenceSpinBox, SLOT(minimumChanged(int)));
  connect(maxValenceSpinBox, SIGNAL(valueChanged(int)), minValenceSpinBox, SLOT(maximumChanged(int)));
  QLabel *maxConcaveLabel = new QLabel(tr("Max # concav.:"));
  parametersLayout->addWidget(maxConcaveLabel, 3, 0, Qt::AlignRight);
  QSpinBox *maxConcaveSpinBox = new QSpinBox;
  maxConcaveSpinBox->setMinimum(0);
  maxConcaveSpinBox->setMaximum(std::numeric_limits<n_corners_type>::max());
  maxConcaveSpinBox->setValue(std::numeric_limits<n_corners_type>::max());
  connect(maxConcaveSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setMaxConcavities(int)));
  parametersLayout->addWidget(maxConcaveSpinBox, 3, 1);
  QLabel *qualityThresholdLabel = new QLabel(tr("Quality th.:"));
  parametersLayout->addWidget(qualityThresholdLabel, 4, 0, Qt::AlignRight);
  QDoubleSpinBox *qualityThresholdSpinBox = new QDoubleSpinBox;
  qualityThresholdSpinBox->setMinimum(0);
  qualityThresholdSpinBox->setMaximum(1);
  qualityThresholdSpinBox->setSingleStep(0.01);
  qualityThresholdSpinBox->setValue(qualityThresholdSpinBox->maximum());
  connect(qualityThresholdSpinBox, SIGNAL(valueChanged(double)), this, SLOT(setQualityThreshold(double)));
  parametersLayout->addWidget(qualityThresholdSpinBox, 4, 1);
  QLabel *laplacianLabel = new QLabel(tr("Laplacian weight:"));
  parametersLayout->addWidget(laplacianLabel, 5, 0, Qt::AlignRight);
  QDoubleSpinBox *laplacianSpinBox = new QDoubleSpinBox;
  laplacianSpinBox->setMinimum(0.01);
  laplacianSpinBox->setMaximum(0.99);
  laplacianSpinBox->setValue(patchGeometrizator->getLaplacianWeight());
  laplacianSpinBox->setDecimals(2);
  laplacianSpinBox->setSingleStep(0.01);
  connect(laplacianSpinBox, SIGNAL(valueChanged(double)), this, SLOT(setLaplacianWeight(double)));
  parametersLayout->addWidget(laplacianSpinBox, 5, 1);
  QLabel *ilp_thresholdLabel = new QLabel(tr("ILP th.:"));
  parametersLayout->addWidget(ilp_thresholdLabel, 6, 0, Qt::AlignRight);
  QSpinBox *ilp_thresholdSpinBox = new QSpinBox;
  ilp_thresholdSpinBox->setToolTip(tr("It's the min side length over which flow constraints are added to the ILP."));
  ilp_thresholdSpinBox->setMinimum(0);
  ilp_thresholdSpinBox->setMaximum(std::numeric_limits<int>::max());
  ilp_thresholdSpinBox->setValue(pDBlookup->getMinSizeThresholdILP());
  connect(ilp_thresholdSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setILPthreshold(int)));
  parametersLayout->addWidget(ilp_thresholdSpinBox, 6, 1);
  QLabel *ilp_toleranceLabel = new QLabel(tr("Tolerance:"));
  parametersLayout->addWidget(ilp_toleranceLabel, 7, 0, Qt::AlignRight);
  QSpinBox *ilp_toleranceSpinBox = new QSpinBox;
  ilp_toleranceSpinBox->setToolTip(tr("It's the tolerance on edge alignment of flow constraints which are added to the ILP."));
  ilp_toleranceSpinBox->setMinimum(0);
  ilp_toleranceSpinBox->setMaximum(std::numeric_limits<u_small_int_type>::max());
  ilp_toleranceSpinBox->setValue(pDBlookup->getFlowConstraintsILPtolerance());
  connect(ilp_toleranceSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setTolerance(int)));
  parametersLayout->addWidget(ilp_toleranceSpinBox, 7, 1);
  searchLayout->addLayout(parametersLayout);
  QPushButton *findButton = new QPushButton(tr("Find"));
  findButton->setMinimumWidth(55);
  findButton->setMaximumWidth(70);
  connect(findButton, SIGNAL(clicked()), this, SLOT(findPatches()));
  connect(this, SIGNAL(executing(bool)), findButton, SLOT(setDisabled(bool)));
  searchLayout->addWidget(findButton);
  searchLayout->setAlignment(findButton, Qt::AlignHCenter);
  QPushButton *stopButton = new QPushButton(tr("Stop"));
  stopButton->setDisabled(true);
  stopButton->setMinimumWidth(55);
  stopButton->setMaximumWidth(70);
  connect(stopButton, SIGNAL(clicked()), this, SLOT(stopSearching()));
  connect(this, SIGNAL(executing(bool)), stopButton, SLOT(setEnabled(bool)));
  searchLayout->addWidget(stopButton);
  searchLayout->setAlignment(stopButton, Qt::AlignHCenter);
  searchBox->setLayout(searchLayout);
  layout->addWidget(searchBox);

  QGroupBox *selectPatchBox = new QGroupBox(tr("Select Patch"));
  connect(this, SIGNAL(enable(bool)), selectPatchBox, SLOT(setEnabled(bool)));
  QVBoxLayout *selectPatchLayout = new QVBoxLayout;
  UpdatableMinMaxSpinBox *selectPatchSpinBox = new UpdatableMinMaxSpinBox;
  selectPatchSpinBox->setMinimum(0);
  selectPatchSpinBox->setMaximum(0);
  connect(this, SIGNAL(setMin(int)), selectPatchSpinBox, SLOT(minimumChanged(int)));
  connect(this, SIGNAL(setMax(int)), selectPatchSpinBox, SLOT(maximumChanged(int)));
  connect(this, SIGNAL(setVal(int)), selectPatchSpinBox, SLOT(setValue(int)));
  connect(selectPatchSpinBox, SIGNAL(valueChanged(int)), this, SLOT(selectPatch(int)));
  connect(this, SIGNAL(patchesChangedAt(int)), this, SLOT(selectPatch(int)));
  selectPatchLayout->addWidget(selectPatchSpinBox);

  patchInfoTextEdit = new QTextEdit(tr("No patches."));
  patchInfoTextEdit->setMinimumHeight(100);
  QFont font = patchInfoTextEdit->currentFont();
  font.setPointSize(10);
  patchInfoTextEdit->setFont(font);
  patchInfoTextEdit->setReadOnly(true);
  selectPatchLayout->addWidget(patchInfoTextEdit);
  selectPatchBox->setLayout(selectPatchLayout);
  selectPatchBox->setDisabled(true);
  layout->addWidget(selectPatchBox);

  QWidget *widget = new QWidget;
  widget->setLayout(layout);
  layout->setSizeConstraint(QLayout::SetMinAndMaxSize);
  setWidget(widget);
  setWidgetResizable(true);
  widget->setMinimumSize(50, 200);
  setMinimumSize(50, 200);
}

#ifndef SINGLE_THREAD
/**
 * @brief PatchRetopo::notify
 * @param notificationType
 * @param nTemplatePatches
 * @param nPatches
 * @param duration
 */
void PatchRetopo::notify(NotificationType notificationType,
                         count_type nTemplatePatches, count_type nPatches,
                         std::chrono::nanoseconds duration) {
  std::string durationStr;
  TimeElements timeElements;
  TimeSupport::SplitTimeElements(duration, timeElements);
  TimeSupport::TimeElementsToString(timeElements, durationStr);
  bool emitSignal = false;
  msg = "Elapsed: ";
  msg += durationStr.c_str();
  msg += " Rotated templates: " + QString::number(nTemplatePatches) + '.';

  mutexForTime.lock();
  if (duration - elapsedFromPreviousUIupdated >= std::chrono::milliseconds(100))
    emitSignal = true;
  mutexForTime.unlock();

  switch (notificationType) {
  case STRUCTURE_FILTER:
    msg += " Filtering by structure";
    break;
  case FLOW_FILTER:
    msg += " Filtering by flows";
    break;
  case FLUSHING:
    msg += " Flushing the buffer: " + QString::number(nPatches) + " previously computed patches to delete";
    break;
  case ILP_FILTER:
    msg += " Expanded: " + QString::number(nPatches) + ". Filtering by ILP";
    if (nPatches > 0) {
      mutexForTime.lock();
      if (emitSignal) {
        elapsedFromPreviousUIupdated = duration;
        emit setMax(nPatches);
        emit patchesChangedAt(currentPatch + 1);
        emit setMin(1);
        emit enable(true);
      }
      mutexForTime.unlock();
    }
    break;
  case SORT:
    msg += " Sorted " + QString::number(nPatches) + "patches. Sorting";
    if (nPatches > 0) {
      mutexForTime.lock();
      if (emitSignal) {
        elapsedFromPreviousUIupdated = duration;
        emit setMax(nPatches);
        emit patchesChangedAt(currentPatch + 1);
        emit setMin(1);
        emit enable(true);
      }
      mutexForTime.unlock();
    }
    break;
  case COMPLETE:
    msg += " Found: " + QString::number(nPatches) + ". Complete!";
    emitSignal = true;
    if (nPatches > 0) {
      if (currentPatch < 0)
        currentPatch = 0;
      emit setMax(nPatches);
      emit patchesChangedAt(currentPatch + 1);
      emit setMin(1);
      emit enable(true);
    }
    emit executing(false);
    break;
  default:
    break;
  }

  if (notificationType != EMPTY && notificationType != COMPLETE)
    for (int i = 0; i <= timeElements.s.count() % 3; ++i)
      msg += '.';
  if (emitSignal)
    emit toStatus(msg);
}
#endif

/**
 * @brief PatchRetopo::~PatchRetopo
 */
PatchRetopo::~PatchRetopo() {
#ifndef SINGLE_THREAD
  std::shared_ptr<MT_PatchPriorityLookupNotifiable> shared_this =
      std::dynamic_pointer_cast<MT_PatchPriorityLookupNotifiable>(shared_from_this());
  pDBlookup->deregisterNotifiable(shared_this);
#endif
}

#ifndef SINGLE_THREAD
/**
 * @brief PatchRetopo::registerForNotifications
 */
void PatchRetopo::registerForNotifications() {
  std::shared_ptr<MT_PatchPriorityLookupNotifiable> shared_this =
      std::dynamic_pointer_cast<MT_PatchPriorityLookupNotifiable>(shared_from_this());
  pDBlookup->registerNotifiable(shared_this);
}
#endif

/**
 * @brief PatchRetopo::reset
 */
void PatchRetopo::reset() {
  pDBlookup->clear();
  renderData->baseMesh.clear();
  for (size_t i = 0; i < renderData->meshes.size(); ++i)
    renderData->meshes[i].clear();
  renderData->currentMesh = renderData->meshes.size() / 2;
  renderData->boundary.sides.clear();
  renderData->boundary.colors.clear();
  renderData->flows.clear();
  renderData->flowsColors.clear();
  renderData->nFlows = 0;
  description.clear();
  baseDescription.clear();
  logStr.clear();
  msg.clear();
  patchInfoTextEdit->setText(logStr);
  emit toStatus(msg);
}

/**
 * @brief PatchRetopo::closeEvent
 * @param e
 */
void PatchRetopo::closeEvent(QCloseEvent *e) {
  QScrollArea::closeEvent(e);
  emit closed();
}

/**
 * @brief PatchRetopo::showTriMeshSlot
 * @param show
 */
void PatchRetopo::showTriMeshSlot(bool show) {
  renderData->showTriMesh = show;
  emit changed();
}

void PatchRetopo::loadXML() {
  // select a mesh file
  QString filename = QFileDialog::getOpenFileName(0, tr("Open Mesh Data"), QString(),
                                                  tr("Mesh data (*.xml)"));

  if(!filename.isEmpty()) {
    QFile xmlFile(filename);
    if (!xmlFile.open(QFile::ReadOnly | QFile::Text)) {
      QMessageBox::warning(this, tr("Mesh Data XML"),
                                 tr("Cannot read file %1:\n%2.")
                                   .arg(filename)
                                   .arg(xmlFile.errorString()));
      return;
    }
    QXmlStreamReader xmlReader(&xmlFile);
    if (!xmlReader.readNextStartElement() || xmlReader.name() != "cereal") {
      std::cout << "Error. Non cereal xml." << std::endl;
      return;
    }
    clearInput();
    readXML(xmlReader);
    emit changed();
  }
}

/**
 * @brief PatchRetopo::readXML
 * @param xmlReader
 */
void PatchRetopo::readXML(QXmlStreamReader &xmlReader) {
  Q_ASSERT(xmlReader.isStartElement() && xmlReader.name() == "cereal");

  bool numOK = false;
  ScalarType num;
  size_t ind, ind2, fInd;
  num_type nEdges;
  PointContainer vertices;
  PointContainer verticesUV;
  std::deque< std::vector<size_t> > faces;
  std::deque<num_type> boundaryLen;
  PointContainer bVertices;
  PointContainer bVerticesUV;
  struct stroke {
    bool isLoop;
    PointContainer points;
    PointContainer pointsUV;
  };
  std::deque<stroke> strokes;
  QString name;
  int valueSize = QString("value").length();
  while (xmlReader.readNextStartElement()) {
    if (xmlReader.name() == "basemesh_vertices") {
      while (xmlReader.readNextStartElement()) {
        name = xmlReader.name().toString();
        ind = name.right(name.length() - valueSize).toULongLong(&numOK);
        assert(numOK);
        if (ind >= vertices.size())
          vertices.resize(ind + 1, PointType(0, 0, 0));
        if (ind >= verticesUV.size())
          verticesUV.resize(ind + 1, PointType(0, 0, 0));
        while (xmlReader.readNextStartElement()) {
          if (xmlReader.name() == "x") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            vertices[ind][0] = num;
          } else if (xmlReader.name() == "y") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            vertices[ind][1] = num;
          } else if (xmlReader.name() == "z") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            vertices[ind][2] = num;
          } else if (xmlReader.name() == "u") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            verticesUV[ind][0] = num;
          } else if (xmlReader.name() == "v") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            verticesUV[ind][1] = num;
          } else
            xmlReader.skipCurrentElement();
        }
      }
    } else if (xmlReader.name() == "basemesh_faces") {
      while (xmlReader.readNextStartElement()) {
        name = xmlReader.name().toString();
        ind = name.right(name.length() - valueSize).toULongLong(&numOK);
        assert(numOK);
        if (ind >= faces.size())
          faces.resize(ind + 1, std::vector<size_t>(3, 0));
        while (xmlReader.readNextStartElement()) {
          name = xmlReader.name().toString();
          ind2 = name.right(name.length() - valueSize).toULongLong(&numOK);
          assert(numOK);
          assert(ind2 < 3);
          fInd = xmlReader.readElementText().toULongLong(&numOK);
          assert(numOK);
          faces[ind][ind2] = fInd;
        }
      }
    } else if (xmlReader.name() == "num_subdiv") {
      while (xmlReader.readNextStartElement()) {
        name = xmlReader.name().toString();
        ind = name.right(name.length() - valueSize).toULongLong(&numOK);
        assert(numOK);
        if (ind >= boundaryLen.size())
          boundaryLen.resize(ind + 1, 0);
        nEdges = xmlReader.readElementText().toULongLong(&numOK);
        assert(numOK);
        boundaryLen[ind] = nEdges;
      }
    } else if (xmlReader.name() == "boundary_vertices") {
      while (xmlReader.readNextStartElement()) {
        name = xmlReader.name().toString();
        ind = name.right(name.length() - valueSize).toULongLong(&numOK);
        assert(numOK);
        if (ind >= bVertices.size())
          bVertices.resize(ind + 1, PointType(0, 0, 0));
        if (ind >= bVerticesUV.size())
          bVerticesUV.resize(ind + 1, PointType(0, 0, 0));
        while (xmlReader.readNextStartElement()) {
          if (xmlReader.name() == "x") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            bVertices[ind][0] = num;
          } else if (xmlReader.name() == "y") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            bVertices[ind][1] = num;
          } else if (xmlReader.name() == "z") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            bVertices[ind][2] = num;
          } else if (xmlReader.name() == "u") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            bVerticesUV[ind][0] = num;
          } else if (xmlReader.name() == "v") {
            num = xmlReader.readElementText().toDouble(&numOK);
            assert(numOK);
            bVerticesUV[ind][1] = num;
          } else
            xmlReader.skipCurrentElement();
        }
      }
    } else if (xmlReader.name() == "strokes") {
      while (xmlReader.readNextStartElement()) {
        name = xmlReader.name().toString();
        ind = name.right(name.length() - valueSize).toULongLong(&numOK);
        assert(numOK);
        if (ind >= strokes.size())
          strokes.resize(ind + 1);
        while (xmlReader.readNextStartElement()) {
          if (xmlReader.name() == "first") {
            if (xmlReader.readElementText().toLower() == "true")
              strokes[ind].isLoop = true;
            else
              strokes[ind].isLoop = false;
          } else if (xmlReader.name() == "second") {
            while (xmlReader.readNextStartElement()) {
              name = xmlReader.name().toString();
              ind2 = name.right(name.length() - valueSize).toULongLong(&numOK);
              assert(numOK);
              if (ind2 >= strokes[ind].points.size())
                strokes[ind].points.resize(ind2 + 1, PointType(0, 0, 0));
              if (ind2 >= strokes[ind].pointsUV.size())
                strokes[ind].pointsUV.resize(ind2 + 1, PointType(0, 0, 0));
              while (xmlReader.readNextStartElement()) {
                if (xmlReader.name() == "x") {
                  num = xmlReader.readElementText().toDouble(&numOK);
                  assert(numOK);
                  strokes[ind].points[ind2][0] = num;
                } else if (xmlReader.name() == "y") {
                  num = xmlReader.readElementText().toDouble(&numOK);
                  assert(numOK);
                  strokes[ind].points[ind2][1] = num;
                } else if (xmlReader.name() == "z") {
                  num = xmlReader.readElementText().toDouble(&numOK);
                  assert(numOK);
                  strokes[ind].points[ind2][2] = num;
                } else if (xmlReader.name() == "u") {
                  num = xmlReader.readElementText().toDouble(&numOK);
                  assert(numOK);
                  strokes[ind].pointsUV[ind2][0] = num;
                } else if (xmlReader.name() == "v") {
                  num = xmlReader.readElementText().toDouble(&numOK);
                  assert(numOK);
                  strokes[ind].pointsUV[ind2][1] = num;
                } else
                  xmlReader.skipCurrentElement();
              }
            }
          } else
            xmlReader.skipCurrentElement();
        }
      }
    } else {
      xmlReader.skipCurrentElement();
    }
  }
  assert(vertices.size() == verticesUV.size());
  assert(bVertices.size() == bVerticesUV.size());
  TMesh &tMesh = patchGeometrizator->getTriMesh();
  VertexIterator vIt = vcg::tri::Allocator<TMesh>::AddVertices(tMesh, vertices.size());
  for (ind = 0; ind < vertices.size(); ++ind, ++vIt) {
    vIt->P() = vertices[ind];
    vIt->T().U() = verticesUV[ind][0];
    vIt->T().V() = verticesUV[ind][1];
  }
  FaceIterator fIt = vcg::tri::Allocator<TMesh>::AddFaces(tMesh, faces.size());
  for (ind = 0; ind < faces.size(); ++ind, ++fIt)
    for (ind2 = 0; ind2 < 3; ++ind2)
      fIt->V(ind2) = &tMesh.vert[faces[ind][ind2]];
  vcg::tri::UpdateTopology<TMesh>::FaceFace(tMesh);
  vcg::tri::UpdateNormal<TMesh>::PerFaceNormalized(tMesh);
  vcg::tri::UpdateNormal<TMesh>::PerVertexFromCurrentFaceNormal(tMesh);
  vcg::tri::UpdateNormal<TMesh>::NormalizePerVertex(tMesh);
  vcg::tri::UpdateBounding<TMesh>::Box(tMesh);
  vcg::tri::UpdateColor<TMesh>::PerFaceConstant(tMesh, SAND);
  patchGeometrizator->getNumberOfCorners() = boundaryLen.size();
  patchGeometrizator->getBoundaryLen().resize(boundaryLen.size(), 0);
  for (size_t i = 0; i < boundaryLen.size(); ++i)
    patchGeometrizator->getBoundaryLen()[i] = boundaryLen[i];
  patchGeometrizator->getBoundary().resize(patchGeometrizator->getBoundaryLen().size());
  patchGeometrizator->getBoundaryUV().resize(patchGeometrizator->getBoundaryLen().size());
  PointContainerConstIterator bvIt = bVertices.begin();
  PointContainerConstIterator bvUVit = bVerticesUV.begin();
  for (size_t s = 0; s <patchGeometrizator->getBoundaryLen().size(); ++s) {
    patchGeometrizator->getBoundary()[s].assign(bvIt, bvIt + patchGeometrizator->getBoundaryLen()[s]);
    patchGeometrizator->getBoundaryUV()[s].assign(bvUVit, bvUVit + patchGeometrizator->getBoundaryLen()[s]);
    bvIt += patchGeometrizator->getBoundaryLen()[s];
    bvUVit += patchGeometrizator->getBoundaryLen()[s];
  }
  for (size_t s = 0; s < patchGeometrizator->getBoundaryLen().size(); ++s) {
    patchGeometrizator->getBoundary()[s].push_back(patchGeometrizator->getBoundary()[(s+1)%patchGeometrizator->getBoundary().size()].front());
    patchGeometrizator->getBoundaryUV()[s].push_back(patchGeometrizator->getBoundaryUV()[(s+1)%patchGeometrizator->getBoundaryUV().size()].front());
  }
  for (size_t f = 0; f < strokes.size(); ++f) {
    PatchRetopologizer::PointsToFlow(strokes[f].points, strokes[f].isLoop, patchGeometrizator->getBoundary(),
                                     patchGeometrizator->getFlowMap(), NSAMPLES);
    PatchRetopologizer::PointsToFlow(strokes[f].pointsUV, strokes[f].isLoop, patchGeometrizator->getBoundaryUV(),
                                     patchGeometrizator->getFlowMapUV(), NSAMPLES);
  }
  patchGeometrizator->getTriMesh(renderData->tMesh);
  renderData->vertexRadius = Coloring::ComputeVertexRadius(tMesh);
  patchGeometrizator->setRadius(tMesh.bbox.Diag() / 2);
  patchGeometrizator->getBoundary(renderData->boundary.sides);
  renderData->boundary.colors.resize(renderData->boundary.sides.size());
  for (size_t side = 0; side < renderData->boundary.colors.size(); ++side)
    renderData->boundary.colors[side] = vcg::Color4b::Scatter(renderData->boundary.colors.size(), side);
  patchGeometrizator->getFlowMap(renderData->flows);
  std::deque<vcg::Color4b> flowColors;
  for (FlowMapIterator flowIt = renderData->flows.begin(); flowIt != renderData->flows.end(); ++flowIt) {
    flowColors.resize(flowIt->second.size());
    for (size_t f = 0; f < flowIt->second.size(); ++f) {
      flowColors[f] = vcg::Color4b::Scatter(renderData->boundary.sides.size() + 20,
                                            renderData->boundary.sides.size() + renderData->nFlows);
      ++renderData->nFlows;
    }
    renderData->flowsColors.insert(std::make_pair(flowIt->first, flowColors));
  }
}

/**
 * @brief PatchRetopo::clearInput
 */
void PatchRetopo::clearInput() {
  pDBlookup->clear();
  patchGeometrizator->getTriMesh().Clear();
  patchGeometrizator->getNumberOfCorners() = 0;
  patchGeometrizator->getFlowMap().clear();
  patchGeometrizator->getFlowMapUV().clear();
  patchGeometrizator->getBoundaryLen().clear();
  patchGeometrizator->getBoundary().clear();
  patchGeometrizator->getBoundaryUV().clear();
  renderData->tMesh.Clear();
  renderData->baseMesh.clear();
  renderData->boundary.sides.clear();
  renderData->boundary.colors.clear();
  renderData->flows.clear();
  renderData->flowsColors.clear();
  renderData->nFlows = 0;
  for (size_t i = 0; i < renderData->meshes.size(); ++i)
    renderData->meshes[i].clear();
  renderData->currentMesh = renderData->meshes.size() / 2;
  description.clear();
  baseDescription.clear();
  logStr.clear();
  msg.clear();
  patchInfoTextEdit->setText(logStr);
  emit toStatus(msg);
  emit setMin(0);
  emit setVal(0);
  emit setMax(0);
  emit enable(false);
  emit changed();
}

/**
 * @brief PatchRetopo::selectSortCriterion
 * @param c
 */
void PatchRetopo::selectSortCriterion(int c) {
  std::list<CompareCriteria> criteria;
  switch (c) {
//  case 0:
//    criteria.push_back(CompareCriteria::BLEND_FLOW_QUADS);
//    criteria.push_back(CompareCriteria::FLOW_DISTANCE);
//    criteria.push_back(CompareCriteria::QUADS_QUALITY);
//    criteria.push_back(CompareCriteria::SINGULARITYNUM);
//    criteria.push_back(CompareCriteria::SINGULARITYTYPES);
//    break;
  case 1:
    criteria.push_back(CompareCriteria::FLOW_DISTANCE);
    criteria.push_back(CompareCriteria::SINGULARITYNUM);
    criteria.push_back(CompareCriteria::SINGULARITYTYPES);
    criteria.push_back(CompareCriteria::QUADS_QUALITY);
    break;
  case 2:
    criteria.push_back(CompareCriteria::QUADS_QUALITY);
    criteria.push_back(CompareCriteria::SINGULARITYNUM);
    criteria.push_back(CompareCriteria::SINGULARITYTYPES);
    criteria.push_back(CompareCriteria::FLOW_DISTANCE);
    break;
  case 3:
    criteria.push_back(CompareCriteria::SINGULARITYNUM);
    criteria.push_back(CompareCriteria::SINGULARITYTYPES);
    criteria.push_back(CompareCriteria::FLOW_DISTANCE);
    criteria.push_back(CompareCriteria::QUADS_QUALITY);
    break;
  default:
    break;
  }
  compareCriteria = criteria.front();
  pDBlookup->setSortCriteria(std::move(criteria));
}

/**
 * @brief PatchRetopo::showQuality
 * @param show
 */
void PatchRetopo::showQuality(bool show) {
  showQualityColor = show;
  for (size_t i = 0; i < renderData->meshes.size(); ++i) {
    if (showQualityColor)
      vcg::tri::UpdateColor<PolyMesh>::PerFaceQualityRamp(renderData->meshes[i].mesh, 0, 1);
    else
      vcg::tri::UpdateColor<PolyMesh>::PerFaceConstant(renderData->meshes[i].mesh, LIGHTGRAY);
  }
}

/**
 * @brief PatchRenderer::smoothWithFlowConstraints
 * @param smooth
 */
void PatchRetopo::smoothWithFlowConstraints(bool smooth) {
  patchGeometrizator->setSmoothWithConstraints(smooth);
}

/**
 * @brief PatchRetopo::setMaxSingularities
 * @param maxS
 */
void PatchRetopo::setMaxSingularities(int maxS) {
  if (pDBlookup && maxS >= 0)
    pDBlookup->setMaxSingularities(maxS);
}

/**
 * @brief PatchRetopo::setMinValence
 * @param minV
 */
void PatchRetopo::setMinValence(int minV) {
  if (pDBlookup && minV >= 2 && minV <= std::numeric_limits<valence_type>::max())
    pDBlookup->setMinValence(minV);
}

/**
 * @brief PatchRetopo::setMaxValence
 * @param maxV
 */
void PatchRetopo::setMaxValence(int maxV) {
  if (pDBlookup && maxV >= 2 && maxV <= std::numeric_limits<valence_type>::max())
    pDBlookup->setMaxValence(maxV);
}

/**
 * @brief PatchRetopo::setMaxConcavities
 * @param maxC
 */
void PatchRetopo::setMaxConcavities(int maxC) {
  if (pDBlookup && maxC >= 0 && maxC <= std::numeric_limits<n_corners_type>::max())
    pDBlookup->setMaxConcavities(maxC);
}

/**
 * @brief PatchRetopo::setQualityThreshold
 * @param th
 */
void PatchRetopo::setQualityThreshold(double th) {
  if (pDBlookup && th >= 0 && th <= 1) {
    patchGeometrizator->setQualityThreshold(th);
    pDBlookup->clear();
  }
}

/**
 * @brief PatchRetopo::setILPthreshold
 * @param th
 */
void PatchRetopo::setILPthreshold(int th) {
  if (pDBlookup && th >= 0)
    pDBlookup->setMinSizeThresholdILP(th);
}

/**
 * @brief PatchRetopo::setTolerance
 * @param tolerance
 */
void PatchRetopo::setTolerance(int tolerance) {
  if (pDBlookup && tolerance >= 0)
    pDBlookup->setFlowConstraintsILPtolerance(tolerance);
}

/**
 * @brief PatchRetopo::setLaplacianWeight
 * @param weight
 */
void PatchRetopo::setLaplacianWeight(double weight) {
  patchGeometrizator->setLaplacianWeight(weight);
}

/**
 * @brief PatchRetopo::findPatches
 */
void PatchRetopo::findPatches() {
  // build the boundary
  std::vector<num_type> boundary;
  patchGeometrizator->getBoundaryLen(boundary);

  num_type sum = std::accumulate(boundary.begin(), boundary.end(), 0);
  if (sum < 4) {
    QMessageBox::critical(0, tr("Wrong boundary conditions"), tr("Boundary edges must be at least 4!"));
    return;
  }
  if (sum % 2 != 0) {
    QMessageBox::critical(0, tr("Wrong boundary conditions"), tr("Odd boundary conditions!"));
    return;
  }
  if (boundary.size() != 4 || sum != 4)
    for (size_t i = 0; i < boundary.size(); ++i)
      if (boundary[i] < 2) {
        QMessageBox::critical(0, tr("Wrong boundary conditions"), tr("Boundary sides' length must be at least 2!"));
        return;
      }

  // update GUI
  emit setMin(0);
  emit setVal(0);
  emit setMax(0);
  emit enable(false);
  renderData->baseMesh.clear();
  for (size_t i = 0; i < renderData->meshes.size(); ++i)
    renderData->meshes[i].clear();
#if __cplusplus >= 201103L || _MSC_VER >= 1800
  emit executing(true);
#endif
  msg = "Searching...";
  emit toStatus(msg);

  // search
  currentPatch = -1;
  elapsedFromPreviousUIupdated = std::chrono::nanoseconds::zero();
  pDBlookup->findPatchesFromBoundaryLength(patchGeometrizator->getNumberOfCorners(), boundary,
                                           patchGeometrizator->getFlowMapUV(), patchGeometrizator);
#ifdef SINGLE_THREAD
  count_type nPatches = pDBlookup->numberOfPatches();
  if (nPatches > 0) {
    emit setMin(1);
    emit setMax(nPatches);
    emit enable(true);
    currentPatch = 0;
    emit patchesChangedAt(currentPatch + 1);
    emit executing(false);
    msg += " Expanded: " + QString::number(nPatches) + ". Complete!";
    emit toStatus(msg);
  }
#endif
}

/**
 * @brief PatchRetopo::stopSearching
 */
void PatchRetopo::stopSearching() {
  if (pDBlookup)
    pDBlookup->stop();
}

/**
 * @brief PatchRetopo::selectPatch
 * @param i
 */
void PatchRetopo::selectPatch(int i) {
  std::string tmpStr;
  if (i > 0 && (count_type)i <= pDBlookup->numberOfPatches()) {
    currentPatch = i-1;

    // expand patches
    if (renderData->meshes.empty())
      renderData->meshes.resize(1);
    Patch<PolyMesh> emptyPatch;
    std::deque< Patch<PolyMesh> > patches;
    count_type startIndex = std::max(currentPatch - (int)renderData->currentMesh, 0);
    count_type offset = renderData->meshes.size() - (renderData->currentMesh - (currentPatch - startIndex));
    pDBlookup->at(startIndex, offset, patches);
    for (int nEmptyBefore = currentPatch - (int)renderData->currentMesh; nEmptyBefore < 0; ++nEmptyBefore)
      patches.push_front(emptyPatch);
    while (patches.size() < renderData->meshes.size())
      patches.push_back(emptyPatch);
    for (size_t j = 0; j < renderData->meshes.size(); ++j)
      patchToMeshData(patches[j], renderData->meshes[j], showQualityColor);

    // description
    description.clear();
    description += "Topology: " + QString::fromStdString(patches.at(renderData->currentMesh).getTopologyStr()) + '\n';
    patches.at(renderData->currentMesh).getBoundaryIDsStr(tmpStr);
    description += "BoundaryIDs: " + QString::fromStdString(tmpStr) + '\n';
    patches.at(renderData->currentMesh).getBoundaryVarStr(tmpStr);
    description += "BoundaryVars: " + QString::fromStdString(tmpStr) + '\n';
    patches.at(renderData->currentMesh).getBoundaryIDsEndsStr(tmpStr);
    description += "BoundaryIDsEnds: " + QString::fromStdString(tmpStr) + '\n';
    patches.at(renderData->currentMesh).getBoundaryLenStr(tmpStr);
    description += "BoundaryLens: " + QString::fromStdString(tmpStr) + '\n';
    description += "# singularities: " + QString::number(patches.at(renderData->currentMesh).getNumberOfSingularities()) + '\n';
    description += "Singularities: " + QString::fromStdString(patches.at(renderData->currentMesh).getSingularityStr()) + '\n';
    description += "# corners: " + QString::number(patches.at(renderData->currentMesh).getNumberOfCorners()) + '\n';
    description += "# faces: " + QString::number(patches.at(renderData->currentMesh).getNumberOfFaces()) + '\n';
    description += "# vertices: " + QString::number(patches.at(renderData->currentMesh).getNumberOfVertices()) + '\n';
    description += "# occurrences: " + QString::number(patches.at(renderData->currentMesh).getNumberOfOccurrences()) + '\n';
    description += "Perimeter (reconstructed patch): " + QString::number(patches.at(renderData->currentMesh).getPerimeter()) + '\n';
    description += "# rotations (reconstructed patch): " + QString::number(patches.at(renderData->currentMesh).getNumberOfRotations()) + '\n';
    description += "# faces (reconstructed patch): " + QString::number(renderData->meshes[renderData->currentMesh].mesh.FN()) + '\n';
    description += "# vertices (reconstructed patch): " + QString::number(renderData->meshes[renderData->currentMesh].mesh.VN());

    // base mesh
    Patch<PolyMesh> basePatch = patches.at(renderData->currentMesh);
    basePatch.getPolychordMap().clear();
    std::string bIDs;
    basePatch.getBoundaryIDsStr(bIDs);
    basePatch.setBoundaryFromStr(bIDs);
    basePatch.buildExpandedTopologyStr();
    PatchRetopologizer::PatchFlatteningBySmoothing(basePatch.getMesh(), basePatch.getStartCorner(),
                                                   basePatch.getBoundaryLenVec(), PatchRetopologizer::RADIUS);
    patchToMeshData(basePatch, renderData->baseMesh);
    renderData->baseMesh.vertexRadius = Coloring::ComputeVertexRadius(renderData->baseMesh.mesh);
    // description
    baseDescription.clear();
    baseDescription += "Topology: " + QString::fromStdString(basePatch.getTopologyStr()) + '\n';
    basePatch.getBoundaryIDsStr(tmpStr);
    baseDescription += "BoundaryIDs: " + QString::fromStdString(tmpStr) + '\n';
    basePatch.getBoundaryVarStr(tmpStr);
    baseDescription += "BoundaryVars: " + QString::fromStdString(tmpStr) + '\n';
    basePatch.getBoundaryIDsEndsStr(tmpStr);
    baseDescription += "BoundaryIDsEnds: " + QString::fromStdString(tmpStr) + '\n';
    basePatch.getBoundaryLenStr(tmpStr);
    baseDescription += "BoundaryLens: " + QString::fromStdString(tmpStr) + '\n';
    baseDescription += "Singularities: " + QString::fromStdString(basePatch.getSingularityStr()) + '\n';
    baseDescription += "Perimeter: " + QString::number(basePatch.getPerimeter()) + '\n';
    baseDescription += "# corners: " + QString::number(basePatch.getNumberOfCorners()) + '\n';
    baseDescription += "# faces: " + QString::number(basePatch.getNumberOfFaces()) + '\n';
    baseDescription += "# vertices: " + QString::number(basePatch.getNumberOfVertices()) + '\n';
    baseDescription += "# rotations: " + QString::number(basePatch.getNumberOfRotations()) + '\n';
    baseDescription += "# occurrences: " + QString::number(basePatch.getNumberOfOccurrences());

    logStr = description;
    patchInfoTextEdit->setText(logStr);
  } else {
    currentPatch = 0;
    renderData->baseMesh.clear();
    for (size_t i = 0; i < renderData->meshes.size(); ++i)
      renderData->meshes[i].clear();
    description.clear();
    baseDescription.clear();
    logStr = tr("No patches found.");
    patchInfoTextEdit->setText(logStr);
  }

  emit changed();
}

/**
 * @brief PatchRetopo::patchToMeshData
 * @param patch
 * @param meshData
 */
void PatchRetopo::patchToMeshData(const Patch<PolyMesh> &patch, MeshData &meshData, const bool qualityColor) {
  meshData.clear();
  if (patch.isNull())
    return;
  // copy mesh
  vcg::face::Pos<PolyFace> startPos;
  patch.getMeshAndStartCorner(meshData.mesh, startPos);
  patch.getPolychordMap(meshData.mesh, meshData.polychordMap);
  // initialize mark
  vcg::tri::InitFaceIMark(meshData.mesh);
  vcg::tri::InitVertexIMark(meshData.mesh);
  vcg::tri::IMark(meshData.mesh) = 0;
  // set vertex radius
  meshData.vertexRadius = renderData->vertexRadius;
  // colorize faces
  if (qualityColor)
    vcg::tri::UpdateColor<PolyMesh>::PerFaceQualityRamp(meshData.mesh, 0, 1);
  else
    vcg::tri::UpdateColor<PolyMesh>::PerFaceConstant(meshData.mesh, LIGHTGRAY);
  // colorize vertices
  Coloring::SetVertexColorByValence(meshData.mesh);
  // colorize the boundary
  vcg::face::Pos<PolyFace> runPos = startPos;
  std::list<CoordType> l;
  n_corners_type range = patch.getNumberOfCorners();
  if (range == 0)
    range = 1;
  n_corners_type side = 0;
  vcg::Color4b c = vcg::Color4b::Scatter(range, side);
  do {
    l.push_back(runPos.V()->P());
    runPos.FlipV();
    runPos.FlipE();
    if (!runPos.IsBorder()) {
      runPos.FlipF();
      runPos.FlipE();
      if (!runPos.IsBorder()) {   // concave corner
        while (!runPos.IsBorder()) {
          runPos.FlipF();
          runPos.FlipE();
        }
        l.push_back(runPos.V()->P());
        c = vcg::Color4b::Scatter(range, side);
        meshData.polyLines.push_back(PolyLine(c, l));
        l.clear();
        ++side;
      } else {
        if (runPos == startPos) {
          l.push_back(runPos.V()->P());
        }
      }
    } else {    // convex corner
      l.push_back(runPos.V()->P());
      c = vcg::Color4b::Scatter(range, side);
      meshData.polyLines.push_back(PolyLine(c, l));
      l.clear();
      side++;
    }
  } while (runPos != startPos);
  if (patch.getNumberOfCorners() == 0)
    meshData.polyLines.push_back(PolyLine(c, l));
  // print labels
  std::string bEqTmp;
  patch.getBoundaryIDsStr(bEqTmp);
  size_t ind = 0;
  for (n_corners_type r = 0; r < patch.getNumberOfRotations(); ++r) {
    while (bEqTmp[ind] != SEPARATOR)
      ++ind;
    ++ind;
  }
  std::string bEq(bEqTmp);
  for (size_t i = 0; i < bEq.length(); ++i)
    if (ind >= bEqTmp.length()) {
      bEq[i] = SEPARATOR;
      ind = 0;
    } else {
      bEq[i] = bEqTmp[ind];
      ++ind;
    }
  std::list< std::pair<var_label_type, CoordType> > labelPosList;
  printBoundaryIDs(startPos, bEq, patch.getBoundaryVarVec(), labelPosList);
  std::list< std::pair<var_label_type, CoordType> >::iterator lIt;
  std::stringstream stream;
  std::string str;
  for (lIt = labelPosList.begin(); lIt != labelPosList.end(); ++lIt) {
    str.clear();
    stream.str(str);
    stream << VARLABEL << "<sub>" << std::hex << std::noshowbase << lIt->first << "</sub>";
    meshData.strings.push_back(Text(DARKGRAY, lIt->second, stream.str()));
  }
  switch (compareCriteria) {
  case CompareCriteria::FLOW_DISTANCE:
    meshData.value = patch.getFlowDistance();
    break;
  case CompareCriteria::QUADS_QUALITY:
    meshData.value = patch.getQuadsQuality();
    break;
  case CompareCriteria::SINGULARITYNUM:
    meshData.value = patch.getNumberOfSingularities();
    break;
  default:
    meshData.value = 0;
    break;
  }
}

/**
 * @brief PatchRetopo::printBoundaryIDs
 * @param startPos
 * @param mesh
 * @param bEq
 * @param bVars
 */
void PatchRetopo::printBoundaryIDs(const vcg::face::Pos<PolyFace> &startPos, const std::string &bEq,
                                   const std::vector<num_type> &bVars,
                                   std::list<std::pair<var_label_type, CoordType> > &labels) {
  labels.clear();
  // print labels
  PolyMesh::CoordType v1(0,0,0), v2(0,0,0);
  std::stringstream stream;
  size_t sInd = 0, vInd = 0;
  stream << bEq.substr(sInd + 1, 2 * sizeof(var_label_type)) + std::string(" ");
  stream >> std::hex >> vInd;
  var_label_type count = 0;
  vcg::face::Pos<PolyFace> runPos = startPos;
  do {
    // find location
    v1 = runPos.VFlip()->P() - runPos.V()->P();
    v2[0] = v1[1];
    v2[1] = -v1[0];
    v1 = (runPos.V()->P() + runPos.VFlip()->P()) / 2.0 + v2 * 0.1;
    // build the label
    labels.push_back(std::pair<var_label_type, PolyMesh::CoordType>(vInd, v1));
    // go to next edge
    runPos.FlipV();
    runPos.FlipE();
    if (!runPos.IsBorder()) {
      runPos.FlipF();
      runPos.FlipE();
      if (!runPos.IsBorder()) {   // concave corner
        while (!runPos.IsBorder()) {
          runPos.FlipF();
          runPos.FlipE();
        }
        ++sInd;
      }
    } else {    // convex corner
      // skip SEPARATOR
      ++sInd;
    }
    // next label
    ++count;
    if (count >= bVars[vInd] && runPos != startPos) {
      count = 0;
      sInd += 1/*VARLABEL*/ + 2 * sizeof(var_label_type)/*variable index*/;
      stream.str(bEq.substr(sInd + 1, 2 * sizeof(var_label_type)) + std::string(" "));
      stream >> std::hex >> vInd;
    }
  } while (runPos != startPos);
}
