#ifndef RENDER_DATA_H
#define RENDER_DATA_H

#include <vector>
#include <list>
#include "mesh_utilities.h"

#define NSAMPLES 100

typedef PolyMesh::ScalarType    ScalarType;
typedef PolyMesh::CoordType     CoordType;

struct PolyLine {
  vcg::Color4b            color;
  std::list<CoordType>    points;
  std::list<ScalarType>   arcLengths;
  PolyLine() : color(LIGHTGREEN) { }
  PolyLine(const vcg::Color4b &c) : color(c) { }
  PolyLine(const vcg::Color4b &c, const std::list<CoordType> &pp) : color(c), points(pp) {
    ComputeArcLengths(points, arcLengths);
  }
  void updateArcLengths() {
    ComputeArcLengths(points, arcLengths);
  }
  template < typename PointContainerType >
  static void ComputeArcLengths(const PointContainerType &pp, std::list<ScalarType> &al) {
    al.clear();
    CoordType lastPoint;
    typename PointContainerType::const_iterator pIt = pp.begin();
    if (pIt != pp.end()) {
      lastPoint = *pIt;
      al.push_back(0);
      ++pIt;
    }
    for (; pIt != pp.end(); lastPoint = *pIt, ++pIt)
      al.push_back(al.back() + (*pIt - lastPoint).Norm());
  }
};

struct Text {
  vcg::Color4b            color;
  CoordType               point;
  std::string             text;
  Text() : color(BLACK), point(0, 0, 0) { }
  Text(const vcg::Color4b &c, const CoordType &p, const std::string &str) : color(c), point(p), text(str) { }
};

struct MeshData {
  typedef PolychordSupport<PolyMesh>::PolychordMap     PolychordMap;
  PolyMesh                mesh;
  ScalarType              vertexRadius;
  std::list<Text>         strings;
  std::list<PolyLine>     polyLines;
  PolychordMap            polychordMap;
  ScalarType              value;
  MeshData() : vertexRadius(0), value(0) { }
  MeshData(const MeshData &md) {
    *this = md;
  }
  MeshData & operator = (const MeshData &md) {
    if (&md != this) {
      vcg::tri::Append<PolyMesh, PolyMesh>::MeshCopy(mesh, const_cast<PolyMesh &>(md.mesh), false, true);
      vcg::tri::UpdateTopology<PolyMesh>::FaceFace(mesh);
      vertexRadius = md.vertexRadius;
      strings = md.strings;
      polyLines = md.polyLines;
      polychordMap = md.polychordMap;
      value = md.value;
    }
    return *this;
  }

  void clear() {
    mesh.Clear();
    vertexRadius = 0;
    strings.clear();
    polyLines.clear();
    polychordMap.clear();
    value = std::numeric_limits<ScalarType>::max();
  }
};

struct Boundary {
  typedef PolychordSupport<PolyMesh>::PointContainer   PointContainer;
  std::vector<PointContainer>   sides;
  std::vector<vcg::Color4b>     colors;
};

enum ShowVertices {
  NONE,
  ALL,
  SINGS
};

struct RenderData {
  typedef PolychordSupport<PolyMesh>::FlowMap             FlowMap;
  typedef std::deque<vcg::Color4b>                        FlowColors;
  typedef std::map<FlowTopology, FlowColors>              FlowColorsMap;

  MeshData                baseMesh;
  bool                    showBaseMesh;
  TMesh                   tMesh;
  bool                    showTriMesh;
  MeshData                provMesh;
  bool                    showProvMesh;
  std::vector<MeshData>   meshes;
  size_t                  currentMesh;
  bool                    showPreview;
  bool                    showStrings;
  ShowVertices            showVertices;
  Boundary                boundary;
  ScalarType              vertexRadius;
  ScalarType              pickRadius;
  FlowMap                 flows;
  FlowColorsMap           flowsColors;
  u_small_int_type        nFlows;
  bool                    showPolychords;
  bool                    showSampleVectors;

  RenderData() : showBaseMesh(false)
               , showTriMesh(true)
               , showProvMesh(false)
               , currentMesh(0)
               , showPreview(false)
               , showStrings(false)
               , showVertices(NONE)
               , pickRadius(0)
               , nFlows(0)
               , showPolychords(false)
               , showSampleVectors(false)
               { }
};

#endif // RENDER_DATA_H
