/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the class library                   */
/*       SoPlex --- the Sequential object-oriented simPlex.                  */
/*                                                                           */
/*    Copyright (C) 1996-2014 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SoPlex is distributed under the terms of the ZIB Academic Licence.       */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SoPlex; see the file COPYING. If not email to soplex@zib.de.  */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file  settings.cpp
 * @brief Settings class
 */

#include <string.h>

#include "spxdefines.h"
#include "settings.h"
#include "soplex.h"

namespace soplex
{
   /// default constructor initializing default settings
   Settings::Settings(bool init)
   {
      _defaultsAndBoundsInitialized = init;

      if( !_defaultsAndBoundsInitialized )
      {
         // should lifting be used to reduce range of nonzero matrix coefficients?
         _boolParamName[LIFTING] = "lifting";
         _boolParamDescription[LIFTING] = "should lifting be used to reduce range of nonzero matrix coefficients?";
         _boolParamDefault[LIFTING] = false;

         // should LP be transformed to equality form before a rational solve?
         _boolParamName[EQTRANS] = "eqtrans";
         _boolParamDescription[EQTRANS] = "should LP be transformed to equality form before a rational solve?";
         _boolParamDefault[EQTRANS] = false;

         // objective sense
         _intParamName[OBJSENSE] = "objsense";
         _intParamDescription[OBJSENSE] = "objective sense (-1 - minimize, +1 - maximize)";
         _intParamLower[OBJSENSE] = -1;
         _intParamUpper[OBJSENSE] = 1;
         _intParamDefault[OBJSENSE] = OBJSENSE_MAXIMIZE;

         // type of computational form, i.e., column or row representation
         _intParamName[REPRESENTATION] = "representation";
         _intParamDescription[REPRESENTATION] = "type of computational form (0 - auto, 1 - column representation, 2 - row representation)";
         _intParamLower[REPRESENTATION] = 0;
         _intParamUpper[REPRESENTATION] = 2;
         _intParamDefault[REPRESENTATION] = REPRESENTATION_AUTO;

         // type of algorithm, i.e., enter or leave
         _intParamName[ALGORITHM] = "algorithm";
         _intParamDescription[ALGORITHM] = "type of algorithm (0 - enter, 1 - leave)";
         _intParamLower[ALGORITHM] = 0;
         _intParamUpper[ALGORITHM] = 1;
         _intParamDefault[ALGORITHM] = ALGORITHM_LEAVE;

         // type of LU update
         _intParamName[FACTOR_UPDATE_TYPE] = "factor_update_type";
         _intParamDescription[FACTOR_UPDATE_TYPE] = "type of LU update (0 - eta update, 1 - Forrest-Tomlin update)";
         _intParamLower[FACTOR_UPDATE_TYPE] = 0;
         _intParamUpper[FACTOR_UPDATE_TYPE] = 1;
         _intParamDefault[FACTOR_UPDATE_TYPE] = FACTOR_UPDATE_TYPE_FT;

         ///@todo which value?
         // maximum number of updates without fresh factorization
         _intParamName[FACTOR_UPDATE_MAX] = "factor_update_max";
         _intParamDescription[FACTOR_UPDATE_MAX] = "maximum number of LU updates without fresh factorization";
         _intParamLower[FACTOR_UPDATE_MAX] = 0;
         _intParamUpper[FACTOR_UPDATE_MAX] = INT_MAX;
         _intParamDefault[FACTOR_UPDATE_MAX] = 200;

         // iteration limit (-1 if unlimited)
         _intParamName[ITERLIMIT] = "iterlimit";
         _intParamDescription[ITERLIMIT] = "iteration limit (-1 - no limit)";
         _intParamLower[ITERLIMIT] = -1;
         _intParamUpper[ITERLIMIT] = INT_MAX;
         _intParamDefault[ITERLIMIT] = -1;

         // refinement limit (-1 if unlimited)
         _intParamName[REFLIMIT] = "reflimit";
         _intParamDescription[REFLIMIT] = "refinement limit (-1 - no limit)";
         _intParamLower[REFLIMIT] = 100;
         _intParamUpper[REFLIMIT] = INT_MAX;
         _intParamDefault[REFLIMIT] = -1;

         // stalling refinement limit (-1 if unlimited)
         _intParamName[STALLREFLIMIT] = "stallreflimit";
         _intParamDescription[STALLREFLIMIT] = "stalling refinement limit (-1 - no limit)";
         _intParamLower[STALLREFLIMIT] = -1;
         _intParamUpper[STALLREFLIMIT] = INT_MAX;
         _intParamDefault[STALLREFLIMIT] = -1;

         // display frequency
         _intParamName[DISPLAYFREQ] = "displayfreq";
         _intParamDescription[DISPLAYFREQ] = "display frequency";
         _intParamLower[DISPLAYFREQ] = 1;
         _intParamUpper[DISPLAYFREQ] = INT_MAX;
         _intParamDefault[DISPLAYFREQ] = 200;

         // verbosity level
         _intParamName[VERBOSITY] = "verbosity";
         _intParamDescription[VERBOSITY] = "verbosity level (0 - error, 1 - warning, 2 - debug, 3 - normal, 4 - high, 5 - full)";
         _intParamLower[VERBOSITY] = 0;
         _intParamUpper[VERBOSITY] = 5;
         _intParamDefault[VERBOSITY] = VERBOSITY_NORMAL;

         // type of simplifier
         _intParamName[SIMPLIFIER] = "simplifier";
         _intParamDescription[SIMPLIFIER] = "simplifier (0 - off, 1 - auto)";
         _intParamLower[SIMPLIFIER] = 0;
         _intParamUpper[SIMPLIFIER] = 1;
         _intParamDefault[SIMPLIFIER] = SIMPLIFIER_AUTO;

         // type of scaler
         _intParamName[SCALER] = "scaler";
         _intParamDescription[SCALER] = "scaling (0 - off, 1 - uni-equilibrium, 2 - bi-equilibrium, 3 - geometric, 4 - iterated geometric)";
         _intParamLower[SCALER] = 0;
         _intParamUpper[SCALER] = 4;
         _intParamDefault[SCALER] = SCALER_BIEQUI;

         // type of starter used to create crash basis
         _intParamName[STARTER] = "starter";
         _intParamDescription[STARTER] = "crash basis generated when starting from scratch (0 - none, 1 - weight, 2 - sum, 3 - vector)";
         _intParamLower[STARTER] = 0;
         _intParamUpper[STARTER] = 3;
         _intParamDefault[STARTER] = STARTER_OFF;

         // type of pricer
         _intParamName[PRICER] = "pricer";
         _intParamDescription[PRICER] = "pricing method (0 - auto, 1 - dantzig, 2 - parmult, 3 - devex, 4 - quicksteep, 5 - steep)";
         _intParamLower[PRICER] = 0;
         _intParamUpper[PRICER] = 5;
         _intParamDefault[PRICER] = PRICER_AUTO;

         // type of ratio test
         _intParamName[RATIOTESTER] = "ratiotester";
         _intParamDescription[RATIOTESTER] = "method for ratio test (0 - textbook, 1 - harris, 2 - fast, 3 - boundflipping)";
         _intParamLower[RATIOTESTER] = 0;
         _intParamUpper[RATIOTESTER] = 3;
         _intParamDefault[RATIOTESTER] = RATIOTESTER_FAST;

         // mode for synchronizing real and rational LP
         _intParamName[SYNCMODE] = "syncmode";
         _intParamDescription[SYNCMODE] = "mode for synchronizing real and rational LP (0 - store only real LP, 1 - auto, 2 - manual)";
         _intParamLower[SYNCMODE] = 0;
         _intParamUpper[SYNCMODE] = 2;
         _intParamDefault[SYNCMODE] = SYNCMODE_ONLYREAL;

         // mode for reading LP files
         _intParamName[READMODE] = "readmode";
         _intParamDescription[READMODE] = "mode for reading LP files (0 - floating-point, 1 - rational)";
         _intParamLower[READMODE] = 0;
         _intParamUpper[READMODE] = 1;
         _intParamDefault[READMODE] = READMODE_REAL;

         // mode for iterative refinement strategy
         _intParamName[SOLVEMODE] = "solvemode";
         _intParamDescription[SOLVEMODE] = "mode for iterative refinement strategy (0 - floating-point solve, 1 - auto, 2 - force iterative refinement)";
         _intParamLower[SOLVEMODE] = 0;
         _intParamUpper[SOLVEMODE] = 2;
         _intParamDefault[SOLVEMODE] = SOLVEMODE_AUTO;

         // mode for iterative refinement strategy
         _intParamName[CHECKMODE] = "checkmode";
         _intParamDescription[CHECKMODE] = "mode for a posteriori feasibility checks (0 - floating-point check, 1 - auto, 2 - rational check)";
         _intParamLower[CHECKMODE] = 0;
         _intParamUpper[CHECKMODE] = 2;
         _intParamDefault[CHECKMODE] = CHECKMODE_AUTO;

         // mode for hyper sparse pricing
         _intParamName[HYPER_PRICING] = "hyperpricing";
         _intParamDescription[HYPER_PRICING] = "mode for hyper sparse pricing (0 - off, 1 - auto, 2 - always)";
         _intParamLower[HYPER_PRICING] = 0;
         _intParamUpper[HYPER_PRICING] = 2;
         _intParamDefault[HYPER_PRICING] = HYPER_PRICING_AUTO;

         // primal feasibility tolerance
         _realParamName[FEASTOL] = "feastol";
         _realParamDescription[FEASTOL] = "primal feasibility tolerance";
         _realParamLower[FEASTOL] = 0.0;
         _realParamUpper[FEASTOL] = 1.0;
         _realParamDefault[FEASTOL] = 1e-6;

         // dual feasibility tolerance
         _realParamName[OPTTOL] = "opttol";
         _realParamDescription[OPTTOL] = "dual feasibility tolerance";
         _realParamLower[OPTTOL] = 0.0;
         _realParamUpper[OPTTOL] = 1.0;
         _realParamDefault[OPTTOL] = 1e-6;

         ///@todo define suitable values depending on Real type
         // general zero tolerance
         _realParamName[EPSILON_ZERO] = "epsilon_zero";
         _realParamDescription[EPSILON_ZERO] = "general zero tolerance";
         _realParamLower[EPSILON_ZERO] = 0.0;
         _realParamUpper[EPSILON_ZERO] = 1.0;
         _realParamDefault[EPSILON_ZERO] = DEFAULT_EPS_ZERO;

         ///@todo define suitable values depending on Real type
         // zero tolerance used in factorization
         _realParamName[EPSILON_FACTORIZATION] = "epsilon_factorization";
         _realParamDescription[EPSILON_FACTORIZATION] = "zero tolerance used in factorization";
         _realParamLower[EPSILON_FACTORIZATION] = 0.0;
         _realParamUpper[EPSILON_FACTORIZATION] = 1.0;
         _realParamDefault[EPSILON_FACTORIZATION] = DEFAULT_EPS_FACTOR;

         ///@todo define suitable values depending on Real type
         // zero tolerance used in update of the factorization
         _realParamName[EPSILON_UPDATE] = "epsilon_update";
         _realParamDescription[EPSILON_UPDATE] = "zero tolerance used in update of the factorization";
         _realParamLower[EPSILON_UPDATE] = 0.0;
         _realParamUpper[EPSILON_UPDATE] = 1.0;
         _realParamDefault[EPSILON_UPDATE] = DEFAULT_EPS_UPDATE;

         ///@todo define suitable values depending on Real type
         // pivot zero tolerance used in factorization
         _realParamName[EPSILON_PIVOT] = "epsilon_pivot";
         _realParamDescription[EPSILON_PIVOT] = "pivot zero tolerance used in factorization";
         _realParamLower[EPSILON_PIVOT] = 0.0;
         _realParamUpper[EPSILON_PIVOT] = 1.0;
         _realParamDefault[EPSILON_PIVOT] = DEFAULT_EPS_PIVOT;

         ///@todo define suitable values depending on Real type
         // infinity threshold
         _realParamName[INFTY] = "infty";
         _realParamDescription[INFTY] = "infinity threshold";
         _realParamLower[INFTY] = 1e10;
         _realParamUpper[INFTY] = 1e100;
         _realParamDefault[INFTY] = DEFAULT_INFINITY;

         // time limit in seconds (INFTY if unlimited)
         _realParamName[TIMELIMIT] = "timelimit";
         _realParamDescription[TIMELIMIT] = "time limit in seconds";
         _realParamLower[TIMELIMIT] = 0.0;
         _realParamUpper[TIMELIMIT] = DEFAULT_INFINITY;
         _realParamDefault[TIMELIMIT] = DEFAULT_INFINITY;

         // lower limit on objective value
         _realParamName[OBJLIMIT_LOWER] = "objlimit_lower";
         _realParamDescription[OBJLIMIT_LOWER] = "lower limit on objective value";
         _realParamLower[OBJLIMIT_LOWER] = -DEFAULT_INFINITY;
         _realParamUpper[OBJLIMIT_LOWER] = DEFAULT_INFINITY;
         _realParamDefault[OBJLIMIT_LOWER] = -DEFAULT_INFINITY;

         // upper limit on objective value
         _realParamName[OBJLIMIT_UPPER] = "objlimit_upper";
         _realParamDescription[OBJLIMIT_UPPER] = "upper limit on objective value";
         _realParamLower[OBJLIMIT_UPPER] = -DEFAULT_INFINITY;
         _realParamUpper[OBJLIMIT_UPPER] = DEFAULT_INFINITY;
         _realParamDefault[OBJLIMIT_UPPER] = DEFAULT_INFINITY;

         // working tolerance for feasibility in floating-point solver during iterative refinement
         _realParamName[FPFEASTOL] = "fpfeastol";
         _realParamDescription[FPFEASTOL] = "working tolerance for feasibility in floating-point solver during iterative refinement";
         _realParamLower[FPFEASTOL] = 1e-12;
         _realParamUpper[FPFEASTOL] = 1.0;
         _realParamDefault[FPFEASTOL] = 1e-6;

         // working tolerance for optimality in floating-point solver during iterative refinement
         _realParamName[FPOPTTOL] = "fpopttol";
         _realParamDescription[FPOPTTOL] = "working tolerance for optimality in floating-point solver during iterative refinement";
         _realParamLower[FPOPTTOL] = 1e-12;
         _realParamUpper[FPOPTTOL] = 1.0;
         _realParamDefault[FPOPTTOL] = 1e-6;

         // maximum increase of scaling factors between refinements
         _realParamName[MAXSCALEINCR] = "maxscaleincr";
         _realParamDescription[MAXSCALEINCR] = "maximum increase of scaling factors between refinements";
         _realParamLower[MAXSCALEINCR] = 1.0;
         _realParamUpper[MAXSCALEINCR] = DEFAULT_INFINITY;
         _realParamDefault[MAXSCALEINCR] = 1e12;

         // lower threshold in lifting (nonzero matrix coefficients with smaller absolute value will be reformulated)
         _realParamName[LIFTMINVAL] = "liftminval";
         _realParamDescription[LIFTMINVAL] = "lower threshold in lifting (nonzero matrix coefficients with smaller absolute value will be reformulated)";
         _realParamLower[LIFTMINVAL] = 0.0;
         _realParamUpper[LIFTMINVAL] = 0.1;
         _realParamDefault[LIFTMINVAL] = 0.000976562; // = 1/1024

         // upper threshold in lifting (nonzero matrix coefficients with larger absolute value will be reformulated)
         _realParamName[LIFTMAXVAL] = "liftmaxval";
         _realParamDescription[LIFTMAXVAL] = "lower threshold in lifting (nonzero matrix coefficients with smaller absolute value will be reformulated)";
         _realParamLower[LIFTMAXVAL] = 10.0;
         _realParamUpper[LIFTMAXVAL] = DEFAULT_INFINITY;
         _realParamDefault[LIFTMAXVAL] = 1024.0;

         // threshold for using sparse pricing (no. of violations need to be smaller than threshold * dimension of problem)
         _realParamName[SPARSITY_THRESHOLD] = "sparsity_threshold";
         _realParamDescription[SPARSITY_THRESHOLD] = "sparse pricing threshold (#violations < dimension * SPARSITY_THRESHOLD activates sparse pricing)";
         _realParamLower[SPARSITY_THRESHOLD] = 0.0;
         _realParamUpper[SPARSITY_THRESHOLD] = 1.0;
         _realParamDefault[SPARSITY_THRESHOLD] = 0.6;

         _defaultsAndBoundsInitialized = true;
      }

      for( int i = 0; i < BOOLPARAM_COUNT; i++ )
         _boolParamValues[i] = _boolParamDefault[i];

      for( int i = 0; i < INTPARAM_COUNT; i++ )
         _intParamValues[i] = _intParamDefault[i];

      for( int i = 0; i < REALPARAM_COUNT; i++ )
         _realParamValues[i] = _realParamDefault[i];
   }

   /// copy constructor
   Settings::Settings(const Settings& settings)
   {
      *this = settings;
   }

   /// assignment operator
   Settings& Settings::operator=(const Settings& settings)
   {
      for( int i = 0; i < BOOLPARAM_COUNT; i++ )
         _boolParamValues[i] = settings._boolParamValues[i];

      for( int i = 0; i < INTPARAM_COUNT; i++ )
         _intParamValues[i] = settings._intParamValues[i];

      for( int i = 0; i < REALPARAM_COUNT; i++ )
         _realParamValues[i] = settings._realParamValues[i];

      _defaultsAndBoundsInitialized = true;

      return *this;
   }

   /// returns boolean parameter value
   bool boolParam(const BoolParam param) const
   {
      assert(param >= 0);
      assert(param < BOOLPARAM_COUNT);
      return _boolParamValues[param];
   }



   /// returns integer parameter value
   int intParam(const IntParam param) const
   {
      assert(param >= 0);
      assert(param < INTPARAM_COUNT);
      return _intParamValues[param];
   }



   /// returns real parameter value
   Real realParam(const RealParam param) const
   {
      assert(param >= 0);
      assert(param < REALPARAM_COUNT);
      return _realParamValues[param];
   }



   /// returns current parameter settings
   const Settings& settings() const
   {
      return *this;
   }



   /// sets boolean parameter value; returns true on success
   bool setBoolParam(const BoolParam param, const bool value, const bool quiet, const bool init)
   {
      assert(param >= 0);
      assert(param < BOOLPARAM_COUNT);
      assert(init || SoPlex::_isConsistent());

      if( !init && value == boolParam(param) )
         return true;

      switch( param )
      {
      case LIFTING:
         break;
      case EQTRANS:
         break;
      default:
         return false;
      }

      _boolParamValues[param] = value;
      return true;
   }



   /// sets integer parameter value; returns true on success
   bool setIntParam(const IntParam param, const int value, const bool quiet, const bool init)
   {
      assert(param >= 0);
      assert(param < INTPARAM_COUNT);
      assert(init || SoPlex::_isConsistent());

      if( !init && value == intParam(param) )
         return true;

      // check for a valid parameter value wrt bounds
      if( value < _intParamLower[param] || value > _intParamUpper[param] )
         return false;

      switch( param )
      {
      // objective sense
      case OBJSENSE:
         if( value != OBJSENSE_MAXIMIZE && value != OBJSENSE_MINIMIZE )
            return false;
         SoPlex::_realLP->changeSense(value == OBJSENSE_MAXIMIZE ? SPxLPReal::MAXIMIZE : SPxLPReal::MINIMIZE);
         if( SoPlex::_rationalLP != 0 )
            SoPlex::_rationalLP->changeSense(value == OBJSENSE_MAXIMIZE ? SPxLPRational::MAXIMIZE : SPxLPRational::MINIMIZE);
         SoPlex::_invalidateSolution();
         break;

      // type of computational form, i.e., column or row representation
      case REPRESENTATION:
         if( value != REPRESENTATION_COLUMN && value != REPRESENTATION_ROW && value != REPRESENTATION_AUTO )
            return false;
         break;

      // type of algorithm, i.e., enter or leave
      case ALGORITHM:
         if( value != ALGORITHM_ENTER && value != ALGORITHM_LEAVE )
            return false;
         SoPlex::_solver.setType(value == ALGORITHM_ENTER ? SPxSolver::ENTER : SPxSolver::LEAVE);
         break;

      // type of LU update
      case FACTOR_UPDATE_TYPE:
         if( value != FACTOR_UPDATE_TYPE_ETA || value != FACTOR_UPDATE_TYPE_FT )
            return false;
         SoPlex::_slufactor.setUtype(value == FACTOR_UPDATE_TYPE_ETA ? SLUFactor::ETA : SLUFactor::FOREST_TOMLIN);
         break;

      // maximum number of updates before fresh factorization
      case FACTOR_UPDATE_MAX:
         if( value <= 1 )
            return false;
         SoPlex::_solver.basis().setMaxUpdates(value);
         break;

      // iteration limit (-1 if unlimited)
      case ITERLIMIT:
         if( value < -1 )
            return false;
         break;

      // display frequency
      case DISPLAYFREQ:
         if( value <= 0 )
            return false;
         SoPlex::_solver.setDisplayFreq(value);
         break;

      // verbosity level
      case VERBOSITY:
         Param::setVerbose(value);
         break;

      // type of simplifier
      case SIMPLIFIER:
         switch( value )
         {
         case SIMPLIFIER_OFF:
            SoPlex::_simplifier = 0;
            break;
         case SIMPLIFIER_AUTO:
            SoPlex::_simplifier = &SoPlex::_simplifierMainSM;
            assert(SoPlex::_simplifier != 0);
            break;
         default:
            return false;
         }
         break;

      // type of scaler
      case SCALER:
         switch( value )
         {
         case SCALER_OFF:
            SoPlex::_scaler = 0;
            break;
         case SCALER_UNIEQUI:
            SoPlex::_scaler = &SoPlex::_scalerUniequi;
            break;
         case SCALER_BIEQUI:
            SoPlex::_scaler = &SoPlex::_scalerBiequi;
            break;
         case SCALER_GEO1:
            SoPlex::_scaler = &SoPlex::_scalerGeo1;
            break;
         case SCALER_GEO8:
            SoPlex::_scaler = &SoPlex::_scalerGeo8;
            break;
         default:
            return false;
         }
         break;

      // type of starter used to create crash basis
      case STARTER:
         switch( value )
         {
         case STARTER_OFF:
            SoPlex::_starter = 0;
            break;
         case STARTER_WEIGHT:
            SoPlex::_starter = &SoPlex::_starterWeight;
            break;
         case STARTER_SUM:
            SoPlex::_starter = &SoPlex::_starterSum;
            break;
         case STARTER_VECTOR:
            SoPlex::_starter = &SoPlex::_starterVector;
            break;
         default:
            return false;
         }
         break;

      // type of pricer
      case PRICER:
         switch( value )
         {
         case PRICER_AUTO:
            SoPlex::_solver.setPricer(&SoPlex::_pricerAuto);
            break;
         case PRICER_DANTZIG:
            SoPlex::_solver.setPricer(&SoPlex::_pricerDantzig);
            break;
         case PRICER_PARMULT:
            SoPlex::_solver.setPricer(&SoPlex::_pricerParMult);
            break;
         case PRICER_DEVEX:
            SoPlex::_solver.setPricer(&SoPlex::_pricerDevex);
            break;
         case PRICER_QUICKSTEEP:
            SoPlex::_solver.setPricer(&SoPlex::_pricerQuickSteep);
            break;
         case PRICER_STEEP:
            SoPlex::_solver.setPricer(&SoPlex::_pricerSteep);
            break;
         default:
            return false;
         }
         break;

      // mode for synchronizing real and rational LP
      case SYNCMODE:
         switch( value )
         {
         case SYNCMODE_ONLYREAL:
            if( _rationalLP != 0 )
            {
               SoPlex::_rationalLP->~SPxLPRational();
               spx_free(SoPlex::_rationalLP);
            }
            break;
         case SYNCMODE_AUTO:
            if( intParam(param) == SYNCMODE_ONLYREAL )
               SoPlex::_syncLPRational();
            break;
         case SYNCMODE_MANUAL:
            SoPlex::_ensureRationalLP();
            break;
         default:
            return false;
         }
         break;

      // mode for reading LP files; nothing to do but change the value if valid
      case READMODE:
         switch( value )
         {
         case READMODE_REAL:
         case READMODE_RATIONAL:
            break;
         default:
            return false;
         }
         break;

      // mode for iterative refinement strategy; nothing to do but change the value if valid
      case SOLVEMODE:
         switch( value )
         {
         case SOLVEMODE_REAL:
         case SOLVEMODE_AUTO:
         case SOLVEMODE_RATIONAL:
            break;
         default:
            return false;
         }
         break;

      // mode for a posteriori feasibility checks; nothing to do but change the value if valid
      case CHECKMODE:
         switch( value )
         {
         case CHECKMODE_REAL:
         case CHECKMODE_AUTO:
         case CHECKMODE_RATIONAL:
            break;
         default:
            return false;
         }
         break;

      // type of ratio test
      case RATIOTESTER:
         switch( value )
         {
         case RATIOTESTER_TEXTBOOK:
            SoPlex::_solver.setTester(&SoPlex::_ratiotesterTextbook);
            break;
         case RATIOTESTER_HARRIS:
            SoPlex::_solver.setTester(&SoPlex::_ratiotesterHarris);
            break;
         case RATIOTESTER_FAST:
            SoPlex::_solver.setTester(&SoPlex::_ratiotesterFast);
            break;
         case RATIOTESTER_BOUNDFLIPPING:
            SoPlex::_solver.setTester(&SoPlex::_ratiotesterBoundFlipping);
            break;
         default:
            return false;
         }
         break;

      // mode of hyper pricing
      case HYPER_PRICING:
         switch( value )
         {
         case HYPER_PRICING_OFF:
         case HYPER_PRICING_AUTO:
         case HYPER_PRICING_ON:
            break;
         default:
            return false;
         }
         break;

      default:
         return false;
      }

      _intParamValues[param] = value;
      return true;
   }



   /// sets real parameter value; returns true on success
   bool setRealParam(const RealParam param, const Real value, const bool quiet, const bool init)
   {
      assert(param >= 0);
      assert(param < REALPARAM_COUNT);
      assert(init || SoPlex::_isConsistent());

      if( !init && value == realParam(param) )
         return true;

      if( value < _realParamLower[param] || value > _realParamUpper[param] )
         return false;

      switch( param )
      {
      // primal feasibility tolerance
      case FEASTOL:
         SoPlex::_solver.setFeastol(value);
         break;

      // dual feasibility tolerance
      case OPTTOL:
         SoPlex::_solver.setOpttol(value);
         break;

      // general zero tolerance
      case EPSILON_ZERO:
         Param::setEpsilon(value);
         break;

      // zero tolerance used in factorization
      case EPSILON_FACTORIZATION:
         Param::setEpsilonFactorization(value);
         break;

      // zero tolerance used in update of the factorization
      case EPSILON_UPDATE:
         Param::setEpsilonUpdate(value);
         break;

      // pivot zero tolerance used in factorization (declare numerical singularity for small LU pivots)
      case EPSILON_PIVOT:
         Param::setEpsilonPivot(value);
         break;

      // infinity threshold
      case INFTY:
         break;

      // time limit in seconds (INFTY if unlimited)
      case TIMELIMIT:
         break;

      // lower limit on objective value is set in solveReal()
      case OBJLIMIT_LOWER:
         break;

      // upper limit on objective value is set in solveReal()
      case OBJLIMIT_UPPER:
         break;

      // working tolerance for feasibility in floating-point solver
      case FPFEASTOL:
         break;

      // working tolerance for optimality in floating-point solver
      case FPOPTTOL:
         break;

      // maximum increase of scaling factors between refinements
      case MAXSCALEINCR:
         break;

      // lower threshold in lifting (nonzero matrix coefficients with smaller absolute value will be reformulated)
      case LIFTMINVAL:
         break;

      // upper threshold in lifting (nonzero matrix coefficients with larger absolute value will be reformulated)
      case LIFTMAXVAL:
         break;

      // threshold for sparse pricing
      case SPARSITY_THRESHOLD:
         break;

      default:
         return false;
      }

      _realParamValues[param] = value;
      return true;
   }



#ifdef SOPLEX_WITH_RATIONALPARAM
   /// sets rational parameter value; returns true on success
   bool setRationalParam(const RationalParam param, const Rational value, const bool quiet, const bool init)
   {
      assert(param >= 0);
      assert(param < RATIONALPARAM_COUNT);
      assert(init || SoPlex::_isConsistent());

      if( !init && value == rationalParam(param) )
         return true;

      if( value < _rationalParamLower[param] || value > _rationalParamUpper[param] )
         return false;

      switch( param )
      {
      default:
         // currently, there are no rational-valued parameters
         return false;
      }

      _rationalParamValues[param] = value;
      return true;
   }
#endif



   /// sets parameter settings; returns true on success
   bool setSettings(const Settings& newSettings, const bool quiet, const bool init)
   {
      assert(init || _isConsistent());

      bool success = true;

      *this = newSettings;

      for( int i = 0; i < BOOLPARAM_COUNT; i++ )
         success &= setBoolParam((BoolParam)i, _boolParamValues[i], quiet, init);

      for( int i = 0; i < INTPARAM_COUNT; i++ )
         success &= setIntParam((IntParam)i, _intParamValues[i], quiet, init);

      for( int i = 0; i < REALPARAM_COUNT; i++ )
         success &= setRealParam((RealParam)i, _realParamValues[i], quiet, init);

#ifdef SOPLEX_WITH_RATIONALPARAM
      for( int i = 0; i < RATIONALPARAM_COUNT; i++ )
         success &= setRationalParam((RationalParam)i, _rationalParamValues[i], quiet, init);
#endif

      assert(_isConsistent());

      return success;
   }



   /// print non-default parameter values
   void printUserSettings()
   {
      bool printedValue = false;

      for( int i = 0; i < BOOLPARAM_COUNT; i++ )
      {
         if( _boolParamValues[i] == _boolParamDefault[i] )
            continue;

         spxout, "bool:" << _boolParamName[i] << " = " << (_boolParamValues[i] ? "true\n" : "false\n");
         printedValue = true;
      }

      for( int i = 0; i < INTPARAM_COUNT; i++ )
      {
         if( _intParamValues[i] == _intParamDefault[i] )
            continue;

         spxout, "int:" << _intParamName[i] << " = " << _intParamValues[i] << "\n";
         printedValue = true;
      }

      for( int i = 0; i < REALPARAM_COUNT; i++ )
      {
         if( _realParamValues[i] == _realParamDefault[i] )
            continue;

         spxout, "real:" << _realParamName[i] << " = " << _realParamValues[i] << "\n";
         printedValue = true;
      }

#ifdef SOPLEX_WITH_RATIONALPARAM
      for( int i = 0; i < RATIONALPARAM_COUNT; i++ )
      {
         if( _rationalParamValues[i] == _rationalParamDefault[i] )
            continue;

         spxout, "rational:" << _rationalParamName[i] << " = " << _rationalParamValues[i] << "\n";
         printedValue = true;
      }
#endif
      if( printedValue )
         spxout, std::endl;
   }



   /// writes settings file; returns true on success
   bool saveSettingsFile(const char* filename, const bool onlyChanged) const
   {
      assert(filename != 0);

      std::ofstream file(filename);
      if( !file.good() )
         return false;

      file.setf(std::ios::left);
      file << "# SoPlex version " << SOPLEX_VERSION / 100 << "." << (SOPLEX_VERSION / 10) % 10 << "." << SOPLEX_VERSION % 10 << "." << SOPLEX_SUBVERSION << "\n";

      for( int i = 0; i < BOOLPARAM_COUNT; i++ )
      {
         if( onlyChanged && _boolParamValues[i] == _boolParamDefault[i] )
            continue;

         file << "\n";
         file << "# " << _boolParamDescription[i] << "\n";
         file << "# range {true, false}, default " << (_boolParamDefault[i] ? "true\n" : "false\n");
         file << "bool:" << _boolParamName[i] << " = " << (_boolParamValues[i] ? "true\n" : "false\n");
      }

      for( int i = 0; i < INTPARAM_COUNT; i++ )
      {
         if( onlyChanged && _intParamValues[i] == _intParamDefault[i] )
            continue;

         file << "\n";
         file << "# " << _intParamDescription[i] << "\n";
         file << "# range [-2147483648,2147483647], default " << _intParamDefault[i] << "\n";
         file << "int:" << _intParamName[i] << " = " << _intParamValues[i] << "\n";
      }

      for( int i = 0; i < REALPARAM_COUNT; i++ )
      {
         if( onlyChanged && _realParamValues[i] == _realParamDefault[i] )
            continue;

         file << "\n";
         file << "# " << _realParamDescription[i] << "\n";
         file << "# range [" << _realParamLower[i] << "," << _realParamUpper[i]
            << "], default " << _realParamDefault[i] << "\n";
         file << "real:" << _realParamName[i] << " = " << _realParamValues[i] << "\n";
      }

      return true;
   }



   /// reads settings file; returns true on success
   bool loadSettingsFile(const char* filename)
   {
      assert(filename != 0);

      // start timing
      SoPlex::_statistics->readingTime.start();

      MSG_INFO1( spxout, "Loading settings file <" << filename << "> . . .\n" );

      // open file
      spxifstream file(filename);

      if( !file )
      {
         MSG_ERROR( "Error opening settings file.\n" );
         return false;
      }

      // read file
      char line[SET_MAX_LINE_LEN];
      int lineNumber = 0;
      bool readError = false;
      bool parseError = false;

      while( !readError && !parseError)
      {
         lineNumber++;
         readError = !file.getline(line, sizeof(line));
         if( !readError )
            parseError = !_parseSettingsLine(line, lineNumber);
      }
      readError = readError && !file.eof();

      if( readError && strlen(line) == SET_MAX_LINE_LEN - 1 )
      {
         MSG_ERROR( "Error reading settings file: line " << lineNumber << " in settings file exceeds " << SET_MAX_LINE_LEN - 2 << " characters.\n" );
      }
      else if( readError )
      {
         MSG_ERROR( "Error reading settings file: line " << lineNumber << ".\n" );
      }

      // stop timing
      SoPlex::_statistics->readingTime.stop();

      return !readError && !parseError;
   }

   /// parses one setting string and returns true on success
   bool parseSettingsString(char* string)
   {
      assert(string != 0);
      if( string == 0 )
         return false;

      char parseString[SET_MAX_LINE_LEN];
      strncpy(parseString, string, SET_MAX_LINE_LEN);

      char* line = parseString;

      // find the start of the parameter type
      while( *line == ' ' || *line == '\t' || *line == '\r' )
         line++;
      if( *line == '\0' || *line == '\n' || *line == '#' )
         return true;
      char* paramTypeString = line;

      // find the end of the parameter type
      while( *line != ' ' && *line != '\t' && *line != '\r' && *line != '\n' && *line != '#' && *line != '\0' && *line != ':' )
         line++;
      if( *line == ':' )
      {
         *line = '\0';
         line++;
      }
      else
      {
         *line = '\0';
         line++;

         // search for the ':' char in the line
         while( *line == ' ' || *line == '\t' || *line == '\r' )
            line++;
         if( *line != ':' )
         {
            MSG_ERROR( "Error parsing setting string: no ':' separating parameter type and name.\n" );
            return false;
         }
         line++;
      }

      // find the start of the parameter name
      while( *line == ' ' || *line == '\t' || *line == '\r' )
         line++;
      if( *line == '\0' || *line == '\n' || *line == '#' )
      {
         MSG_ERROR( "Error parsing setting string: no parameter name.\n");
         return false;
      }
      char* paramName = line;

      // find the end of the parameter name
      while( *line != ' ' && *line != '\t' && *line != '\r' && *line != '\n' && *line != '#' && *line != '\0' && *line != '=' )
         line++;
      if( *line == '=' )
      {
         *line = '\0';
         line++;
      }
      else
      {
         *line = '\0';
         line++;

         // search for the '=' char in the line
         while( *line == ' ' || *line == '\t' || *line == '\r' )
            line++;
         if( *line != '=' )
         {
            MSG_ERROR( "Error parsing setting string: no '=' after parameter name.\n" );
            return false;
         }
         line++;
      }

      // find the start of the parameter value string
      while( *line == ' ' || *line == '\t' || *line == '\r' )
         line++;
      if( *line == '\0' || *line == '\n' || *line == '#' )
      {
         MSG_ERROR( "Error parsing setting string: no parameter value.\n");
         return false;
      }
      char* paramValueString = line;

      // find the end of the parameter value string
      while( *line != ' ' && *line != '\t' && *line != '\r' && *line != '\n' && *line != '#' && *line != '\0' )
         line++;
      if( *line != '\0' )
      {
         // check, if the rest of the line is clean
         *line = '\0';
         line++;
         while( *line == ' ' || *line == '\t' || *line == '\r' )
            line++;
         if( *line != '\0' && *line != '\n' && *line != '#' )
         {
            MSG_ERROR( "Error parsing setting string: additional character '" << *line << "' after parameter value.\n" );
            return false;
         }
      }

      // check whether we have a bool parameter
      if( strncmp(paramTypeString, "bool", 4) == 0 )
      {
         for( int param = 0; ; param++ )
         {
            if( param >= BOOLPARAM_COUNT )
            {
               MSG_ERROR( "Error parsing setting string: unknown parameter name <" << paramName << ">.\n" );
               return false;
            }
            else if( strncmp(paramName, _boolParamName[param].c_str(), SET_MAX_LINE_LEN) == 0 )
            {
               if( strncasecmp(paramValueString, "true", 4) == 0
                  || strncasecmp(paramValueString, "TRUE", 4) == 0
                  || strncasecmp(paramValueString, "t", 4) == 0
                  || strncasecmp(paramValueString, "T", 4) == 0
                  || strtol(paramValueString, NULL, 4) == 1 )
               {
                  setBoolParam((BoolParam)param, true);
                  break;
               }
               else if( strncasecmp(paramValueString, "false", 5) == 0
                  || strncasecmp(paramValueString, "FALSE", 5) == 0
                  || strncasecmp(paramValueString, "f", 5) == 0
                  || strncasecmp(paramValueString, "F", 5) == 0
                  || strtol(paramValueString, NULL, 5) == 0 )
               {
                  setBoolParam((BoolParam)param, false);
                  break;
               }
               else
               {
                  MSG_ERROR( "Error parsing setting string: invalid value <" << paramValueString << "> for bool parameter <" << paramName << ">.\n" );
                  return false;
               }
            }
         }

         return true;
      }

      // check whether we have an integer parameter
      if( strncmp(paramTypeString, "int", 3) == 0 )
      {
         for( int param = 0; ; param++ )
         {
            if( param >= INTPARAM_COUNT )
            {
               MSG_ERROR( "Error parsing setting string: unknown parameter name <" << paramName << ">.\n" );
               return false;
            }
            else if( strncmp(paramName, _intParamName[param].c_str(), SET_MAX_LINE_LEN) == 0 )
            {
               int value;

               if( sscanf(paramValueString, "%d", &value) == 1 && setIntParam((IntParam)param, value) )
                  break;
               else
               {
                  MSG_ERROR( "Error parsing setting string: invalid value <" << paramValueString << "> for int parameter <" << paramName << ">.\n" );
                  return false;
               }
            }
         }

         return true;
      }

      // check whether we have a real parameter
      if( strncmp(paramTypeString, "real", 4) == 0 )
      {
         for( int param = 0; ; param++ )
         {
            if( param >= REALPARAM_COUNT )
            {
               MSG_ERROR( "Error parsing setting string: unknown parameter name <" << paramName << ">.\n" );
               return false;
            }
            else if( strncmp(paramName, _realParamName[param].c_str(), SET_MAX_LINE_LEN) == 0 )
            {
               Real value;

               if( sscanf(paramValueString, "%" REAL_FORMAT, &value) == 1 && setRealParam((RealParam)param, value) )
                  break;
               else
               {
                  MSG_ERROR( "Error parsing setting string: invalid value <" << paramValueString << "> for real parameter <" << paramName << ">.\n" );
                  return false;
               }
            }
         }

         return true;
      }

      MSG_ERROR( "Error parsing setting string: invalid parameter type <" << paramTypeString << "> for parameter <" << paramName << ">.\n" );

      return false;
   }

      /// parses one line in a settings file and returns true on success; note that the string is modified
   bool Settings::_parseSettingsLine(char* line, const int lineNumber)
   {
      assert(line != 0);

      // find the start of the parameter type
      while( *line == ' ' || *line == '\t' || *line == '\r' )
         line++;
      if( *line == '\0' || *line == '\n' || *line == '#' )
         return true;
      char* paramTypeString = line;

      // find the end of the parameter type
      while( *line != ' ' && *line != '\t' && *line != '\r' && *line != '\n' && *line != '#' && *line != '\0' && *line != ':' )
         line++;
      if( *line == ':' )
      {
         *line = '\0';
         line++;
      }
      else
      {
         *line = '\0';
         line++;

         // search for the ':' char in the line
         while( *line == ' ' || *line == '\t' || *line == '\r' )
            line++;
         if( *line != ':' )
         {
            MSG_ERROR( "Error parsing settings file: no ':' separating parameter type and name in line " << lineNumber << ".\n" );
            return false;
         }
         line++;
      }

      // find the start of the parameter name
      while( *line == ' ' || *line == '\t' || *line == '\r' )
         line++;
      if( *line == '\0' || *line == '\n' || *line == '#' )
      {
         MSG_ERROR( "Error parsing settings file: no parameter name in line " << lineNumber << ".\n");
         return false;
      }
      char* paramName = line;

      // find the end of the parameter name
      while( *line != ' ' && *line != '\t' && *line != '\r' && *line != '\n' && *line != '#' && *line != '\0' && *line != '=' )
         line++;
      if( *line == '=' )
      {
         *line = '\0';
         line++;
      }
      else
      {
         *line = '\0';
         line++;

         // search for the '=' char in the line
         while( *line == ' ' || *line == '\t' || *line == '\r' )
            line++;
         if( *line != '=' )
         {
            MSG_ERROR( "Error parsing settings file: no '=' after parameter name in line " << lineNumber << ".\n" );
            return false;
         }
         line++;
      }

      // find the start of the parameter value string
      while( *line == ' ' || *line == '\t' || *line == '\r' )
         line++;
      if( *line == '\0' || *line == '\n' || *line == '#' )
      {
         MSG_ERROR( "Error parsing settings file: no parameter value in line " << lineNumber << ".\n");
         return false;
      }
      char* paramValueString = line;

      // find the end of the parameter value string
      while( *line != ' ' && *line != '\t' && *line != '\r' && *line != '\n' && *line != '#' && *line != '\0' )
         line++;
      if( *line != '\0' )
      {
         // check, if the rest of the line is clean
         *line = '\0';
         line++;
         while( *line == ' ' || *line == '\t' || *line == '\r' )
            line++;
         if( *line != '\0' && *line != '\n' && *line != '#' )
         {
            MSG_ERROR( "Error parsing settings file: additional character '" << *line << "' after parameter value in line " << lineNumber << ".\n" );
            return false;
         }
      }

      // check whether we have a bool parameter
      if( strncmp(paramTypeString, "bool", 4) == 0 )
      {
         for( int param = 0; ; param++ )
         {
            if( param >= BOOLPARAM_COUNT )
            {
               MSG_ERROR( "Error parsing settings file: unknown parameter name <" << paramName << "> in line " << lineNumber << ".\n" );
               return false;
            }
            else if( strncmp(paramName, _boolParamName[param].c_str(), SET_MAX_LINE_LEN) == 0 )
            {
               if( strncasecmp(paramValueString, "true", 4) == 0
                  || strncasecmp(paramValueString, "TRUE", 4) == 0
                  || strncasecmp(paramValueString, "t", 4) == 0
                  || strncasecmp(paramValueString, "T", 4) == 0
                  || strtol(paramValueString, NULL, 4) == 1 )
               {
                  setBoolParam((BoolParam)param, true);
                  break;
               }
               else if( strncasecmp(paramValueString, "false", 5) == 0
                  || strncasecmp(paramValueString, "FALSE", 5) == 0
                  || strncasecmp(paramValueString, "f", 5) == 0
                  || strncasecmp(paramValueString, "F", 5) == 0
                  || strtol(paramValueString, NULL, 5) == 0 )
               {
                  setBoolParam((BoolParam)param, false);
                  break;
               }
               else
               {
                  MSG_ERROR( "Error parsing settings file: invalid value <" << paramValueString << "> for bool parameter <" << paramName << "> in line " << lineNumber << ".\n" );
                  return false;
               }
            }
         }

         return true;
      }

      // check whether we have an integer parameter
      if( strncmp(paramTypeString, "int", 3) == 0 )
      {
         for( int param = 0; ; param++ )
         {
            if( param >= INTPARAM_COUNT )
            {
               MSG_ERROR( "Error parsing settings file: unknown parameter name <" << paramName << "> in line " << lineNumber << ".\n" );
               return false;
            }
            else if( strncmp(paramName, _intParamName[param].c_str(), SET_MAX_LINE_LEN) == 0 )
            {
               int value;

               if( sscanf(paramValueString, "%d", &value) == 1 && setIntParam((IntParam)param, value) )
                  break;
               else
               {
                  MSG_ERROR( "Error parsing settings file: invalid value <" << paramValueString << "> for int parameter <" << paramName << "> in line " << lineNumber << ".\n" );
                  return false;
               }
            }
         }

         return true;
      }

      // check whether we have a real parameter
      if( strncmp(paramTypeString, "real", 4) == 0 )
      {
         for( int param = 0; ; param++ )
         {
            if( param >= REALPARAM_COUNT )
            {
               MSG_ERROR( "Error parsing settings file: unknown parameter name <" << paramName << "> in line " << lineNumber << ".\n" );
               return false;
            }
            else if( strncmp(paramName, _realParamName[param].c_str(), SET_MAX_LINE_LEN) == 0 )
            {
               Real value;

               if( sscanf(paramValueString, "%" REAL_FORMAT, &value) == 1 && setRealParam((RealParam)param, value) )
                  break;
               else
               {
                  MSG_ERROR( "Error parsing settings file: invalid value <" << paramValueString << "> for real parameter <" << paramName << "> in line " << lineNumber << ".\n" );
                  return false;
               }
            }
         }

         return true;
      }

#ifdef SOPLEX_WITH_RATIONALPARAM
      // check whether we have a rational parameter
      if( strncmp(paramTypeString, "rational", 8) == 0 )
      {
         for( int param = 0; ; param++ )
         {
            if( param >= RATIONALPARAM_COUNT )
            {
               MSG_ERROR( "Error parsing settings file: unknown parameter name <" << paramName << "> in line " << lineNumber << ".\n" );
               return false;
            }
            else if( strncmp(paramName, _rationalParamName[param].c_str(), SET_MAX_LINE_LEN) == 0 )
            {
               Rational value;

               if( readStringRational(paramValueString, value) && setRationalParam((RationalParam)param, value) )
                  break;
               else
               {
                  MSG_ERROR( "Error parsing settings file: invalid value <" << paramValueString << "> for rational parameter <" << paramName << "> in line " << lineNumber << ".\n" );
                  return false;
               }
            }
         }

         return true;
      }
#endif

      MSG_ERROR( "Error parsing settings file: invalid parameter type <" << paramTypeString << "> for parameter <" << paramName << "> in line " << lineNumber << ".\n" );

      return false;
   }
}
